/*##############################################################################

nIP - nano IP stack

File        : dhcp.h

Description : Dynamic Host Configuration Protocol

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#ifndef _DHCP_H_
#define _DHCP_H_

#include "nip_init.h"
#include "nip_types.h"
#include "net/net_if.h"
#include "net/udp.h"
#include NIP_ARCH_FILE(endianness.h)

#define NIP_DHCP_INIT_BACKOFF 4 /**< initial delay for exponential backoff algorithm */
#define NIP_DHCP_MAX_BACKOFF  64/**< maximum delay for exponential backoff algorithm */

#define NIP_DHCP_OP_REQUEST   1
#define NIP_DHCP_OP_REPLY     2

#define DHCP_OPTION_PADDING              0
#define DHCP_OPTION_SUBNETMASK           1
#define DHCP_OPTION_ROUTER               3
#define DHCP_OPTION_DNS                  6
#define DHCP_OPTION_HOST_NAME            12
#define DHCP_OPTION_DOMAINNAME           15
#define DHCP_OPTION_ROUTERDSCVR          31
#define DHCP_OPTION_NTPSERVERS           42
#define DHCP_OPTION_REQUEST_IP           50
#define DHCP_OPTION_LEASETIME            51
#define DHCP_OPTION_MSGTYPE              53
#define   DHCP_OPTION_MSGTYPE_DISCOVER     1
#define   DHCP_OPTION_MSGTYPE_OFFER        2
#define   DHCP_OPTION_MSGTYPE_REQUEST      3
#define   DHCP_OPTION_MSGTYPE_DECLINE      4
#define   DHCP_OPTION_MSGTYPE_ACK          5
#define   DHCP_OPTION_MSGTYPE_NAK          6
#define   DHCP_OPTION_MSGTYPE_RELEASE      7
#define DHCP_OPTION_SERVERID             54
#define DHCP_OPTION_PARAMREQLIST         55
#define DHCP_OPTION_RENEWAL_TIME         58
#define DHCP_OPTION_REBINDING_TIME       59
#define DHCP_OPTION_VENDOR_CLASSID       60
#define DHCP_OPTION_CLIENT_IDENTIFIER    61
#define DHCP_OPTION_AUTO_CONFIGURE       116
#define DHCP_OPTION_END                  255


// #define NIP_DHCP_HTYPE_ETH    1
// #define NIP_DHCP_HLEN_ETH     6

/** DHCP Magic Cookie Option in Network Byte Order (decimal 99.130.82.99) */
#if ( ENDIANNESS == BIG_ENDIAN )
#define NIP_DHCP_MAGIC_COOKIE   0x63825363
#else  // ENDIANNESS
#define NIP_DHCP_MAGIC_COOKIE   0x63538263
#endif // ENDIANNESS

typedef enum
{
	NIP_DHCP_STAT_DISABLED = 0,
	NIP_DHCP_STAT_INIT,
	NIP_DHCP_STAT_SELECTING,
	NIP_DHCP_STAT_REQUESTING,
	NIP_DHCP_STAT_BOUND,
	NIP_DHCP_STAT_RENEWING,
	NIP_DHCP_STAT_REBINDING,
	NIP_DHCP_STAT_REBOOTING
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_dhcp_stat_t;

/** structure to evaluate known dhcp options */
struct nip_dhcp_option
{
	uint8_t type;
	uint8_t valsize;
	union
	{
		uint8_t  msgtype;
		uint8_t  ipaddr[4];
		uint32_t time;
	} value;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

struct nip_dhcp_packet
{
	uint8_t op;
	uint8_t htype;
	uint8_t hlen;
	uint8_t hops;
	uint32_t xid;
	uint16_t secs;
	uint8_t flags[2];
	uint8_t ciaddr[4];
	uint8_t yiaddr[4];
	uint8_t siaddr[4];
	uint8_t giaddr[4];
	uint8_t chaddr[16];
	uint8_t sname[64];
	uint8_t file[128];
	uint32_t magic_cookie;
	uint8_t options[60];
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

struct nip_dhcp_config {
	nip_dhcp_stat_t   state;
	uint32_t          xid;
	uint8_t           backoff_delay;
	nip_time_t        leasetime; /** halt network when lease time is reached */
	nip_time_t        t1;  /** renew (unicast) DHCP lease when timer t1 expires */
	nip_time_t        t2;  /** renew (broadcast) DHCP lease when timer t2 expires*/
	nip_time_t        ts;  /** time a packet will have to be resent at */
	struct nip_udp_sock_addr serv_addr; /** address of dhcp server */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

void nip_dhcp_start( nip_net_if_id_t net_if );
void nip_dhcp_disable( nip_net_if_id_t net_if );
void nip_dhcp_disp_check( void );

#endif // _DHCP_H_
