/*##############################################################################

nIP - Testapplication

File        : mdns.h

Description : Multicast Domain Name System

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#ifndef _MDNS_H_
 #define _MDNS_H_

#include "nip_init.h"
#include "nip_types.h"
#include "os_core.h"
#include "mem.h"
#include "net/net_if.h"
#include "app/dns.h"

#define NIP_MDNS_ADDR0  224
#define NIP_MDNS_ADDR1  0
#define NIP_MDNS_ADDR2  0
#define NIP_MDNS_ADDR3  251

#ifndef NIP_MDNS_MAX_QUERIES
	/** maximum number of open MDNS queries. */
	#define NIP_MDNS_MAX_QUERIES    5
#endif

#ifndef NIP_MDNS_DISCARD_ON_OUT_OF_MEMORY
	/** If this is set to 1 we won't wait for memory to become free, but discard
	 * any incoming packet if we don't have enough memory available to serve it */
	#define NIP_MDNS_DISCARD_ON_OUT_OF_MEMORY 1
#endif

#ifndef NIP_MDNS_MAX_NAMESTRINGS
	/** maximum number of cached DNS name strings. The hereby meant name strings
	 * are the parts in between two dots of a DNS name, not the entire name. */
	#define NIP_MDNS_MAX_NAMESTRINGS 20
#endif

#ifndef NIP_MDNS_MAX_CACHE_SIZE
	/** maximum number of entries in MDNS cache */
	#define NIP_MDNS_MAX_CACHE_SIZE  20
#endif

#ifndef NIP_MDNS_DEFEND_HOSTNAME
	/** Should mdns-hostname be unique in the network? */
	#define NIP_MDNS_DEFEND_HOSTNAME 1
#endif

#ifndef NIP_MDNS_MAX_NAME_DEPTH
	/** The currently used system maximizes the size of MDNS names to 6 dotted
	 * parts which is the maximum size needed for IPv4 in-addr.arpa pointers.
	 * More flexible implementations would require a more complexly programmed
	 * system of name concatenation.
	 */
	#define NIP_MDNS_MAX_NAME_DEPTH 6
#endif

#ifndef NIP_MDNS_CREATE_IN_ADDR_ARPA_PTR
	#define NIP_MDNS_CREATE_IN_ADDR_ARPA_PTR 1
#endif

// double check: we cannot create in-addr.arpa pointers if max-name-depth < 6
#if ( NIP_MDNS_MAX_NAME_DEPTH < 6 )
	#undef NIP_MDNS_CREATE_IN_ADDR_ARPA_PTR
	#define NIP_MDNS_CREATE_IN_ADDR_ARPA_PTR 0
#endif

#ifndef NIP_MDNS_COUNTER_SIZE
	/** Byte-Size of temporary query and answer counters for transmission. DNS
	 * defines these counters to be 16-Bit, but for internal use we can limit
	 * them to 8-Bit, as we can -- depending on available memory -- assure, that
	 * a counter value of 255 will never be exceeded.
	 * If by any chance any query/answer counter might exceed 255 you SHOULD set
	 * NIP_MDNS_COUNTER_SIZE to 2, as the program does not perform any boundary
	 * checks. Setting it to 2 will only lead to 4 additional Bytes of RAM being
	 * used, which will most probbly not hurt, if you feel able to transmit more
	 * than 255 queries or answers in one packet.
	 */
	#define NIP_MDNS_COUNTER_SIZE  1
#endif

/** Query timer with this value is disabled until set to a different value. */
#define NIP_MDNS_NO_QUERY_TIMER 255

/** Maximum number of probes to send until assuming no conflict */
#define NIP_MDNS_PROBE_COUNT    3

/** Maximum number of announcements to send until assuming no conflict */
#define NIP_MDNS_ANNOUNCE_COUNT 3

#define NIP_MDNS_DEFAULT_TTL_A 12  /**< default TTL for records with a host name \
                                     *in rname or rdata, 2min */
#define NIP_MDNS_DEFAULT_TTL   4500 /**< default TTL for not-A-records, 75min */


#if   ( NIP_MDNS_MAX_QUERIES < 0x3F   )
	typedef uint8_t nip_mdns_query_id_t;
	typedef uint8_t nip_mdns_query_cnt_t;
	#define NIP_MDNS_QUERY_ID_MASK       0x3F
	#define NIP_MDNS_NO_QUERY            0x3F
	#define NIP_MDNS_QUERY_DELETE        0x80
	#define NIP_MDNS_QUERY_ACTIVE        0x00
	#define NIP_MDNS_NFY_NONE            0x00
	#define NIP_MDNS_NFY_LOCAL           0x80
	#define NIP_MDNS_NFY_REMOTE          0x40
#elif ( NIP_MDNS_MAX_QUERY < 0x7FFF )
	typedef uint8_t nip_mdns_query_id_t;
	typedef uint16_t nip_mdns_query_cnt_t;
	#define NIP_MDNS_QUERY_ID_MASK       0x3FFF
	#define NIP_MDNS_NO_QUERY            0x3FFF
	#define NIP_MDNS_QUERY_DELETE        0x8000
	#define NIP_MDNS_QUERY_ACTIVE        0x0000
	#define NIP_MDNS_NFY_LOCAL           0x8000
	#define NIP_MDNS_NFY_REMOTE          0x4000
#else
	typedef uint8_t nip_mdns_query_id_t;
	typedef uint32_t nip_mdns_query_cnt_t;
	#define NIP_MDNS_QUERY_ID_MASK       0x3FFFFFFF
	#define NIP_MDNS_NO_QUERY            0x3FFFFFFF
	#define NIP_MDNS_QUERY_DELETE        0x80000000
	#define NIP_MDNS_QUERY_ACTIVE        0x00000000
	#define NIP_MDNS_NFY_LOCAL           0x80000000
	#define NIP_MDNS_NFY_REMOTE          0x40000000
#endif

#define NIP_MDNS_NFY_ALL        (NIP_MDNS_NFY_LOCAL|NIP_MDNS_NFY_REMOTE)
#define NIP_MDNS_NFY_NONE       0x00

#define NIP_MDNS_NO_NAMESTRING     0
#if   ( NIP_MDNS_MAX_NAMESTRINGS < 0xFF   )
	typedef uint8_t nip_mdns_namestring_id_t;
	typedef uint8_t nip_mdns_namestring_cnt_t;
#elif ( NIP_MDNS_MAX_NAMESTRINGS < 0xFFFF )
	typedef uint16_t nip_mdns_namestring_id_t;
	typedef uint16_t nip_mdns_namestring_cnt_t;
#else
	typedef uint32_t nip_mdns_namestring_id_t;
	typedef uint32_t nip_mdns_namestring_cnt_t;
#endif

/** Per standard, the size of the name string length is limited to 6 Bit */
typedef uint8_t nip_mdns_namestring_len_t;

#define NIP_MDNS_NO_CACHE_ID   0
#if   ( NIP_MDNS_MAX_CACHE_SIZE < 0xFF   )
	typedef uint8_t nip_mdns_cache_id_t;
	typedef uint8_t nip_mdns_cache_cnt_t;
#elif ( NIP_MDNS_MAX_CACHE_SIZE < 0xFFFF )
	typedef uint16_t nip_mdns_cache_id_t;
	typedef uint16_t nip_mdns_cache_cnt_t;
#else
	typedef uint32_t nip_mdns_cache_id_t;
	typedef uint32_t nip_mdns_cache_cnt_t;
#endif

/** Size of usage counter assumes that average usage of a namestring will not
 * exceed two occurances per cache entry and one per query. If that assumption
 * fails to be true, NIP_E_OUT_OF_MEMORY errors will occur upon registering
 * names beyond that point.
 */
#define NIP_MDNS_MAX_NAMESTRING_USG (NIP_MDNS_MAX_CACHE_SIZE/2 + NIP_MDNS_MAX_QUERIES)
#if   ( NIP_MDNS_MAX_NAMESTRING_USG < 0xFF   )
	typedef uint8_t nip_mdns_namestring_usg_t;
#elif ( NIP_MDNS_MAX_NAMESTRING_USG < 0XFFFF )
	typedef uint16_t nip_mdns_namestring_usg_t;
#else
	typedef uint32_t nip_mdns_namestring_usg_t;
#endif


struct nip_mdns_name
{
	nip_mdns_namestring_id_t part[ NIP_MDNS_MAX_NAME_DEPTH ];
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

struct nip_mdns_rdata_a
{
	uint8_t ip_addr[4];
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;
/** PTR data entry only holds instance name as app, proto and domain are implicit */
#define NIP_MDNS_RDATA_SRV_SIZE  (3*sizeof(uint16_t)+sizeof(struct nip_mdns_data_a))
struct nip_mdns_rdata_srv
{
	uint16_t  priority;
	uint16_t  weight;
	uint16_t  port;
	struct nip_mdns_name target;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

/** PTR data entry only holds instance name as app, proto and domain are implicit */
#define NIP_MDNS_RDATA_PTR_SIZE  sizeof(nip_mdns_namestring_id_t)
struct nip_mdns_rdata_ptr
{
	struct nip_mdns_name name;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

struct nip_mdns_rdata_oth
{
	nip_mem_handle_t string;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;


typedef union
{
	struct nip_mdns_rdata_a   a;
	struct nip_mdns_rdata_srv srv;
	struct nip_mdns_rdata_ptr ptr;
	struct nip_mdns_rdata_oth oth;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_rdata_t;

typedef enum
{
	NIP_MDNS_TYPE_A     =  1,
	NIP_MDNS_TYPE_NS    =  2, // currently not used
	NIP_MDNS_TYPE_CNAME =  5, // currently not used
	NIP_MDNS_TYPE_SOA   =  6, // currently not used
	NIP_MDNS_TYPE_PTR   = 12,
	NIP_MDNS_TYPE_MX    = 15,  // currently not used
	NIP_MDNS_TYPE_TXT   = 16,
	NIP_MDNS_TYPE_AAAA  = 28,  // currently not used
	NIP_MDNS_TYPE_SRV   = 33,
	NIP_MDNS_TYPE_ANY   = 255
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_type_t;

typedef enum
{
	NIP_MDNS_STAT_DOWN = 0, /**< mDNS inactive */
	NIP_MDNS_STAT_INIT,     /**< mDNS trying to go up, initialization pending */
	NIP_MDNS_STAT_UP,       /**< mDNS up -> statemachines running */
	NIP_MDNS_STAT_CACHE_CHECK,
	NIP_MDNS_STAT_QUERY_CHECK,
	NIP_MDNS_STAT_RESET,    /**< mDNS waiting for reset */
	NIP_MDNS_STAT_BYE       /**< mDNS going down */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_status_t;

/** All Cache entries with the NIP_MDNS_CSTAT_AUTHORATIVE_BIT status bit set are
 * considered authorative and will be used to reply to remote queries and will
 * be defended upon conflicts. */
#define NIP_MDNS_CSTAT_AUTH_BIT    0x01
#define NIP_MDNS_CSTAT_AUTH_MASK   0x0F
/** All Cache entries with the NIP_MDNS_CSTAT_LEARNED_BIT status bit set are
 * considered learned and bound to one or more local queries. */
#define NIP_MDNS_CSTAT_LEARNED_BIT        0x80
#define NIP_MDNS_CSTAT_LEARNED_MASK       0xF0
typedef enum
{
	NIP_MDNS_CSTAT_PROBING       = 0x02, /**< item is being registered */
	NIP_MDNS_CSTAT_PROBECONFLICT = 0x04, /**< probing failed, select new name */
	NIP_MDNS_CSTAT_AUTHORATIVE   = 0x01, /**< item has been registered but must
	                                      * not be defended (need not be unique) */
	NIP_MDNS_CSTAT_UNIQUE_AUTH   = 0x09, /**< item has been registered and must
	                                      * be defended (is unique) */
	NIP_MDNS_CSTAT_ANNOUNCING    = 0x03, /**< item is being announced */
	NIP_MDNS_CSTAT_CONFLICT      = 0x05, /**< conflict on authorative item */
	NIP_MDNS_CSTAT_DELETE        = 0x00, /**< item is scheduled for deletion -> notify
	                                       * subscribers */
	NIP_MDNS_CSTAT_LEARNED   = 0x80, /**< item has been learned through MDNS */
	NIP_MDNS_CSTAT_REFRESH   = 0x90, /**< item timed out and we're trying to refresh */
	NIP_MDNS_CSTAT_OUTDATED1 = 0xA0, /**< someone else asked for this item */
	NIP_MDNS_CSTAT_OUTDATED2 = 0xB0, /**< someone asked for this item again, previous question unanswered */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_cache_status_t;

typedef enum
{
	NIP_MDNS_TASK_UNREGISTER = 0x00,  /**< unregister authorative record */
	NIP_MDNS_TASK_REGISTER = 0x01,  /**< register authorative record */
	NIP_MDNS_TASK_REGUNIQUE= 0x11,  /**< register UNIQUE authorative record */
	NIP_MDNS_TASK_LEARN    = 0x20,  /**< learning remote record */
	NIP_MDNS_TASK_SEARCH   = 0x30,  /**< search record */
	NIP_MDNS_TASK_GET      = 0x80,  /**< return record data for given, learned Cache ID */
	NIP_MDNS_TASK_GET_LEARNED=0x90, /**< return record data for given, authorative Cache ID */
	NIP_MDNS_TASK_RESET    = 0x40   /**< reset cache */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_task_t;

typedef enum
{
	NIP_MDNS_NFYTASK_NOTIFY      = 0x01,/**< notify query about cache change */
	NIP_MDNS_NFYTASK_SUBSCRIBE   = 0x11,/**< add cache subscription */
	NIP_MDNS_NFYTASK_UNSUBSCRIBE = 0x10,/**< remove subscription    */
	NIP_MDNS_NFYTASK_NOTIFIED    = 0x20 /**< remove notification marker */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_nfy_task_t;

typedef enum
{
	NIP_MDNS_NTASK_FIND,        /**< check if name from query is in cache */
	NIP_MDNS_NTASK_WRITE,       /**< write name to query */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_name_task_t;

#define NIP_MDNS_QTYPE_TX_BIT     0x80
#define NIP_MDNS_QTYPE_LOCAL_BIT  0x01
#define NIP_MDNS_QTYPE_REMOTE_BIT 0x02

/** Query types. The enumeration values are chosen so that the lower two bits
 * decide what packet type will result from the query. That is: LOCAL, PROBE and
 * REMLOC queries will result in a Query-Packet being sent when the timer triggers.
 * REMOTE and REMLOC queries will result in a Response-Packet being sent, whenever
 * there are active Remote-Subscriptions. The High-order Bit indicates Queries
 * that are actively transmitted on the network.
 */
typedef enum
{
	NIP_MDNS_QTYPE_LOCAL    = 0x81, /**< local query that we're caching remote records for */
	NIP_MDNS_QTYPE_REMOTE   = 0x02, /**< remote query that we have authorative records for */
	NIP_MDNS_QTYPE_REMLOC   = 0x83, /**< query is remote and local */
	NIP_MDNS_QTYPE_PROBE    = 0x82, /**< probe query */
	NIP_MDNS_QTYPE_ANNOUNCE = 0x12  /**< Fake remote query to collect announcement answers */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_query_type_t;

/** The lower two Bits of Packet type describe wether a Query or a Response
 * packet will be sent. The higher Bits are used to code additional information
 * that is needed to determine timing and answer coding. */
typedef enum
{
	NIP_MDNS_PTYPE_NONE          = 0x00,
	NIP_MDNS_PTYPE_QUERY         = 0x01,
	NIP_MDNS_PTYPE_RESPONSE      = 0x02,
	NIP_MDNS_PTYPE_PROBE         = 0x11,
	NIP_MDNS_PTYPE_ANSWCONTINUED = 0x21,/**< continuation of known answers */
	NIP_MDNS_PTYPE_PROBERESPONSE = 0x12 /**< special type to signal urgency.*/
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_packet_type_t;

struct nip_mdns_query
{
	nip_mdns_query_id_t   id;       /**< unique query ID */
	nip_mdns_query_type_t type;     /**< query type */
	uint8_t               timer;    /**< transmit when timer reaches 0 */
	struct nip_mdns_name  rr_name;  /**< resource record name */
	nip_mdns_type_t       rr_type;  /**< resource record type */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

#define NIP_MDNS_STRINGOFF (sizeof(nip_mdns_namestring_id_t)+sizeof(nip_mdns_namestring_usg_t))
struct nip_mdns_namestring
{
	nip_mdns_namestring_id_t   id;   /**< unique namestring ID */
	nip_mdns_namestring_usg_t  usg;  /**< namestring usage counter */
	struct nip_dns_string      str;  /**< dns string */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;
#define NIP_MDNS_BASE_RECORD_SIZE (sizeof(struct nip_mdns_record)-sizeof(nip_mdns_rdata_t))
#define NIP_MDNS_OTH_RECORD_SIZE (NIP_MDNS_BASE_RECORD_SIZE+sizeof(struct nip_mdns_rdata_oth))
#define NIP_MDNS_A_RECORD_SIZE   (NIP_MDNS_BASE_RECORD_SIZE+sizeof(struct nip_mdns_rdata_a  ))
#define NIP_MDNS_SRV_RECORD_SIZE (NIP_MDNS_BASE_RECORD_SIZE+sizeof(struct nip_mdns_rdata_srv))
#define NIP_MDNS_PTR_RECORD_SIZE (NIP_MDNS_BASE_RECORD_SIZE+sizeof(struct nip_mdns_rdata_ptr))
typedef uint8_t nip_mdns_record_size_t;

struct nip_mdns_record
{
	nip_mdns_type_t       type; /**< DNS type of record */
	struct nip_mdns_name  name; /**< DNS name of record, depending on type */
	uint16_t              ttl;  /**< time to live */

	// "data" MUST be last attribute of nip_mdns_record
	nip_mdns_rdata_t      data; /**< record body, depending on type */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

#define NIP_MDNS_BASE_CACHE_SIZE (sizeof(struct nip_mdns_cache)-sizeof(struct nip_mdns_record))
struct nip_mdns_cache
{
	nip_mdns_cache_id_t id;
	uint16_t            age;  /**< age of entry. entry will be deleted if age
	                            * exceeds record.ttl */
	nip_mdns_cache_status_t stat;
	struct nip_mdns_record record;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

struct nip_mdns_cache_subscription
{
	nip_mdns_cache_id_t   c_id; /**< cache id */
	nip_mdns_query_id_t   q_id; /**< query id */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

#if NIP_MDNS_COUNTER_SIZE == 1
	typedef uint8_t nip_mdns_qa_counter_t;
#else
	typedef uint16_t nip_mdns_qa_counter_t;
#endif

#define NIP_MDNS_QA_COUNTERS_OFFSET 4 /**< counter offset in dns header */
#define NIP_MDNS_QA_COUNTERS_NUMBER 4 /**< number of query/answer counters */
/** Section types in DNS packet. The values are also used as qa_count-array
 * item selectors and as 16-bit offsets in DNS header starting at the
 * NIP_MDNS_QA_COUNTERS_OFFSET position of the header. DO NOT ALTER THE VALUES
 * unless you change a whole bunch of code along with it. */
typedef enum
{
	NIP_MDNS_QDCOUNT_OFFSET = 0,
	NIP_MDNS_ANCOUNT_OFFSET = 1,
	NIP_MDNS_NSCOUNT_OFFSET = 2,
	NIP_MDNS_ARCOUNT_OFFSET = 3
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_qa_offset_t;

typedef enum
{
	NIP_MDNS_PTASK_INC_COUNTER, /**< increment one of the section counters. */
	NIP_MDNS_PTASK_SECTION_PTR, /**< move pointer to specific section */
	NIP_MDNS_PTASK_SET_TC,      /**< set truncated flag (TC-flag) in header */
	NIP_MDNS_PTASK_NEXT_ITEM    /**< forward pointer to next item in section */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mdns_packet_task_t;



struct nip_mdns_check
{
	nip_mdns_status_t      stat; /**< state of mdns-check function */
	nip_mdns_packet_type_t recv; /**< type of packet in receive queue */
	uint16_t               recv_size;
	nip_mdns_packet_type_t send; /**< type of packet in send queue */
// 	nip_mdns_qa_counter_t  qa_count[ NIP_MDNS_QA_COUNTERS_NUMBER ];
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

// UNUSED /** mDNS configuration structure */
// struct nip_mdns_config
// {
// 	nip_mdns_status_t        stat;
// 	nip_mdns_cache_id_t      hostname; /** ID of hostname cachid */
// 	nip_mdns_namestring_id_t domain;   /** registration of local domain-name */
// 	nip_mem_handle_t         cs;       /** cache subscriptions */
// 	nip_mem_handle_t         cache;    /** DNS cache */
// 	nip_mem_handle_t         names;    /** DNS name strings */
// 	nip_mem_handle_t         queries;  /** DNS queries */
// }
// #ifndef DOXYGEN
// __attribute__((packed))
// #endif
// ;
/** @name Macros for Management of DNS names */
//@{
/** Register name string. That is just one part of a name.
 * @return namestring ID */
#define NIP_MDNS_REG_STRING( name_ptr ) nip_mdns_register_name( NULL, 0, name_ptr, 1 )
/** Register dns-encoded name. Name will be decoded from name_ptr. At most idcnt
 * part-IDs will be written to idarr.
 * @return ID of last part on SUCCESS */
#define NIP_MDNS_REG_NAME( idarr, idcnt, name_ptr ) nip_mdns_register_name( idarr, idcnt, name_ptr, 1 )

/** Increment name usage. Usage of at most idcnt items in idarr will be increased by 1
 * @return NIP_MDNS_NO_NAMESTRING on error*/
#define NIP_MDNS_NAME_USE( idarr, idcnt ) nip_mdns_register_name( idarr, idcnt, NULL, 1 )
/** Unregister name occurence. Usage of at most idcnt items in idarr will be decreased by 1
 * @return NIP_MDNS_NO_NAMESTRING on error */
#define NIP_MDNS_UNREG_NAME( idarr, idcnt ) nip_mdns_register_name( idarr, idcnt, NULL, -1 )
/** Does name exist in list?
 * @return NIP_MDNS_NO_NAMESTRING on failure*/
#define NIP_MDNS_CHECK_NAME( name_ptr ) nip_mdns_register_name( NULL, 0, name_ptr, 0 )
/** Write mdns-encoded name to buffer. At most idcnt items of idarr will be
 * concatenated and written to buffer.
 * @todo has yet to be implemented
 * @return NIP_MDNS_NO_NAMESTRING on error */
#define NIP_MDNS_WRITE_NAME( idarr, idcnt, buf_ptr ) nip_mdns_register_name( idarr, idcnt, buf_ptr, 0 )
//@}

/** @name Macros for Management of DNS Queries */
//@{
/** Send PROBE request for record */
#define NIP_MDNS_PROBE( name ) \
nip_mdns_query( NIP_MDNS_QTYPE_PROBE, name, NIP_MDNS_TYPE_ANY )

/** Send Announcements for record */
#define NIP_MDNS_ANNOUNCE( name ) \
nip_mdns_query( NIP_MDNS_QTYPE_ANNOUNCE, name, NIP_MDNS_TYPE_ANY )

#define NIP_MDNS_SUBSCRIBE_REMOTE( c_id, q_id ) \
nip_mdns_notify( NIP_MDNS_NFYTASK_SUBSCRIBE, c_id, q_id, NIP_MDNS_NFY_REMOTE )

#define NIP_MDNS_UNSUBSCRIBE_QUERY( q_id ) \
nip_mdns_notify( NIP_MDNS_NFYTASK_UNSUBSCRIBE, NIP_MDNS_NO_CACHE_ID, q_id, NIP_MDNS_NFY_ALL )

/** Cache subscriptions for (Remote) Queries will be removed upon notification */
#define NIP_MDNS_QUERY_NOTIFIED( c_id, q_id ) \
nip_mdns_notify( NIP_MDNS_NFYTASK_NOTIFIED, c_id, q_id, NIP_MDNS_NFY_REMOTE )

//@}

/** @name Macros for Management of DNS Cache */
//@{
#define NIP_MDNS_REGISTER( cache_id_ptr, record_ptr ) \
nip_mdns_cache_task( NIP_MDNS_TASK_REGISTER, cache_id_ptr, record_ptr )

#define NIP_MDNS_UNREGISTER( cache_id_ptr ) \
nip_mdns_cache_task( NIP_MDNS_TASK_UNREGISTER, cache_id_ptr, NULL )

#define NIP_MDNS_REGUNIQUE( cache_id_ptr, record_ptr ) \
nip_mdns_cache_task( NIP_MDNS_TASK_REGUNIQUE, cache_id_ptr, record_ptr )

#define NIP_MDNS_RESET_CACHE() \
nip_mdns_cache_task( NIP_MDNS_TASK_RESET, NULL, NULL )

#define NIP_MDNS_GET_RECORD( cache_id_ptr, record_ptr ) \
nip_mdns_cache_task( NIP_MDNS_TASK_GET_LEARNED, cache_id_ptr, record_ptr )

#define NIP_MDNS_GET_LEARNED_CSTAT( item_ptr ) \
((item_ptr)->stat & NIP_MDNS_CSTAT_LEARNED_MASK)

#define NIP_MDNS_GET_AUTH_CSTAT( item_ptr ) \
((item_ptr)->stat & NIP_MDNS_CSTAT_AUTH_MASK)

#define NIP_MDNS_SET_LEARNED_CSTAT( item_ptr, newstat ) \
(item_ptr)->stat = ((item_ptr)->stat & ~NIP_MDNS_CSTAT_LEARNED_MASK) | newstat

#define NIP_MDNS_SET_AUTH_CSTAT( item_ptr, newstat ) \
(item_ptr)->stat = ((item_ptr)->stat & ~NIP_MDNS_CSTAT_LEARNED_MASK) | newstat

//@}


extern nip_mdns_query_id_t      nip_mdns_hostname_queries;


nip_error_t nip_mdns_init( void );

void nip_mdns_interface_up( nip_net_if_id_t if_id );
void nip_mdns_interface_down( void );

nip_mdns_query_id_t nip_mdns_query( nip_mdns_query_type_t qtype, struct nip_mdns_name *name, nip_mdns_type_t type );
nip_mdns_cache_id_t nip_mdns_fetch_results( nip_mdns_query_id_t r_id );
nip_error_t nip_mdns_close( nip_mdns_query_id_t r_id );

nip_mdns_namestring_id_t nip_mdns_register_name( nip_mdns_namestring_id_t *id, uint8_t cnt, nip_mem_ptr_t *name_ptr, nip_mdns_namestring_usg_t usg );

struct nip_mdns_cache *nip_mdns_cache_ptr( nip_mdns_cache_id_t id );
nip_error_t nip_mdns_cache_task( nip_mdns_task_t task, nip_mdns_cache_id_t *id, struct nip_mdns_record *record );
void nip_mdns_disp_check( void );

#endif /* _MDNS_H_ */
