/*##############################################################################

nIP - nano IP stack

File        : mem.h

Description : Memory Management

Copyright notice:

Copyright (C) 2007 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#ifndef _NIP_MEM_H
 #define _NIP_MEM_H

#include "nip_init.h"
#include "nip_error.h"

/** Maximum number of memory blocks. Raising the number above 255 may lead to
  * an increase of used program memory, due to larger identifiers. Look at the
  * definition of nip_mem_handle_t for further explanation.
  */
#ifndef NIP_MEM_MAX_BLOCKS
	#define NIP_MEM_MAX_BLOCKS 255
#endif

/** Number of bytes reserved for dynamic buffer management. The default value
 * is 600 Bytes which is assumed to be needed for the reception of datagrams
 * with the standard maximum IP size of 576 Bytes.
 * \note Not all of the reservied memory will actually be available for
 * allocation, as some of it will be spend on management information.
 */
#ifndef NIP_MEM_DYNAMIC_SIZE
	#define NIP_MEM_DYNAMIC_SIZE 600
#endif


#ifndef NIP_MEM_DEBUG
	#define NIP_MEM_DEBUG   0
#endif


// Define type for memory size
#if ( NIP_MEM_DYNAMIC_SIZE <= 256 )
	typedef uint8_t nip_mem_size_t;
	typedef uint8_t nip_mem_cnt_t;
#elif ( NIP_MEM_DYNAMIC_SIZE <= 65563 )
	typedef uint16_t nip_mem_size_t;
	typedef uint16_t nip_mem_cnt_t;
#else
	typedef uint32_t nip_mem_size_t;
	typedef uint32_t nip_mem_cnt_t;
#endif


/** States of memory allocation function */
typedef enum
{
	NIP_MEM_ALLOC_DEFRAG,
	NIP_MEM_ALLOC_SIZE,
	NIP_MEM_ALLOC_MIN_SIZE,
	NIP_MEM_ALLOC_BLOCK_SIZE,
	NIP_MEM_ALLOC_FAIL,
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mem_alloc_stat_t;



/** Handle for block of dynamically managed memory. */
#if ( NIP_MEM_MAX_BLOCKS < 256 )
	typedef uint8_t nip_mem_handle_t;
#elif ( NIP_MEM_MAX_BLOCKS < 65536 )
	typedef uint16_t nip_mem_handle_t;
#else
	typedef uint32_t nip_mem_handle_t;
#endif

/** Type definition for memory block flags */
typedef uint8_t nip_mem_flags_t;

#define NIP_MEM_NULL        0x00  /**< handle for undefined memory block */

/** \name Flags for block of dynamically managed memory. */
//@{
#define NIP_MEM_FLG_USED    0x01  /**< Block in use */
#define NIP_MEM_FLG_LOCKED  0x02  /**< Block locked and may not be moved/defragmented at this time. */
#define NIP_MEM_FLG_DELREAD 0x04  /**< Data will be deleted automatically when read from block. */

#define NIP_MEM_FLGS_UNUSED 0x00  /**< Flag configuration for unused block */
//@}

/** Header for block of dynamically managed memory */
typedef struct
{
	nip_mem_handle_t id;         //!< Unique identifier. 1 <= id <= NIP_MEM_MAX_BLOCKS
	nip_mem_size_t   res_length; //!< reserved length. Size of reserved memory for this block. (incl. this header)
	nip_mem_size_t   min_length; //!< minimum length. Amount of memory that is guaranteed to be reserved for this block's data.
	nip_mem_size_t   used_length;//!< used length. Amount of memory currently used for data.
	nip_mem_flags_t  flags;      //!< flags to control access to the block \todo use bit masks (see below) to save program memory
/*	struct
	{
		uint8_t used:1;
		uint8_t locked:1;
		uint8_t delread:1;
		uint8_t unused_flag:1;
		uint8_t unused_flag:1;
		uint8_t unused_flag:1;
		uint8_t unused_flag:1;
		uint8_t unused_flag:1;
	}flags;      //!< flags to control access to the block
*/


}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mem_block_t;

/** Structure to handle pointers to managed or unmanaged memory. If the id
 * attribute of the structure is NIP_MEM_NULL, then ptr holds an actual pointer,
 * for direct usage. If id is not NIP_MEM_NULL, ptr defines and offset from the
 * beginning of the data part of the managed memory block specified by id.
 */
typedef struct
{
	nip_mem_handle_t id;   /**< ID of memory block or NIP_MEM_NULL */
	uint8_t          *ptr; /**< offset in block memory, or pointer to data if id
	                        * is NIP_MEM_NULL */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_mem_ptr_t;

#define NIP_MEM_BLOCK_USED( b )   (b->flags & NIP_MEM_FLG_USED)
#define NIP_MEM_BLOCK_LOCKED( b ) (b->flags & NIP_MEM_FLG_LOCKED)

void nip_mem_init();
nip_mem_handle_t nip_mem_alloc( nip_mem_size_t size, nip_mem_size_t min_length, nip_mem_flags_t flags, nip_mem_size_t *res_length);
void nip_mem_free( nip_mem_handle_t block_id );
nip_error_t nip_mem_set_min_length( nip_mem_handle_t block_id, nip_mem_size_t min_length );
nip_error_t nip_mem_write( nip_mem_handle_t block_id, void *buf, nip_mem_size_t size );
nip_mem_size_t nip_mem_read_at_pos( nip_mem_handle_t block_id, uint8_t *buf,
                                    nip_mem_size_t size, nip_mem_size_t pos);
nip_mem_size_t nip_mem_read( nip_mem_handle_t block_id, uint8_t *buf, nip_mem_size_t size );
void *nip_mem_obtain_ptr( nip_mem_handle_t block_id );
void nip_mem_release_block( nip_mem_handle_t block_id );
void *nip_mem_ptr( nip_mem_ptr_t *p );
void nip_mem_release_ptr( nip_mem_ptr_t *p );
nip_mem_size_t nip_mem_buf_used( nip_mem_handle_t block_id );
void nip_mem_set_used( nip_mem_handle_t block_id, nip_mem_size_t new_size );
nip_mem_size_t nip_mem_move( nip_mem_handle_t dst_block, nip_mem_handle_t src_block, nip_mem_size_t size );
nip_error_t nip_mem_insert( nip_mem_ptr_t *dst, nip_mem_ptr_t *src, nip_mem_size_t len );
nip_error_t nip_mem_cut( nip_mem_ptr_t *location, nip_mem_size_t len );
#endif /* _NIP_MEM_H */
