/*##############################################################################

nIP - nano IP stack

File        : net_if.h

Description : Network interface (header).

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#ifndef _NIP_NET_IF_H
 #define _NIP_NET_IF_H

#include "nip_init.h"
#include "mem.h"
#include "os_core.h"
//#include "net/ip.h"

#ifndef NIP_NET_IF_COUNT
	/** Define number of Network Interfaces. Defaults to 1. */
	#define NIP_NET_IF_COUNT          1
#endif

#ifndef NIP_NET_IF_TRANS_COUNT
	/** Define maximum number of queued transmissions per interface
	 * Defaults to 2.*/
	#define NIP_NET_IF_TRANS_COUNT 2
#endif

#define NIP_NET_MAX_TRANSMISSIONS NIP_NET_IF_TRANS_COUNT

#ifndef NIP_MAX_HW_ADDR_SIZE
	/** Define maximum hardware address size. Defaults to 6 for ethernet addressing.*/
	#define NIP_MAX_HW_ADDR_SIZE      6
#endif

#ifndef NIP_MAX_NET_ADDR_SIZE
	/** Define maximum network address size. Defaults to 4 for ip addressing.*/
	#define NIP_MAX_NET_ADDR_SIZE     4
#endif

#ifndef NIP_NET_LOOPBACK_ENABLE
	/** Enable local loopback of network packages */
	#define NIP_NET_LOOPBACK_ENABLE  1
#endif

/** to be used as NULL indicator on transmission ids */
#define NIP_NET_NULL_TRANSMISSION    NIP_NET_MAX_TRANSMISSIONS

// Define type of transmission ids
#if   ( NIP_NET_MAX_TRANSMISSIONS <  255   )
	typedef uint8_t nip_net_trans_id_t;
#elif ( NIP_NET_MAX_TRANSMISSIONS <  65535 )
	typedef uint16_t nip_net_trans_id_t;
#else
	typedef uint32_t nip_net_trans_id_t;
#endif



// Define type of Network Interface ID
#if ( NIP_NET_IF_COUNT < 0xFF )
	typedef uint8_t nip_net_if_id_t;
	#define NIP_NET_NO_IF   0xFF
#elif ( NIP_NET_IF_COUNT < 0xFFFF )
	typedef uint16_t nip_net_if_id_t;
	#define NIP_NET_NO_IF   0xFFFF
#else
	typedef uint32_t nip_net_if_id_t;
	#define NIP_NET_NO_IF   0xFFFFFFFF
#endif


/** @name Physical Interface Flags
 *
 * Interface flags (IFF) are used to influence the behavior of network interfaces.
 * @warning The flags should be set using the nip_set_if_flags function and may
 * not be altered directly, as that might not be noticed by the interface.
 */
//@{
#define NIP_PHY_IFF_UP        0x01  /**< Interface is active and ready to transfer packets. */
#define NIP_PHY_IFF_BROADCAST 0x02  /**< Interface allows broadasting. */
#define NIP_PHY_IFF_MULTICAST 0x04  /**< Interface allows multicasting. */
#define NIP_PHY_IFF_NOARP     0x08  /**< Interface does not use ARP. Hard_header method takes care of addressing. */
#define NIP_PHY_IFF_LOOPBACK  0x10  /**< Interface does local loopback. \todo implement loopback! */
#define NIP_PHY_IFF_CNT_OVW   0x20  /**< Interface should count receive buffer overruns. */
#define NIP_PHY_IFF_CNT_RXE   0x40  /**< Interface should count receive errors. */
#define NIP_PHY_IFF_CNT_TXE   0x80  /**< Interface should count transmit errors. */

//@}

/** @name Macros for Interface statistics */
//@{

/** Macro to count receive buffer overruns */
#ifdef NIP_NET_IF_CNT_OVERRUN
	#define NIP_PHY_LOG_OVW(net_if_ptr) net_if_ptr->phy_conf.stats.cnt_overrun++
#else
	#define NIP_PHY_LOG_OVW(net_if_ptr)
#endif

/** Macro to count receive errors */
#ifdef NIP_NET_IF_CNT_RXE
	#define NIP_PHY_LOG_RXE(net_if_ptr) net_if_ptr->phy_conf.stats.cnt_rx_err++
#else
	#define NIP_PHY_LOG_RXE(net_if_ptr)
#endif

/** Macro to count transmit errors */
#ifdef NIP_NET_IF_CNT_TXE
	#define NIP_PHY_LOG_TXE(net_if_ptr) net_if_ptr->phy_conf.stats.cnt_tx_err++
#else
	#define NIP_PHY_LOG_TXE(net_if_ptr)
#endif


//@}

/** Type to be used when referencing IFFs */
typedef uint8_t nip_phy_if_flags_t;

/** \file net_if.h \todo remove Physical Interface function typedefs, if code works without these typedefs */
/*typedef uint8_t nip_net_if_send( nip_net_if_id_t net_if_id );
typedef uint8_t nip_net_if_receive( nip_net_if_id_t net_if_id );
*/

/** Network Interface status */
typedef enum
{
	NIP_NET_IF_IDLE,        /**< idle -- ready to send or receive data */
	NIP_NET_IF_RECEIVING,   /**< busy -- receiving data */
	NIP_NET_IF_ROUTE,       /**< transmission waiting to be routed */
	NIP_NET_IF_ROUTING,     /**< transmission busy routing */
	NIP_NET_IF_ROUTED,      /**< transmission has been routed */
	NIP_NET_IF_RESOLV_ADDR, /**< waiting for hardware address to be resolved */
	NIP_NET_IF_RTS,         /**< ready to send, hardware address resolved */
	NIP_NET_IF_SENDING,     /**< busy -- sending data */
	NIP_NET_IF_SEND_HNF,    /**< sending data failed, Host not found, waiting to be unlocked.*/
	NIP_NET_IF_SEND_NRT,    /**< sending data failed, No route to host, waiting to be unlocked.*/
	NIP_NET_IF_DONE         /**< sending/receiving data successful, waiting to be unlocked.*/
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_net_trans_status_t;


/** Physical layer interface types.
 * @note Type numbers are taken from ARP section of "Assigned Numbers" RFC and
 * may be used for various applications (ARP, DHCP,...).
 */
typedef enum
{
	NIP_PHY_IF_OTHER,       /**< Other Network Interface. Use this type if non of the below apply. */
	NIP_PHY_IF_ETH = 1      /**< Ethernet interface */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_phy_if_type_t;

typedef union
{
	/** parameters for ip*/
	struct
	{
		uint8_t dst_nw_addr[NIP_MAX_NET_ADDR_SIZE];  /**< network (IP) address of transmission destination. */
		uint8_t src_nw_addr[NIP_MAX_NET_ADDR_SIZE];  /**< network (IP) address of transmission source. */
		uint8_t ll_nw_addr[NIP_MAX_NET_ADDR_SIZE];   /**< network (IP) address of Local Link target. */
		uint8_t ttl;
		uint16_t id;
		uint8_t protocol;
	}
	#ifndef DOXYGEN
	__attribute__((packed))
	#endif
	ip;

	/** parameters for tcp*/
	struct
	{
		uint8_t  dst_nw_addr[NIP_MAX_NET_ADDR_SIZE];  /**< network (IP) address of transmission destination. */
		uint8_t  src_nw_addr[NIP_MAX_NET_ADDR_SIZE];  /**< network (IP) address of transmission source. */
		uint16_t dst_port;
		uint16_t src_port;
	}
	#ifndef DOXYGEN
	__attribute__((packed))
	#endif
	tcp;



}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_net_if_trans_params_t;

/** flags to define transmission behavior. */
typedef struct
{
	uint8_t free_payload:1; /**< free payload upon transmission reset */
	uint8_t reset_status:1; /**< reset status upon transmission reset */
	uint8_t ll_broadcast:1; /**< local link broadcast */
	uint8_t deliverlocal:1; /**< transmit packet locally */
	uint8_t forward:1;      /**< forward packet on transmission's interface */
	uint8_t unused:3;       /**< unused flag space */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_net_if_trans_flags_t;

/** Network Interface transmission */
typedef struct
{
	nip_net_if_id_t     if_id;       /**< id of network interface to transmit on */
	nip_net_trans_status_t status;      /**< Transmission status */
	nip_time_t          time;        /**< This is needed for ARP timeouts, etc. */
	uint8_t             retrans_cnt; /**< Retransmission Counter */
	uint16_t            header_size; /**< header size */
	uint16_t            payload_size;/**< payload size */
	uint16_t            payload_off; /**< offset of payload in payload buffer */
	nip_mem_handle_t    header; /**< Handle to buffer containing a transmission header. */
	nip_mem_handle_t    payload;/**< Handle to buffer containing the payload to be transmitted. */
	nip_net_if_trans_params_t params;/**< transmission params */
	nip_net_if_trans_flags_t  flags; /**< transmission flags */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_net_if_trans_t;


typedef uint16_t nip_net_stats_cnt_t;

/** Network Interface statistics */
typedef struct
{
	#ifdef NIP_NET_IF_CNT_OVERRUN
		nip_net_stats_cnt_t cnt_overrun;   /**< Number of receive buffer overruns. */
	#endif

	#ifdef NIP_NET_IF_CNT_TXE
		nip_net_stats_cnt_t cnt_tx_err;    /**< Number of transmit errors. */
	#endif

	#ifdef NIP_NET_IF_CNT_RXE
		nip_net_stats_cnt_t cnt_rx_err;    /**< Number of transmit errors. */
	#endif

	#ifdef NIP_NET_IF_CNT_DISCARDED
		nip_net_stats_cnt_t cnt_discarded; /**< Number of discarded packets. */
	#endif

	#ifdef NIP_NET_IF_CNT_RECEIVED
		nip_net_stats_cnt_t cnt_received;  /**< Number of received packets. */
	#endif

}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_net_if_stats_t;


/** Physical layer driver interface. Some of the properties are in the style of
  * Linux device driver attributes and methods.
  * \todo add interface functions
  */
typedef struct
{
	nip_phy_if_type_t  type;  /**< Interface type which will be used by ARP for address resolution. */
	nip_net_if_stats_t stats; /**< Physical Interface statistics (packet loss, etc) */
	nip_phy_if_flags_t flags; /**< Physical Interface Flags	 */
	uint8_t            hw_addr[NIP_MAX_HW_ADDR_SIZE]; /**< Hardware address */
	uint8_t            hw_brdcst_addr[NIP_MAX_HW_ADDR_SIZE]; /**< Hardware broadcast address */
	uint8_t            hw_addr_size; /** Size of hardware address */
	uint16_t           rx_bytes; /** bytes pending to be read from phy memory. After read-initialization this will be the packet length.*/
	uint16_t           rx_pos;   /** position in receive packet buffer */
	uint16_t           tx_bytes; /** bytes still to be sent in current transmission */
	uint16_t           tx_pos;   /** position in transmit packet buffer */

	nip_success_t (*open)( nip_net_if_id_t net_if_id ); /**< Opens network interface.
	     * The open method should register any system resource it needs (I/O
		 * ports, IRQ, etc)
		 * @return
		 *   -  NIP_SUCCESS if network interface has been opened, or
		 *   -  NIP_ERROR if opening network interface failed.
		 */

	nip_success_t (*close)( nip_net_if_id_t net_if_id );/**< Stops network interface.
		 * Reverses anything that's been done at open.
		 * @return
		 *   -  NIP_SUCCESS if network interface has been closed, or
		 *   -  NIP_ERROR if closing network interface failed.
		 */

	nip_error_t (*hard_send_init) (nip_net_if_id_t net_if_id, void *ll_target_addr,
		uint16_t type, void *daddr, void *saddr, uint16_t len);
		/**< Builds hardware header and initializes packet transmission. If
		 * daddr is NULL, the destination address should be resolved using the
		 * given ll_target_addr. For ethernet devices ARP is being used.
		 * Ethernet device drivers can rely on nip_arp_lookup( uint8_t* addr)
		 * to be available. If the Address cannot be resolved locally the hardware
		 * specific address resolution mechanism should be initiated and the
		 * return value set to NIP_RESOLVING_ADDR. It's up to the caller to handle
		 * the return value correctly and retry transmission after a reasonable
		 * timeout.
		 * @return
		 *   - NIP_E_OK             transmission initialized successfully
		 *   - NIP_E_RESOLVING_ADDR address resolution started, or
		 *   - NIP_E_PARAM_FAILED   source or destination address was invalid
		 *   - NIP_E_HARD_ERROR     error connecting to network hardware
		 */

	uint16_t (*hard_send) (nip_net_if_id_t net_if_id, uint8_t* buffer, uint16_t count);
		/**< Transmit count bytes from given buffer.
		 * @return
		 *   - number of bytes transmitted, or
		 *   - 0 if error occured
		 */

	uint16_t (*hard_read) (nip_net_if_id_t net_if_id, uint8_t* buffer, uint16_t count);
		/**< Read data from interface and copy it into recv_buf. If size is 0 the current
		 * reception will be terminated, all pending bytes discarded and network interface
		 * reset to read next packet.
		 * @return number of bytes copied.
		 */

	nip_success_t (*hard_read_init) ( nip_net_if_id_t net_if_id );
		/**< Initialize Reading packet from interface.
		 * @return packet length or 0 if packet buffer is empty.
		 */

	void (*hard_disp_recv)( void );
		/**< Function to be called by dispatcher to initiate reception of data */

}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_phy_if_t;

typedef enum
{
	NIP_IP_CONF_STAT_UNCONFIGURED = 0,
	NIP_IP_CONF_STAT_CONFIGURED
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_if_ip_conf_stat_t;

/** Network Interface IP configuration */
typedef struct nip_if_ip_conf
{
	nip_if_ip_conf_stat_t state;
	uint8_t addr[4];
	uint8_t brdcst_addr[4];
	uint8_t subnet_mask[4];
	uint8_t std_gateway[4];
	#if NIP_DNS_ENABLE == 1
	uint8_t name_server[4];
	#endif
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_if_ip_conf_t;

/** Network Interface */
typedef struct nip_net_if
{
	nip_net_if_id_t    id;       /**< Network Interface ID */
	nip_if_ip_conf_t   ip_conf;  /**< Network Interface IP configuration */
	nip_phy_if_t       phy_conf; /**< Physical layer interface */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_net_if_t;

#define NIP_IP_ADDR( if_id ) nip_net_if_list[if_id].ip_conf.addr

extern nip_net_if_t nip_net_if_list[];
extern nip_net_if_trans_t nip_net_trans_list[];


nip_net_if_id_t nip_net_if_create( nip_phy_if_type_t type );
void            nip_net_if_free  ( nip_net_if_id_t id );
nip_success_t   nip_set_if_flags ( nip_net_if_id_t id, nip_phy_if_flags_t flags);
nip_net_if_t   *nip_net_if_ptr   ( nip_net_if_id_t );
nip_success_t   nip_net_if_up    ( nip_net_if_id_t );
nip_success_t   nip_net_if_down  ( nip_net_if_id_t );
void nip_reset_trans( nip_net_if_trans_t *trans );

#endif /* _NIP_NET_IF_H */

