/*##############################################################################

nIP - nano IP stack

File        : nip_init.h

Description : Initialize NIP Internetstack (HEADER)

Copyright notice:

Copyright (C) 2007 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#ifndef _NIP_INIT_H
 #define _NIP_INIT_H

#include "nip_types.h"

#define QMAKESTR(s) #s
#define MAKESTR(s)  QMAKESTR(s)

/** @name Architecture */
//@{
/** define what architecture the NIP network stack will be compiled for. The
 * architecture has to match a directory name in the `'arch'' directory.
 */
#define NIP_ARCH       avr

/** path for architecture-dependent files. */
#define NIP_ARCH_PATH  arch/NIP_ARCH

/** Build filename for includes depending on the architecture. */
#define NIP_ARCH_FILE( file ) MAKESTR(NIP_ARCH_PATH/file)

//@}

/** @name nIP Modules */
//@{

#ifndef NIP_MEM_ENABLE
	#define NIP_MEM_ENABLE    1 /**< enable memory management */
#endif
#ifndef NIP_NET_ENABLE
	#define NIP_NET_ENABLE    1 /**< enable network interfaces */
#endif
#ifndef NIP_DISP_ENABLE
	#define NIP_DISP_ENABLE   1 /**< enable dispatcher */
#endif
#ifndef NIP_ETH_ENABLE
	#define NIP_ETH_ENABLE    1 /**< enable support of ethernet interfaces */
#endif
#ifndef NIP_RTL_ENABLE
	#define NIP_RTL_ENABLE    1 /**< enable realtek network driver */
#endif
#ifndef NIP_IP_ENABLE
	#define NIP_IP_ENABLE     1 /**< enable IP */
#endif
#ifndef NIP_ARP_ENABLE
	#define NIP_ARP_ENABLE    1 /**< enable ARP */
#endif
#ifndef NIP_ICMP_ENABLE
	#define NIP_ICMP_ENABLE   1 /**< enable ICMP */
#endif
#ifndef NIP_UDP_ENABLE
	#define NIP_UDP_ENABLE    1 /**< enable UDP */
#endif
#ifndef NIP_TCP_ENABLE
	#define NIP_TCP_ENABLE    1 /**< enable TCP */
#endif
#ifndef NIP_DHCP_ENABLE
	#define NIP_DHCP_ENABLE   1 /** Dynamic Host Configuration Protocol. Requires NIP_UDP_ENABLE. */
#endif
#ifndef NIP_AUTOIP_ENABLE
	#define NIP_AUTOIP_ENABLE 1 /** Dynamic Configuration of IPv4 Link-Local Addresses */
#endif
#ifndef NIP_MULTICAST_ENABLE
	#define NIP_MULTICAST_ENABLE 1 /** Send and receive IP multicast messages */
#endif
#ifndef NIP_IGMP_ENABLE
	#define NIP_IGMP_ENABLE   0 /** needed to receive multihop multicast messages. Requires NIP_MULTICAST_ENABLE */
#endif
#ifndef NIP_MDNS_ENABLE
	#define NIP_MDNS_ENABLE   1 /** mDNS */
#endif
#ifndef NIP_DNSSD_ENABLE
	#define NIP_DNSSD_ENABLE   1 /** DNS-SD */
#endif

//@}

/** @name Network Interface Configuration */
//@{

/** Define Number of network interfaces */
#define NIP_NET_IF_COUNT 1

/** Define maximum number of queued transmissions per interface. Defaults to 2.*/
#define NIP_NET_RX_TRANS_COUNT 2

/** Define maximum number transmissions in global send queue. Defaults to 2.*/
#define NIP_NET_TX_TRANS_COUNT 2

/** Define maximum hardware address size. Defaults to 6 for ethernet addresses.*/
#define NIP_MAX_HW_ADDR_SIZE 6

/** Define maximum network address size. Defaults to 4 for ip addressing.*/
#define NIP_MAX_NET_ADDR_SIZE 4

/** Enable local loopback of network packages */
#define NIP_NET_LOOPBACK_ENABLE 1

/** number of entries in ARP cache */
#define NIP_ARP_CACHE_SIZE 4

//@}

/** @name Network Interface Statistics */
//@{

/** count network interface receive buffer overruns */
#define NIP_NET_IF_CNT_OVERRUN

/** count receive errors */
#define NIP_NET_IF_CNT_RXE

/** count transmit errors */
#define NIP_NET_IF_CNT_TXE

/** count discarded packets */
#define NIP_NET_IF_CNT_DISCARDED

/** count received packets  */
#define NIP_NET_IF_CNT_RECEIVED

//@}


/** @name Transport Layer Configuration */
//@{

//** Define maximum number of transport layer sockets. */
//#define NIP_MAX_SOCKETS 10

/** Use fixed socket array (static allocation of memory for fixed numer of
	* sockets) or use dynamic array of sockets with at most NIP_UDP_MAX_SOCKETS
	* sockets. */
#define NIP_UDP_USE_FIXED_SOCKETS  1

/** Fixed number of UDP sockets to be used for fixed socket configuration */
#define NIP_UDP_FIXED_SOCKETS      4

/** Fixed size of UDP queue for incoming packets */
#define NIP_UDP_PACKET_QUEUE_SIZE  2


//@}



/** @name Memory management */
//@{

/** Define maximum number of memory blocks.
  * \note Any number larger than 255 will cause memory addresses to be larger
  * than 8 Bit, which  may be disadvantagous on 8-bit architectures. */
#define NIP_MEM_MAX_BLOCKS 255

/** Number of bytes reserved for dynamic buffer management.
 * \note Not all of the reservied memory will actually be available for
 * allocation, as some of it will be spent on management information.
 */
#define NIP_MEM_DYNAMIC_SIZE 1000


/** enable additional sanity check which may be useful points to halt the
 * debugger at. Check for DEBUG comments in nip_mem.c.
 */
#define NIP_MEM_DEBUG  1

//@}




void nip_init( void );

#endif /* _NIP_INIT */

