/*##############################################################################

nIP - nano IP stack

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/
/** @file mdns.c
 *
 * Standard compliant, minimal implementation of Multiast DNS.
 * @todo Make it standard compliant. ;) (timing, conflicts, local queries...)
 *
 */

#include "nip_init.h"
#include "app/mdns.h"
#include "dispatcher.h"
#include "net/udp.h"
#include "app/ports.h"
#include "inet.h"

#if ( NIP_MDNS_ENABLE == 1 )



struct nip_mdns_check    nip_mdns;

nip_net_if_id_t          nip_mdns_if_id = NIP_NET_NO_IF;
nip_udp_sock_id_t        nip_mdns_sock  = NIP_UDP_NO_SOCKET;
nip_mdns_cache_id_t      nip_mdns_hostname = NIP_MDNS_NO_CACHE_ID;
nip_mdns_cache_id_t      nip_mdns_host_ptr = NIP_MDNS_NO_CACHE_ID;
nip_mdns_namestring_id_t nip_mdns_domain   = NIP_MDNS_NO_NAMESTRING;
nip_mem_handle_t nip_mdns_cs      = NIP_MEM_NULL; /**< cache subscriptions */
nip_mem_ptr_t    nip_mdns_cache   = {NIP_MEM_NULL,NULL}; /**< MDNS cache */
nip_mem_handle_t nip_mdns_names   = NIP_MEM_NULL; /**< MDNS name strings */
nip_mem_ptr_t    nip_mdns_queries = {NIP_MEM_NULL,NULL}; /**< MDNS queries */

void nip_mdns_register_hostname( struct nip_dns_string *name );
void nip_mdns_disp_check_queries( void );
nip_error_t nip_mdns_parse_packet( nip_udp_buf_type_t udp_buffer,
nip_mdns_packet_task_t task, nip_mem_ptr_t *p, nip_mdns_qa_offset_t section );

/** initialize mdns memory
 *
 * @return NIP_E_OK or NIP_E_OUT_OF_MEMORY
 */
nip_error_t nip_mdns_init( void )
{
	// use union to save memory
	union
	{
		struct nip_udp_sock_addr addr;
		// The decision on how to initialize strings is tough, as it either eats up
		// RAM or program memory, both of which we don't have much. Currently we
		// follow the pro-RAM con-Program-Memory approach, that is to initialize all
		// chars separately, to allocate the string locally.
		/// @todo check on moving fixed strings to the EEPROM and reading from there.
		uint8_t  local[6];
		uint8_t  avr[4];
		uint8_t  mc_addr[4];
	} tmp;
	nip_mem_ptr_t name_ptr;

	// init buffer space
	if ( nip_mdns_cs    == NIP_MEM_NULL )
		nip_mdns_cs    = nip_mem_alloc( 0, 0, 0, NULL );
	if ( nip_mdns_cache.id == NIP_MEM_NULL )
		nip_mdns_cache.id = nip_mem_alloc( 0, 0, 0, NULL );
	if ( nip_mdns_names == NIP_MEM_NULL )
		nip_mdns_names = nip_mem_alloc( 0, 0, 0, NULL );
	if ( nip_mdns_queries.id == NIP_MEM_NULL )
		nip_mdns_queries.id = nip_mem_alloc( 0, 0, 0, NULL );

	// register names
	if ( nip_mdns_domain == NIP_MDNS_NO_NAMESTRING )
	{
		tmp.local[0] = 5;
		tmp.local[1] = 'l';
		tmp.local[2] = 'o';
		tmp.local[3] = 'c';
		tmp.local[4] = 'a';
		tmp.local[5] = 'l';
		name_ptr.id  = NIP_MEM_NULL;
		name_ptr.ptr = tmp.local;
		nip_mdns_domain = NIP_MDNS_REG_STRING( &name_ptr );
	}

	if ( nip_mdns_domain   == NIP_MDNS_NO_NAMESTRING
	  || nip_mdns_cache.id == NIP_MEM_NULL
	  || nip_mdns_cs       == NIP_MEM_NULL )
	{
		return NIP_E_OUT_OF_MEMORY;
	}

	// Don't continue if MDNS is down
	if ( nip_mdns.stat == NIP_MDNS_STAT_DOWN )
		return NIP_E_NOT_CONFIGURED;

	// register hostname -- if necessary
	if ( nip_mdns_hostname == NIP_MDNS_NO_CACHE_ID )
	{
		tmp.avr[0] = 3;
// 		tmp.avr[1] = 0xC3; // Unicode for Ä
// 		tmp.avr[2] = 0x84; // Unicode for Ä
		tmp.avr[1] = 'A';
		tmp.avr[2] = 'V';
		tmp.avr[3] = 'R';

		/// @todo get hostname from EEPROM?
		nip_mdns_register_hostname( (struct nip_dns_string *)tmp.avr );
		if ( nip_error != NIP_E_OK )
			return nip_error;
	}

	// init UDP socket
	if ( nip_mdns_sock == NIP_UDP_NO_SOCKET && nip_mdns_if_id != NIP_NET_NO_IF )
	{
		// only listen on configured interfacew
		tmp.addr.port = NIP_UDP_PORT_MDNS;
		nip_memcpy( tmp.addr.ip, NIP_IP_ADDR( nip_mdns_if_id ), 4 );
/*		tmp.addr.ip[0] = 0;
		tmp.addr.ip[1] = 0;
		tmp.addr.ip[2] = 0;
		tmp.addr.ip[3] = 0;*/
		nip_mdns_sock = nip_udp_socket(
			&tmp.addr,
			NIP_UDP_SOCK_FLG_LISTENING | NIP_UDP_SOCK_FLG_NON_BLOCK
		);
		if ( nip_mdns_sock == NIP_UDP_NO_SOCKET )
			return nip_error;
	}

	// Register Multicast Group
	tmp.mc_addr[0]=NIP_MDNS_ADDR0;
	tmp.mc_addr[1]=NIP_MDNS_ADDR1;
	tmp.mc_addr[2]=NIP_MDNS_ADDR2;
	tmp.mc_addr[3]=NIP_MDNS_ADDR3;
	return nip_ip_join_host_group( nip_mdns_if_id, tmp.mc_addr );
}


/** Perform dns name parsing-operation, depending on task parameter.
 * @param task
 *  - NIP_MDNS_NTASK_FIND   check the local name list and fill the given
 *    name pointer (if not NULL) if the search was sucessfull.
 *  - NIP_MDNS_NTASK_WRITE  write
 * @param name  pointer to dns name structure
 * @param buf   pointer to name buffer
 * @param size  size of buffer. May be 0 for ..WRITE-task, if memory block is
 *              being used.
 * operation will parse the name from
 */
// nip_error_t nip_mdns_name_task( nip_mdns_name_task_t task, struct nip_mdns_name *name,
// nip_mem_ptr_t *buf, nip_mem_size_t size )
// {
// 	nip_mem_size_t  listsize;
// 	nip_error_t     res = NIP_E_OK;
//
// 	listsize = nip_mem_buf_used( buf->id );
//
// 	/// @todo call nip_mdns_register_name depending on task
//
// 	return res;
// }

/** Compare UTF-8 namestrings, case insensitively, according to mDNS draft
 * chapter "18. Multicast DNS Character Set".
 *
 * @note n1 or n2 MUST NOT be NULL.
 *
 * @return NIP_E_OK or NIP_E_NO_MATCH
 */
nip_error_t nip_mdns_name_cmp( struct nip_dns_string *n1, struct nip_dns_string *n2)
{
	uint8_t pos;
	uint8_t utf8_head;
	uint8_t c1, c2;

	if ( n1->len == n2->len )
	{
		pos = 0;
		utf8_head = 0;
		do
		{
			c1 =n1->s[pos]; c2 = n2->s[pos];

			// check for UTF-8 multibyte-characters
			if ( (utf8_head & 0x80) == 0 )
				utf8_head   = c1;
			else
				utf8_head <<= 1;

			// compare characters a-zA-Z case insensitively (convert to lower case)
			// only exception: UTF-8 parts.
			if ( (utf8_head & 0x80) == 0 )
			{
				if ( c1 >= 'A' && c1 <='Z' ) c1 += 'a'-'A';
				if ( c2 >= 'A' && c2 <='Z' ) c2 += 'a'-'A';
			}

			if ( c1 != c2 )
				goto abort_no_match;

		} while ( ++pos < n1->len );
		return NIP_E_OK;
	}

	abort_no_match:
	return NIP_E_NO_MATCH;
}

/** Registers new MDNS namestring or retrieves already registered entry for that
 * name. The namestring's usage counter will be increased for every call of
 * nip_mdns_register_name(). This function may also be used to unregister
 * a previously registered namestring by providing a negative usage count.
 * @param id    if set, the usage counter for that namestring will be increased
 *              and name and len parameter will be ignored.
 * @param cnt   if > 1, id will be treated as array of namestrings to be
 *              registered. Must be 0 if id is not set.
 * @param name_ptr  pointer to buffer holding the name. If id is set, the found
 *              name(s) will be appended to that buffer, preceded by their
 *              length. The pointer has to point to a nip_dns_string structure.
 * @param len   length of name
 * @param usg   number of string usages to register (negative to unregister)
 * @return ID of namestring or NIP_MDNS_NO_NAMESTRING on error.
 *
 * @note If usg is 0 a simple lookup of the given namestring will be performed
 * and -- if found -- it's Id be returned.
 *
 * @note The function's result depends very much on the correct combination of
 * the various parameters. Those are chosen to be as minimal as possible, which
 * will most probably lead to erroneous usage of the function. You SHOULD NOT
 * call this function directly but use the provided macros, which, depending on
 * the provided function, require only a subset of the parameters to be set.
 *  - NIP_MDNS_REG_STRING
 *  - NIP_MDNS_REG_NAME
 *  - NIP_MDNS_NAME_USE
 *  - NIP_MDNS_UNREG_NAME
 *  - NIP_MDNS_CHECK_NAME
 *  - NIP_MDNS_WRITE_NAME
 */
nip_mdns_namestring_id_t nip_mdns_register_name( nip_mdns_namestring_id_t *id,
uint8_t cnt, nip_mem_ptr_t *name_ptr, nip_mdns_namestring_usg_t usg )
{
	nip_mdns_namestring_id_t   res = NIP_MDNS_NO_NAMESTRING;
	nip_mdns_namestring_id_t   new_id = 1;
	nip_mem_ptr_t              current;
	nip_mem_ptr_t              name_src;
	nip_error_t                err;
	nip_mem_size_t             len;
	nip_mem_size_t             listsize;
	struct nip_dns_string      *name = NULL;
	uint8_t                    i = 0;
	struct nip_mdns_namestring *l; // pointer to list item and list end

	listsize = nip_mem_buf_used( nip_mdns_names );

	do
	{
		current.id  = nip_mdns_names;
		current.ptr = (void*)0;

		// For all remaining operations a name string is required
		if ( name == NULL && (name = nip_mem_ptr( name_ptr )) == NULL && id == NULL )
		{
			new_id = NIP_MDNS_NO_NAMESTRING;
			break;
		}

		// search for name in list
		while ( (l = nip_mem_ptr( &current ) ) != NULL )
		{
			// Search Namestring by ID
			if ( id != NULL && ( name_ptr == NULL || usg == 0 ) )
			{
				new_id = NIP_MDNS_NO_NAMESTRING;
				if ( id[i] == NIP_MDNS_NO_NAMESTRING )
					i = cnt;
				else if ( id[i] == l->id )
				{
					l->usg  += usg;
					name_src.id  = current.id;
					name_src.ptr = current.ptr + NIP_MDNS_STRINGOFF;
					len      = l->str.len;
					// release memory pointers so nip_mem_insert() can work
					// unbothered by locked memory blocks.
					l = NULL;
					nip_mem_release_block( current.id );
					nip_mem_release_ptr( name_ptr );

					// insert length/name to given buffer
					nip_mem_insert( name_ptr, &name_src, len+1 );

				}
				else
				{
					// check next name
					goto next;
				}
				// Reached end of name? -> Insert 0-Character
				// The user has put cnt==NIP_MDNS_MAX_NAME_DEPTH, which means he
				// wants to print an entire name, which will always be 0-padded.
				if ( i >= NIP_MDNS_MAX_NAME_DEPTH-1 )
				{
					name_src.id  = NIP_MEM_NULL;
					name_src.ptr = NULL;
					nip_mem_insert( name_ptr, &name_src, 1 );
				}

				break;
			}

			// compare names
			if ( nip_mdns_name_cmp( &(l->str), name ) == NIP_E_OK )
			{
				// found name -> increase usage pointer and return ID
				new_id = NIP_MDNS_NO_NAMESTRING;
				res     = l->id;
				l->usg += usg;
				break;
			}

			// check free ID
			if ( new_id == l->id )
			{
				new_id++;
				current.ptr = (void*)0;
				goto release;
			}

			next:
				// advance pointer to next name
				current.ptr += sizeof( struct nip_mdns_namestring ) + l->str.len;

			release:
				nip_mem_release_block( current.id );
				l = NULL;
		}
		if ( l != NULL )
			nip_mem_release_block( current.id );

		// Add new String to list
		if ( new_id != NIP_MDNS_NO_NAMESTRING && new_id < NIP_MDNS_MAX_NAMESTRINGS && usg != 0 )
		{
			len = name->len;

			// release pointer temporarily, so it won't be blocking the following
			// memory operation, which may include defragmentation.
			nip_mem_release_ptr( name_ptr );

			// End of name reached?
			if ( len == 0 )
			{
				id[i] = 0;
				goto quit;
			}


			// allocate memory for block
			err = nip_mem_write( nip_mdns_names, NULL, sizeof( struct nip_mdns_namestring ) + len );

			if ( err != NIP_E_OK )
				goto add_error;

			// get new data pointers and make sure there's nothing wrong with them
			name = nip_mem_ptr( name_ptr );
			if ( name == NULL )
				goto add_error;

			l = nip_mem_ptr( &current);
			if ( l == NULL )
				goto add_error;

			// fill header of new item
			l->id  = new_id;
			l->usg = usg;
			l->str.len = name->len;

			// copy name to list
			nip_memcpy( l->str.s, name->s, name->len );

			res = new_id;
			if ( id != NULL )
			{
				id[i] = new_id;
			}
			nip_mem_release_ptr( &current );

			// forward name pointer
			name_ptr->ptr += len+1;
			name           = (struct nip_dns_string*)(((uint8_t*)name) + len+1);

			continue;

			add_error:
				// reset list size (delete all new items)
				nip_mem_set_used( nip_mdns_names, listsize );
				goto release_name;
		}
	}
	while ( ++i < cnt );


	release_name:

	// release name pointer
	if ( name != NULL )
		nip_mem_release_ptr( name_ptr );

	quit:
	return res;
}

/** Execute MDNS lookup query.
 * The query will be registered within the system and a query ID will be
 * returned. That query has to be ultimately closed by calling nip_mdns_close().
 * Answers to the query can be polled by calling nip_mdns_fetch_results().
 *
 * @param name pointer to DNS name
 * @param type DNS type
 *
 * @return query-ID or NIP_MDNS_NO_QUERY
 */
nip_mdns_query_id_t nip_mdns_query( nip_mdns_query_type_t type,
struct nip_mdns_name *rr_name, nip_mdns_type_t rr_type )
{
	nip_mdns_query_id_t   res = NIP_MDNS_NO_QUERY;
	nip_mdns_query_id_t   new_id = 0;
	nip_mem_size_t        listsize;
	struct nip_mdns_query *l, *end; // pointer to list item and list end
	struct nip_mdns_query query;

	listsize = nip_mem_buf_used( nip_mdns_queries.id );

	// search for name in list
	l = nip_mem_obtain_ptr( nip_mdns_queries.id );
	if ( l == NULL )
	{
		goto quit;
	}

	end = (void *)((uint8_t*)l + listsize );
	while ( l != end )
	{
		// compare names.
		if ( nip_memcmp( &l->rr_name, rr_name, sizeof( struct nip_mdns_name ) ) == 0 )
		{
			// Probes will only be registered once per name
			if ( l->type == NIP_MDNS_QTYPE_PROBE )
			{
				if ( type == NIP_MDNS_QTYPE_PROBE )
				{
					res  = (l->id & NIP_MDNS_QUERY_ID_MASK);
					break;
				}
			}

			// only register one remote instance per remote query
			else if ( type == NIP_MDNS_QTYPE_REMOTE
			  || ( type == NIP_MDNS_QTYPE_LOCAL && l->type == NIP_MDNS_QTYPE_REMLOC))
			{
				if ( l->type == NIP_MDNS_QTYPE_LOCAL )
					l->type = NIP_MDNS_QTYPE_REMLOC;
				res = (l->id & NIP_MDNS_QUERY_ID_MASK);
				break;
			}
		}

		// check free ID
		if ( (new_id & NIP_MDNS_QUERY_ID_MASK) == (l->id & NIP_MDNS_QUERY_ID_MASK) )
		{
			new_id++;
			l = ( struct nip_mdns_query *)((uint8_t*)end - listsize );
			continue;
		}

		// advance to next query
		l = (void*) ((uint8_t *)l + sizeof( struct nip_mdns_query ));
	}
	nip_mem_release_block( nip_mdns_queries.id );

	if ( (res & NIP_MDNS_QUERY_ID_MASK) == NIP_MDNS_NO_QUERY )
	{
		if ( (new_id & NIP_MDNS_QUERY_ID_MASK) < NIP_MDNS_MAX_QUERIES )
		{
			query.id       = new_id | NIP_MDNS_QUERY_ACTIVE;
			query.type     = type;
			query.rr_name  = *rr_name;
			query.rr_type  = rr_type;
			///@todo set timer correctly
			query.timer    = 0;
			if ( NIP_E_OK == nip_mem_write(
					nip_mdns_queries.id,
					(uint8_t*) &query,
					sizeof( struct nip_mdns_query ))
				)
			{
				res = new_id;
			}
			// register name usage
// 			nip_mdns_register_name( query.rr_name.part, NIP_MDNS_MAX_NAME_DEPTH, NULL, 0, 1 );
			NIP_MDNS_NAME_USE( query.rr_name.part, NIP_MDNS_MAX_NAME_DEPTH );
		}
	}


	quit:
	return res;
}

/** Return size of cache record. This function is used to ease usage of
 * dynamically sized cache list items.
 */
nip_mdns_record_size_t nip_mdns_record_size( struct nip_mdns_record *r )
{
	// cache item size depends on item type, because content size may vary greatly
	nip_mdns_record_size_t off;
	switch ( r->type )
	{
		case NIP_MDNS_TYPE_A:
			off = NIP_MDNS_A_RECORD_SIZE;
			break;
		case NIP_MDNS_TYPE_SRV:
			off = NIP_MDNS_SRV_RECORD_SIZE;
			break;
		case NIP_MDNS_TYPE_PTR:
			off = NIP_MDNS_PTR_RECORD_SIZE;
			break;
		// add cases for any other record type here. Their content will be treated
		// as string, without interpreting its meaning.
		default:
			off = NIP_MDNS_OTH_RECORD_SIZE;
	}

	return off;
}

/** Get pointer to specific cache entry.
 * @note You have to call nip_mem_release_block( nip_mdns_cache.id ) to release
 * the retrieved pointer and unlock the cache.
 *
 * @return Pointer to cache entry. On error NULL will be returned and nip_error
 * be set to one of the following values:
 *  - NIP_E_NOT_FOUND  no record found for given ID
 *  - NIP_E_LOCKED     cache locked, try again later
 */
struct nip_mdns_cache *nip_mdns_cache_ptr( nip_mdns_cache_id_t id )
{
	nip_mem_size_t         listsize;
	struct nip_mdns_cache  *l, *end; // pointer to list item and list end

	listsize = nip_mem_buf_used( nip_mdns_cache.id );
	l = nip_mem_obtain_ptr( nip_mdns_cache.id );

	if ( l == NULL )
	{
		nip_error = NIP_E_LOCKED;
		goto quit;
	}

	end = (void*)((uint8_t*)l + listsize );
	while ( l != end )
	{
		// found entry? -> return pointer
		if ( l->id == id )
		{
			return l;
		}

		// advance to next cache item
		l = (struct nip_mdns_cache *)((uint8_t*)l
		      + NIP_MDNS_BASE_CACHE_SIZE
		      + nip_mdns_record_size( &l->record ));
	}

	nip_mem_release_block( nip_mdns_cache.id );

	nip_error = NIP_E_NOT_FOUND;

	quit:
	return NULL;
}

/** Perform mDNS cache operation. Type of operation depends on task parameter.
 *
 *  - The register operation will add a new authorative record.
 *  - The learn operation will add or update a remote DNS record.
 *  - The get operation will retrieve the record data for a given ID.
 *  - The search operation will return the Id of the existing cache record that
 *    matches certain attributes from the given record. The set of attributes
 *    chosen for the match depends on the record type.
 *  - The reset operation removes all learned entries from cache, clears the
 *    subscription array and resets all registered entries into PROBING state.
 *
 * @note This function will not trigger any dispatcher events or update the
 * subscription list as it's just an internal function to be called by other
 * mDNS functions.
 *
 * @param record  Pointer to DNS record to be added or searched
 * @param task    register, learn or search record or reset cache
 * @return Id of registered, updated or found cache entry. Will be set to
 * NIP_MDNS_NO_CACHE_ID on error, nip_error will be set to one of the following
 * then:
 *  - NIP_E_LOCKED             cache locked, try again later
 *  - NIP_E_OUT_OF_MEMORY      unable to allocate memory for new record
 *  - NIP_E_OUT_OF_RESSOURECES cache full
 *  - NIP_E_NOT_FOUND          search operation was unsuccessfull
 *  - NIP_E_EXISTS             register failed because record exists already
 *  - NIP_E_INTERNAL_ERROR     something went wrong where it shouldn't have
 */
nip_error_t nip_mdns_cache_task( nip_mdns_task_t task,
nip_mdns_cache_id_t *id, struct nip_mdns_record *record )
{
	nip_mdns_cache_id_t    res = NIP_MDNS_NO_CACHE_ID;
	nip_mdns_cache_id_t    new_id = 1;
	nip_mem_size_t         listsize;
	nip_mem_ptr_t          list_ptr = { nip_mdns_cache.id, NULL };
	nip_mdns_record_size_t recsize;
	uint8_t                *start;
	struct nip_mdns_cache  *l; // pointer to list item

	listsize = nip_mem_buf_used( nip_mdns_cache.id );
	start = nip_mem_obtain_ptr( nip_mdns_cache.id );
	l     = (struct nip_mdns_cache *)start;

	if ( start == NULL )
	{
		nip_error = NIP_E_LOCKED;
		goto quit;
	}

	while ( (nip_mem_size_t)list_ptr.ptr != listsize )
	{
		l = (struct nip_mdns_cache *)(start + (nip_mem_size_t)list_ptr.ptr);
		recsize = nip_mdns_record_size( &l->record );

		// perform cache reset if requested
		if ( task == NIP_MDNS_TASK_RESET )
		{

			if ( l->stat & NIP_MDNS_CSTAT_AUTH_MASK )
			{
				l->stat = NIP_MDNS_CSTAT_PROBING;
				l->age  = 0;
			}
			else
				l->stat = NIP_MDNS_CSTAT_DELETE;

			goto next;
		}

		if ( task & NIP_MDNS_TASK_GET )
		{
			if ( l->id == *id )
			{
				nip_memcpy( record, &l->record, recsize );
				if ( task == NIP_MDNS_TASK_GET_LEARNED )
					record->ttl -= l->age;
				nip_error = NIP_E_OK;
				goto release;
			}

			nip_error = NIP_E_NOT_FOUND;
			goto next;
		}

		// For all other tasks: Check for matching record type and name
		if ( l->record.type == record->type
		&& nip_memcmp( &l->record.name, &record->name, sizeof( struct nip_mdns_name) ) == 0 )
		{
			if ( record->type != NIP_MDNS_TYPE_PTR
				|| nip_memcmp( &l->record.data, &record->data, sizeof(struct nip_mdns_rdata_a) ))
			///@todo support multiple entries for the same name?
			{
				res = l->id;
				nip_error = NIP_E_EXISTS;
				break;
			}
		}

		// check free ID
		if ( new_id == l->id )
		{
			new_id++;
			list_ptr.ptr = NULL;
			continue;
		}

		next:
			list_ptr.ptr += NIP_MDNS_BASE_CACHE_SIZE + recsize;
	}

	if ( task == NIP_MDNS_TASK_RESET )
	{
		nip_error = NIP_E_OK;
		goto release;
	}


	if ( ( res == NIP_MDNS_NO_CACHE_ID && (task & NIP_MDNS_TASK_REGISTER) )
	  || task == NIP_MDNS_TASK_LEARN )
	{
		recsize    = nip_mdns_record_size( record );

		// register new record names
		/// @todo nip_mdns_register_record_names( record, 1 )
		NIP_MDNS_NAME_USE( record->name.part, NIP_MDNS_MAX_NAME_DEPTH );
		/// @todo on error goto release

		if ( res == NIP_MDNS_NO_CACHE_ID )
		{
			if ( new_id >= NIP_MDNS_MAX_CACHE_SIZE )
			{
				nip_error = NIP_E_OUT_OF_RESSOURCES;
				goto release;
			}

			// release cache pointer, so it can be moved if necessary
			nip_mem_release_block( nip_mdns_cache.id );
			// add record to cache
			if ( NIP_E_OK !=
			nip_mem_write( nip_mdns_cache.id, NULL, NIP_MDNS_BASE_CACHE_SIZE + recsize ))
			{
				nip_error = NIP_E_OUT_OF_MEMORY;
				goto quit;
			}

			// get cache pointer
			l = nip_mem_ptr( &list_ptr );
			if ( l == NULL )
			{
				nip_error = NIP_E_INTERNAL_ERROR;
				// try to reset list size
				nip_mem_set_used( nip_mdns_cache.id, listsize );
				goto quit;
			}

			l->id = new_id;
			res = new_id;
		}
		else
		{
			// Unregister old record names
			/// @todo nip_mdns_register_record_names( l->record, -1 )
			NIP_MDNS_UNREG_NAME( l->record.name.part, NIP_MDNS_MAX_NAME_DEPTH );
			/// @todo on error, try to unregister new record names, set INTERNAL_ERROR
			/// because name-count may now be inconsistant and goto release.
		}

		if ( task == NIP_MDNS_TASK_LEARN )
		{
			///@todo set notification bit
			l->stat |= NIP_MDNS_CSTAT_LEARNED;
		}
		else if ( task == NIP_MDNS_TASK_REGUNIQUE )
		{
			l->stat = NIP_MDNS_CSTAT_PROBING;
		}
		else
		{
			l->stat = NIP_MDNS_CSTAT_ANNOUNCING;
		}

		// fill cache structure
		nip_memcpy( &l->record, record, recsize );
		l->age  = 0;

		// notify dispatcher
		nip_disp_notify( NIP_DISP_CHECK_MDNS );

	}
	else if ( task == NIP_MDNS_TASK_SEARCH )
	{
		if ( res == NIP_MDNS_NO_CACHE_ID )
			nip_error = NIP_E_NOT_FOUND;
		else
			nip_error = NIP_E_OK;
	}

	release:
	nip_mem_release_block( nip_mdns_cache.id );

	quit:
	if ( id != NULL && res != NIP_MDNS_NO_CACHE_ID )
		*id = res;

	return nip_error;
}

/** Edit query notification list
 *
 * @param task  whether to subscribe, unsubscribe or notify
 * @param c_id  ID of changed cache item or NIP_MDNS_NO_CACHE_ID to notify all queries.
 * @param q_id  ID of query item which is needed upon register.
 * @param nfy   Flag, who is to be notified. NIP_MDNS_NFY_LOCAL, or -_REMOTE
 *
 * @note for subscriptions, both c_id and q_id have to be set. No check will be
 * performed on this, so you'll corrupt the list if you feed it wrongly.
 */
nip_error_t nip_mdns_notify(nip_mdns_nfy_task_t task, nip_mdns_cache_id_t c_id,
nip_mdns_query_id_t q_id, uint8_t nfy )
{
	struct nip_mdns_cache_subscription *cs, *end, new;
	nip_mem_size_t listsize;
	nip_error_t res = NIP_E_OK;

	listsize = nip_mem_buf_used( nip_mdns_cs );
	cs = nip_mem_obtain_ptr( nip_mdns_cs );
	if ( cs == NULL )
	{
		res = NIP_E_LOCKED;
		goto quit;
	}

	end = (void*)(((uint8_t*)cs) + listsize);
	while ( cs < end )
	{
		// The following if-statement is TRUE for any matching combination of the
		// given q_id and c_id parameters with the current cache subscription.
		// E.g. q_id and c_id have to match both, if they're both set, or
		//      q_id has to match               , if c_id is not set, and so on.
		if( ( (q_id & NIP_MDNS_QUERY_ID_MASK) == (cs->q_id & NIP_MDNS_QUERY_ID_MASK)
		    && q_id != NIP_MDNS_NO_QUERY
		    &&(c_id == cs->c_id || c_id == NIP_MDNS_NO_CACHE_ID ))
		 || (  c_id == cs->c_id
		    && q_id == NIP_MDNS_NO_QUERY
		    && c_id != NIP_MDNS_NO_CACHE_ID)
		  )
		{
			if ( task & NIP_MDNS_NFYTASK_NOTIFY )
				cs->q_id |= nfy;

			if ( task == NIP_MDNS_NFYTASK_NOTIFIED )
				cs->q_id &= ~nfy;

			if ( task == NIP_MDNS_NFYTASK_SUBSCRIBE )
				break;

			if ( task == NIP_MDNS_NFYTASK_UNSUBSCRIBE )
			{
				// remove subscription
				nip_memmove( cs, cs+1, ((uint8_t*) --end) - ((uint8_t*) cs) );
				listsize -= sizeof( struct nip_mdns_cache_subscription );
			}
		}

		cs++;
	}

	nip_mem_release_block( nip_mdns_cs );

	if ( task == NIP_MDNS_NFYTASK_SUBSCRIBE && cs == end )
	{
		new.c_id     = c_id;
		new.q_id     = q_id | nfy;
		res = nip_mem_write( nip_mdns_cs, &new, sizeof( new ) );
		goto quit;
	}

	if ( task == NIP_MDNS_NFYTASK_UNSUBSCRIBE )
	{
		// set new subscription list size
		nip_mem_set_used( nip_mdns_cs, listsize );
	}

	quit:
	return res;
}


/** Close query.
 */
nip_error_t nip_mdns_close( nip_mdns_query_id_t r_id )
{
	/// @todo implement
	return NIP_E_OK;
}

/** Fetch next result for lookup query. This function has to be called
 * repeatedly, to fetch all results until nip_mdns_close() is called.
 * @param r_id  ID of query to fetch result from
 * @return ID of cache entry to result. If no result was available
 * NIP_MDNS_NO_CACHE_ID will be returned and nip_error be set to NIP_E_AGAIN.
 */
nip_mdns_cache_id_t nip_mdns_fetch_results( nip_mdns_query_id_t r_id )
{
	/// @todo implement
	return NIP_MDNS_NO_CACHE_ID;
}

/** Unregister previosly registered DNS record.
 */
nip_error_t nip_mdns_unregister( nip_mdns_cache_id_t id )
{
	/// @todo implement
	return NIP_E_OK;
}

/** set new host name.
 *
 * nip_error variable contains error code.
 * @todo should this function or part of it be moved to dns-sd?
 */
void nip_mdns_register_hostname( struct nip_dns_string *name )
{
	nip_net_if_t *net_if = &nip_net_if_list[ nip_mdns_if_id ];
	struct nip_mdns_record dnsrec;
	nip_mem_ptr_t          name_ptr = { NIP_MEM_NULL, (uint8_t*)name };
#if NIP_MDNS_CREATE_IN_ADDR_ARPA_PTR == 1
	uint8_t                name_str[8];
	uint8_t                i,r;  // integer and remainder
	uint8_t                part; // number of current name part
	uint8_t                name_pos;
#endif
	// fill DNS record for hostname
	dnsrec.type    = NIP_MDNS_TYPE_A;
	// add A-record TTL
	dnsrec.ttl     = NIP_MDNS_DEFAULT_TTL_A;
// 	dnsrec.name.part[0] = nip_mdns_register_name( NULL, 0, &name_ptr, len, 1 );
	dnsrec.name.part[0] = NIP_MDNS_REG_STRING( &name_ptr );
	dnsrec.name.part[1] = nip_mdns_domain;
	dnsrec.name.part[2] = NIP_MDNS_NO_NAMESTRING;
	nip_memcpy( dnsrec.data.a.ip_addr, net_if->ip_conf.addr, 4 );

	if ( dnsrec.name.part[0] == NIP_MDNS_NO_NAMESTRING )
	{
		goto quit;
	}

	// unregister current hostname
	if ( nip_mdns_hostname != NIP_MDNS_NO_CACHE_ID )
	{
		nip_error = nip_mdns_unregister( nip_mdns_hostname );
		if ( nip_error != NIP_E_OK )
			goto unregister_hostname_usage;
#if NIP_MDNS_CREATE_IN_ADDR_ARPA_PTR == 1
		nip_error = nip_mdns_unregister( nip_mdns_host_ptr );
#endif
	}

	// register DNS record for new hostname
#if NIP_MDNS_DEFEND_HOSTNAME == 1
	NIP_MDNS_REGUNIQUE( &nip_mdns_hostname, &dnsrec );
#else
	NIP_MDNS_REGISTER( &nip_mdns_hostname, &dnsrec );
#endif

	unregister_hostname_usage:
	NIP_MDNS_UNREG_NAME( dnsrec.name.part, 1 );

#if NIP_MDNS_CREATE_IN_ADDR_ARPA_PTR == 1
	// register reverse-ptr record
	if ( nip_mdns_hostname != NIP_MDNS_NO_CACHE_ID )
	{
		// copy hostname to ptr-record
		dnsrec.data.ptr.name.part[0] = dnsrec.name.part[0];
		dnsrec.data.ptr.name.part[1] = dnsrec.name.part[1];
		dnsrec.data.ptr.name.part[2] = dnsrec.name.part[2];
		part = 0;
		// set in-addr.arpa name
		do
		{
			name_ptr.ptr = name_str;
			// convert ip address too 4 namestrings and add them
			if ( part < 4 )
			{
				i = NIP_IP_ADDR( nip_mdns_if_id )[3-part];
				if      ( i>=100 ) name_pos = 3;
				else if ( i>=10  ) name_pos = 2;
				else               name_pos = 1;
				name_str[0] = name_pos;
				do {
					r = i%10;
					name_str[ name_pos-- ] = '0'+r;
				} while ((i/=10) != 0);
			}
			else if ( part == 4 )
			{
				// add "in-addr" domain part
				name_str[0] = 7;
				name_str[1] = 'i';
				name_str[2] = 'n';
				name_str[3] = '-';
				name_str[4] = 'a';
				name_str[5] = 'd';
				name_str[6] = 'd';
				name_str[7] = 'r';
			}
			else if ( part == 5 )
			{
				// add "arpa" domain part
				name_str[0] = 4;
				name_str[1] = 'a';
				name_str[2] = 'r';
				name_str[3] = 'p';
				name_str[4] = 'a';
			}
			dnsrec.name.part[ part ] = NIP_MDNS_REG_STRING( &name_ptr );
			if ( dnsrec.name.part[ part ] == NIP_MDNS_NO_NAMESTRING )
				goto unregister_ptrname_usage;
		} while (++part < 6 );
		// set record type and TTL
		dnsrec.type    = NIP_MDNS_TYPE_PTR;
		dnsrec.ttl     = NIP_MDNS_DEFAULT_TTL_A;

		// register DNS record for new hostname
		NIP_MDNS_REGUNIQUE( &nip_mdns_host_ptr, &dnsrec );
		/// @todo do something if pointer registration fails?

		unregister_ptrname_usage:
		NIP_MDNS_UNREG_NAME( dnsrec.name.part, 6 );
	}
#endif

//	nip_mdns_register_name( &dnsrec.name.part[0], -1, NULL, 0, -1 );

	quit:
	return;
}

/** Register host name and refresh cache
 */
void nip_mdns_interface_up( nip_net_if_id_t if_id )
{

	nip_mdns_if_id	 = if_id;
	nip_mdns.stat   = NIP_MDNS_STAT_INIT;

	// reset cache
	NIP_MDNS_RESET_CACHE();

	nip_disp_notify( NIP_DISP_CHECK_MDNS );
// 	nip_dispatcher();
}

void nip_mdns_interface_down( void )
{
	///@todo send notifications to network

	///@todo reset cache, clean up remote queries

	// close UDP socket
	nip_udp_close( nip_mdns_sock );

	// set status
	nip_mdns_if_id = NIP_NET_NO_IF;
	nip_mdns.stat  = NIP_MDNS_STAT_DOWN;
}

void nip_mdns_disp_check_cache( void )
{
	nip_mem_size_t         listsize;    // size of list (in bytes)
	nip_mdns_record_size_t csize;       // size of current item's dns record
	struct nip_mdns_cache  *l;          // current list item
	struct nip_mdns_cache  c;
	nip_mdns_query_id_t    q_id;
	nip_mdns_query_type_t  q_type;
	uint8_t                *item, *pstart, *name, i;
	nip_mem_size_t         pstart_off;  // offset of packet start in block
	nip_mem_ptr_t          item_ptr, name_ptr;    // pointer to scan position in packet
	nip_error_t            res;
	nip_mdns_type_t        rtype;


	listsize = nip_mem_buf_used( nip_mdns_cache.id );


	while ( (nip_mem_size_t)nip_mdns_cache.ptr < listsize )
	{
		l = nip_mem_ptr( &nip_mdns_cache );
		if ( l == NULL )
			goto quit;

		csize = nip_mdns_record_size( &l->record ) + NIP_MDNS_BASE_CACHE_SIZE;

		/** @todo remove items that are scheduled for deletion where all
		 * subscribers have been notified */
		// if l->stat == DELETE && notificationcount = 0
		// {
		// 	listsize -= recsize;
		// 	nip_memmove( l, (uint8_t*)l + recsize, listsize - (nip_mem_size_t)nip_mdns_cache.ptr);
		// 	nip_mem_release_ptr( &nip_mdns_cache );
		// 	nip_mem_set_used( nip_mdns_cache.id, listsize );
		// 	continue;
		// }

		// make local copy of cache item to work with
		nip_memcpy( &c, l, csize);

		nip_mem_release_ptr( &nip_mdns_cache );

		// Check if current cache record correlates somehow with received query.
		// Only authorative or probing records are relevant for this
		if ( c.stat & NIP_MDNS_CSTAT_AUTH_BIT
		  && nip_mdns.recv & NIP_MDNS_PTYPE_QUERY )
		{
			// set start item_ptr before any item, so the next parse_packet call
			// will get a pointer to the first item.
			nip_udp_data_mem_ptr( nip_mdns_sock, NIP_UDP_RX_BUF, &item_ptr );
			pstart_off   = (nip_mem_size_t)item_ptr.ptr;

			// loop through queries in query-packet
			do
			{
				res = nip_mdns_parse_packet(
						NIP_UDP_RX_BUF,
						NIP_MDNS_PTASK_NEXT_ITEM,
						&item_ptr,
						NIP_MDNS_QDCOUNT_OFFSET );

				pstart = nip_udp_data_ptr( nip_mdns_sock, NIP_UDP_RX_BUF );

				if ( res == NIP_E_OK && pstart != NULL)
				{
					// parse query
					item = pstart + (nip_mem_size_t)item_ptr.ptr - pstart_off;
					name = item;
					i    = 0;
					name_ptr.id  = NIP_MEM_NULL;
					// check name
					do
					{
						name_ptr.ptr = name;

						// name pointer? follow
						if ( *name & 0xC0 )
						{
// 							name = pstart + (ntohs(*((uint16_t*)name)) & 0x3F00 );
							name = pstart + ((name[0]&0x3F)<<8) + name[1];
						}
						// end of name?
						else if ( *name == 0x00 )
						{
							if ( (c.record.name.part[i] != NIP_MDNS_NO_NAMESTRING
							   && i != NIP_MDNS_MAX_NAME_DEPTH - 1 )
							  || i >= NIP_MDNS_MAX_NAME_DEPTH )
								res = NIP_E_NO_MATCH;
						}
						else
						{
							// compare namestrings
							if ( i >= NIP_MDNS_MAX_NAME_DEPTH
							|| NIP_MDNS_CHECK_NAME( &name_ptr ) != c.record.name.part[i] )
								res = NIP_E_NO_MATCH;
							name += *name + 1;
							i++;
						}

						if ( *item < 0xC0 && *item != 0x00)
							item += *item + 1;

						// overflow protection for name, so broken packets wont send
						// us in some infinte loop
						if ( name >= pstart - pstart_off + nip_mdns.recv_size )
						{
							res = NIP_E_EOF;
						}

					} while ( *name != 0x00 && res == NIP_E_OK );
					if (*item == 0x00 ) item++;
					else if (*item &  0xC0 ) item+=2;

					// check type and class
					rtype = ntohs(*(uint16_t*)item);
					if ( (rtype != c.record.type && rtype != NIP_MDNS_TYPE_ANY )
					  || item[3] != NIP_DNS_CLASS_IN )
					{
						res = NIP_E_NO_MATCH;
					}
					///  @todo else check known answers

				}

				// release block
				nip_udp_ptr_release( nip_mdns_sock, NIP_UDP_RX_BUF );

				if ( res == NIP_E_OK )
				{
					// register remote query, we have auth answer for it in cache
					q_id = nip_mdns_query( NIP_MDNS_QTYPE_REMOTE, &c.record.name, c.record.type );
					if ( q_id != NIP_MDNS_NO_QUERY )
					{
						// add cache subscription
						NIP_MDNS_SUBSCRIBE_REMOTE( c.id, q_id );
					}
				}

			} while ( res != NIP_E_EOF );

				/// @todo received Probe that we have Probe for?
					/// @todo -> handle probe conflict

			/// @todo out-of-memory? -> goto release
		}


		/// @todo check if there are unnotified remote queries to this

		/// @todo check if there's still a query interested in this item
		/// @todo update age and possibly trigger delete or refresh

		/// @todo check probe and announcement states of registered cache records
		if ( c.stat == NIP_MDNS_CSTAT_PROBING
		  || NIP_MDNS_GET_AUTH_CSTAT( &c ) == NIP_MDNS_CSTAT_ANNOUNCING)
		{
			/// @todo allow announcement of non-unique records
			if( c.age < NIP_MDNS_PROBE_COUNT + NIP_MDNS_ANNOUNCE_COUNT )
			{
				q_type = NIP_MDNS_QTYPE_PROBE;
				if ( c.age >= NIP_MDNS_PROBE_COUNT )
				{
					q_type = NIP_MDNS_QTYPE_ANNOUNCE;
					if ( c.stat == NIP_MDNS_CSTAT_PROBING )
						c.stat = NIP_MDNS_CSTAT_ANNOUNCING;
// 					NIP_MDNS_SET_AUTH_CSTAT( &c, NIP_MDNS_CSTAT_ANNOUNCING );
				}

				q_id = nip_mdns_query( q_type, &c.record.name, NIP_MDNS_TYPE_ANY );

				if ( q_id != NIP_MDNS_NO_QUERY
				&& NIP_MDNS_SUBSCRIBE_REMOTE( c.id, q_id ) == NIP_E_OK )
				{
					// update timer (probe/announce counter)
					c.age++;
				}
			}
			else
			{
				NIP_MDNS_SET_AUTH_CSTAT( &c, NIP_MDNS_CSTAT_UNIQUE_AUTH );
			}
		}

		// write cache entry
		l = nip_mem_ptr( &nip_mdns_cache );
		if ( l == NULL )
			goto quit;

		nip_memcpy( l, &c, csize );
		nip_mem_release_block( nip_mdns_cache.id );

		/// @todo check if local hostname record changed and update all registered
		///       records if necessary

		// increment list offset
		nip_mdns_cache.ptr = (void*)((nip_mem_size_t)nip_mdns_cache.ptr + csize);
	}

	// Checked all Cache items? -> reset offset
	nip_mdns_cache.ptr = (void*)0;

	quit:
		nip_mdns.stat = NIP_MDNS_STAT_QUERY_CHECK;
		NIP_CURR_CMD  = &nip_mdns_disp_check_queries;
		return;
}

/** Initialize Sending of packet.
 *
 * If successfull the global nip_mdns.send variable will be set to ptype. Only
 * exception is, if that variable is NIP_MDNS_PTYPE_ANSWCONTINUED. In that case
 * only ptypes == NIP_MDNS_PTYPE_QUERY will be accepted and nip_mdns.send will
 * remain untouched to allow further special handling of that situation.
 */
nip_error_t nip_mdns_tx_init( nip_mdns_packet_type_t ptype )
{
	struct nip_dns_header *header;
	nip_error_t           res = NIP_E_OK;

	if ( nip_mdns.send == NIP_MDNS_PTYPE_NONE
	  || ( nip_mdns.send == NIP_MDNS_PTYPE_ANSWCONTINUED
	    && ptype         == NIP_MDNS_PTYPE_QUERY ))
	{
		res    = nip_udp_init_send ( nip_mdns_sock, NIP_DNS_HSIZE  );
		res    = nip_udp_write     ( nip_mdns_sock, NULL, NIP_DNS_HSIZE  );
		header = nip_udp_data_ptr  ( nip_mdns_sock, NIP_UDP_TX_BUF );
		if ( res != NIP_E_OK || header == NULL )
		{
			// Come back and try again later. Probably out-of-memory
			nip_udp_cancel_send( nip_mdns_sock );
			res = NIP_E_AGAIN;
			goto quit;
		}
		if ( nip_mdns.send != NIP_MDNS_PTYPE_ANSWCONTINUED )
			nip_mdns.send = ptype;

		if ( ptype == NIP_MDNS_PTYPE_RESPONSE )
		{
			header->qr_opcode_aa_tc_rd = ( NIP_DNS_QR_MASK | NIP_DNS_AA_MASK );
		}
		nip_udp_ptr_release( nip_mdns_sock, NIP_UDP_TX_BUF );
	} else
		if ( ptype != nip_mdns.send )
			res = NIP_E_AGAIN;

	quit:
	return res;
}

/** Add DNS Name/Type/Class to Buffer. The dst-pointer will be forward to the
 * end of the written name. That can serve as indication of the name length,
 * given that the original pointer has been saved by the callee.
 */
nip_error_t nip_mdns_tx_name( nip_mem_ptr_t *dst, struct nip_mdns_name *rr_name,
nip_mdns_type_t rr_type, uint8_t uclass_byte )
{
	nip_error_t           res;
	nip_mem_ptr_t         typeclass_ptr;
	uint8_t               typeclass[4];

	// add name
	res = NIP_MDNS_WRITE_NAME( rr_name->part, NIP_MDNS_MAX_NAME_DEPTH, dst );
	if ( res != NIP_E_OK )
		goto quit;

	typeclass[0] = 0;
	typeclass[1] = rr_type;
	typeclass[2] = uclass_byte;
	typeclass[3] = NIP_DNS_CLASS_IN;
	typeclass_ptr.id  = NIP_MEM_NULL;
	typeclass_ptr.ptr = typeclass;

	// add type and class
	res = nip_mem_insert( dst, &typeclass_ptr, 4 );
	/// @todo Handle error cases. For that backup the original dst-pointer and
	/// remove all data inbetween the old and the new pointer.

	quit:
	return res;
}

void nip_mdns_transmit( void )
{
	struct nip_udp_sock_addr addr;

	// fill address structure
	addr.port  = NIP_UDP_PORT_MDNS;
	addr.ip[0] = NIP_MDNS_ADDR0;
	addr.ip[1] = NIP_MDNS_ADDR1;
	addr.ip[2] = NIP_MDNS_ADDR2;
	addr.ip[3] = NIP_MDNS_ADDR3;

	// no queries or answers in packet? -> discard, else transmit
	if ( nip_udp_transmit( nip_mdns_sock, &addr, nip_mdns_if_id ) == NIP_E_OK )
	{
		nip_mdns.send = NIP_MDNS_PTYPE_NONE;
	}

}

/** Perform several tasks on DNS packets.
 *
 * For task NIP_MDNS_PTASK_NEXT_ITEM the return value may be NIP_E_EOF if there
 * were no more items in the given section beyond the given pointer p.
 */
nip_error_t nip_mdns_parse_packet( nip_udp_buf_type_t udp_buffer,
nip_mdns_packet_task_t task, nip_mem_ptr_t *p, nip_mdns_qa_offset_t section )
{
	uint8_t  *pos, len, *end;
	uint16_t count;
	nip_error_t           res = NIP_E_OK;
	nip_mem_ptr_t         ptmp;
	struct nip_dns_header *header;
	nip_mem_size_t        pstart_off, bufsize;
	nip_mdns_qa_offset_t  sec = 0;

	nip_udp_data_mem_ptr( nip_mdns_sock, udp_buffer, &ptmp );

	bufsize    = nip_mem_buf_used( ptmp.id );
	header     = nip_mem_ptr( &ptmp );
	pstart_off = (nip_mem_size_t)ptmp.ptr;
	if ( header == NULL )
	{
		res = nip_error;
		goto quit;
	}

	// start right after header
	pos = (uint8_t*)(header+1);
	end = (uint8_t*)(header) + bufsize;

	// set Truncated-Flag
	if ( task == NIP_MDNS_PTASK_SET_TC )
	{
		header->qr_opcode_aa_tc_rd |= NIP_DNS_TC_MASK;
		goto release;
	}


	// move to specified section in packet
	do
	{
		// get number of items in section
		count = ntohs( header->qa_count[ sec ] );

		// found section?
		if ( sec == section )
		{
			if ( task == NIP_MDNS_PTASK_SECTION_PTR )
			{
				goto release;
			}
			if ( task == NIP_MDNS_PTASK_INC_COUNTER )
			{
				header->qa_count[ section ] = htons( count+1 );
				goto release;
			}
		}

		// move to end of section (begin of next section)
		while ( count > 0 )
		{
			ptmp.ptr = (void*)(pstart_off + pos - (uint8_t*)header);

			if ( task == NIP_MDNS_PTASK_NEXT_ITEM && ptmp.ptr > p->ptr)
				goto release;

			// name
			do
			{
				// forward pos pointer to skip name
				len = *pos;
				if ( len >= 0xC0 ) { pos += 2; break; }
				if ( len == 0x00 ) { pos += 1; break; }
				// overflow protection against bad packets
				if ( pos >= end )
				{
					res = NIP_E_EOF;
					goto release;
				}
			} while ( pos+=len+1 );

			// type and class
			pos += 4;

			// record data
			if ( sec != NIP_MDNS_QDCOUNT_OFFSET )
			{
				// TTL
				pos += 4;
				// Len + Data
				pos += 2 + *pos;
			}

			// overflow protection against bad packets
			if ( pos >= end )
			{
				res = NIP_E_EOF;
				goto release;
			}
			count--;
		}

		if ( task == NIP_MDNS_PTASK_NEXT_ITEM && sec == section )
		{
			// no more items in section. -> Error
			res = NIP_E_EOF;
			goto release;
		}

	} while ( ++sec < NIP_MDNS_QA_COUNTERS_NUMBER );

	release:
		nip_mem_release_ptr( &ptmp );

	// set return pointer
	if ( p != NULL && res == NIP_E_OK )
	{
		p->id  = ptmp.id;
		p->ptr = (void*)(pstart_off + pos - (uint8_t*)header);
	}

	quit:
	return res;
}




/** Check list of mDNS queries
 */
void nip_mdns_disp_check_queries( void )
{
	nip_mem_size_t         listsize;    // size of list (in bytes)
	struct nip_mdns_query  *l;          // current list item
	struct nip_mdns_query  q;           // query structure for "offline"-work
	struct nip_mdns_record r;           // record structure for "offline"-work
	uint8_t                count = 0;   // count items in transmit-packet
	nip_error_t            res = NIP_E_OK;
	nip_mem_ptr_t          cs_ptr, insert_ptr, src_ptr;
	nip_mem_size_t         section_off, head_off, len;
	nip_mdns_packet_type_t ptype;
	nip_mdns_qa_offset_t   answer_section;
	nip_mdns_cache_id_t    c_id;
	uint8_t                name_flag;
// 	nip_mdns_namestring_id_t *name;
	struct {
			uint16_t ttl[2];
			uint16_t len;
	} data_header;
	struct nip_mdns_cache_subscription *cs;

	while ( (l = nip_mem_ptr( &nip_mdns_queries )) != NULL )
	{
		// make local copy of cache item to work with
		nip_memcpy( &q, l, sizeof( struct nip_mdns_query ));
		/** @todo remove items that have no more subscribers */
		// if l->stat == DELETE && subscribercount = 0
		// {
		// 	listsize -= sizeof( struct nip_mdns_query );
		// 	nip_memmove( l, (uint8_t*)l + sizeof( struct nip_mdns_query ), listsize - (nip_mem_size_t)nip_mdns_queries.ptr);
		// 	nip_mem_release_ptr( &nip_mdns_queries );
		// 	nip_mem_set_used( nip_mdns_queries.id, listsize );
		// 	continue;
		// }

		nip_mem_release_ptr( &nip_mdns_queries );

		/// @todo loop answers a in Packet
		/// @todo a saisfies q ?
		///          q=local? -> add a to cache, subscribe/notify q,a
		///          q=probe? -> CONFLICT -> select new name, update cache items
		///          q=remote & a=probe? -> schedule immediate response (timeout=0)

		/// @todo update query timer

		/// if timeout==0
		///          q=local?  -> @todo configure query
		///          @todo loop notifications,
		///              q=remote? -> @todo configure query response
		///              @todo add answers
		answer_section = NIP_MDNS_QDCOUNT_OFFSET;
		if ( q.timer == 0 && q.type & NIP_MDNS_QTYPE_TX_BIT )
		{
			// determine packet type
			if ( q.type == NIP_MDNS_QTYPE_PROBE )
			{
				ptype          = NIP_MDNS_PTYPE_PROBE;
				answer_section = NIP_MDNS_NSCOUNT_OFFSET;
			}
			else
			{
				ptype = NIP_MDNS_PTYPE_QUERY;
				answer_section = NIP_MDNS_ANCOUNT_OFFSET;
			}

			// Can only transmit one probe at once, currently.
			res = nip_mdns_tx_init( ptype );

			if ( res == NIP_E_OK )
			{
				// move pointer to query section
				nip_mdns_parse_packet(
				   NIP_UDP_TX_BUF,
				   NIP_MDNS_PTASK_SECTION_PTR,
				   &insert_ptr,
				   NIP_MDNS_QDCOUNT_OFFSET );

				// No new query will be created for NIP_MDNS_PTYPE_ANSWCONTINUED
				if ( nip_mdns.send != NIP_MDNS_PTYPE_ANSWCONTINUED )
				{
					// Add name to packet
					if ( insert_ptr.id == NIP_MEM_NULL )
						goto quit;
					res = nip_mdns_tx_name( &insert_ptr, &q.rr_name, q.rr_type, 0 );
					if ( res != NIP_E_OK )
						goto quit;
					// increment query-counter
					nip_mdns_parse_packet(
						NIP_UDP_TX_BUF,
						NIP_MDNS_PTASK_INC_COUNTER,
						&insert_ptr,
						NIP_MDNS_QDCOUNT_OFFSET );
					count++;
				}
				else
				{
					// continue as if a query has been created. That query was just
					// part of a previous packet.
					nip_mdns.send = NIP_MDNS_PTYPE_QUERY;
				}
			}
			else
			{
				// clear answer selector
				answer_section = NIP_MDNS_QDCOUNT_OFFSET;
			}
		}
		cs_ptr.id  = nip_mdns_cs;
		cs_ptr.ptr = (void*)0;
		src_ptr.id = NIP_MEM_NULL;

		// check if query-answers need be sent
		if (   answer_section != NIP_MDNS_QDCOUNT_OFFSET
		    || (   q.type & NIP_MDNS_QTYPE_REMOTE
		        && (res = nip_mdns_tx_init( NIP_MDNS_PTYPE_RESPONSE )) == NIP_E_OK
		        && (answer_section = NIP_MDNS_ANCOUNT_OFFSET) )
			)
		{
			// move pointer to answer section
			nip_mdns_parse_packet(
				NIP_UDP_TX_BUF,
				NIP_MDNS_PTASK_SECTION_PTR,
				&insert_ptr,
				answer_section );

			// remember section offset
			section_off = (nip_mem_size_t)insert_ptr.ptr;


			// check subscriptions
			// get subscription pointer
			while ( (cs = nip_mem_ptr( &cs_ptr )) != NULL )
			{
				// check for Remote subscriptions. Local subscriptions will not be
				// transmitted on the network as answers, as they mark the notification
				// state of the local user.
				c_id = NIP_MDNS_NO_CACHE_ID;
				if((cs->q_id & NIP_MDNS_NFY_REMOTE)
				&& (cs->q_id & NIP_MDNS_QUERY_ID_MASK) == (q.id & NIP_MDNS_QUERY_ID_MASK))
				{
					c_id = cs->c_id;
				}
				nip_mem_release_block( nip_mdns_cs );
				if ( c_id == NIP_MDNS_NO_CACHE_ID )
					goto next_cs;

				insert_ptr.ptr = (void*)section_off;
				len            = 0;

				// get ressource record
				res = nip_mdns_cache_task( nip_mdns.send == NIP_MDNS_PTYPE_QUERY
				        ? NIP_MDNS_TASK_GET_LEARNED
					     : NIP_MDNS_TASK_GET
					   , &c_id, &r );
				if ( res != NIP_E_OK )
					goto answer_abort;

				// add Ressource Name to packet
				name_flag = 0;
				if ( q.type == NIP_MDNS_QTYPE_ANNOUNCE )
					name_flag = NIP_DNS_CF_FLAG;
				res = nip_mdns_tx_name( &insert_ptr, &r.name, r.type, name_flag );
				if ( res != NIP_E_OK )
					goto answer_abort;

				// remember current offset for later inclusion of data header
				head_off = (nip_mem_size_t)insert_ptr.ptr;

				// add Ressource Record Data
				// special treatment of recognized types
				src_ptr.id  = NIP_MEM_NULL;
				src_ptr.ptr = (uint8_t*)&r.data;
				if ( r.type == NIP_MDNS_TYPE_A )
				{
					len = 4;
				}
				else
				if ( r.type == NIP_MDNS_TYPE_SRV )
				{
					len = 6;
					r.data.srv.priority = htons( r.data.srv.priority );
					r.data.srv.weight   = htons( r.data.srv.weight   );
					r.data.srv.port     = r.data.srv.port;
				}
				else
				if ( r.type == NIP_MDNS_TYPE_PTR )
				{
					res = NIP_MDNS_WRITE_NAME( r.data.ptr.name.part, NIP_MDNS_MAX_NAME_DEPTH, &insert_ptr );
					if ( res != NIP_E_OK )
						goto answer_abort;
				}
				else
				{
					// add strings
					src_ptr.id  = r.data.oth.string;
					src_ptr.ptr = (void*)0;
					len = nip_mem_buf_used( r.data.oth.string );
				}

				res = nip_mem_insert( &insert_ptr, &src_ptr, len );
				if ( res != NIP_E_OK )
				{
					len = 0;
					goto answer_abort;
				}

				if ( r.type == NIP_MDNS_TYPE_SRV )
				{
					// add srv-name
					res = NIP_MDNS_WRITE_NAME( r.data.srv.target.part, NIP_MDNS_MAX_NAME_DEPTH, &insert_ptr );
					if ( res != NIP_E_OK )
						goto answer_abort;
				}

				// add Ressource Record Header before data
				len                = (nip_mem_size_t)insert_ptr.ptr-head_off;
				data_header.ttl[0] = 0;
				data_header.ttl[1] = htons( r.ttl );
				data_header.len    = htons( len );
				insert_ptr.ptr     = (void*)head_off;
				src_ptr.id         = NIP_MEM_NULL;
				src_ptr.ptr        = (uint8_t*)&data_header;
				res = nip_mem_insert( &insert_ptr, &src_ptr, sizeof( data_header ) );
				if ( res != NIP_E_OK )
					goto answer_abort;

				// answer added successfully? -> count
				nip_mdns_parse_packet( NIP_UDP_TX_BUF, NIP_MDNS_PTASK_INC_COUNTER, &insert_ptr, answer_section  );
				count++;

				// unsubscribe probes, unnotify normal queries
				NIP_MDNS_QUERY_NOTIFIED( c_id, q.id );

				next_cs:
				// move on to next subscription
				cs_ptr.ptr += sizeof( struct nip_mdns_cache_subscription );
			}
			goto query_done;

			answer_abort:
					// remove incomplete answer from packet
					len += (nip_mem_size_t)insert_ptr.ptr - section_off;
					insert_ptr.ptr = (void*) section_off;
					nip_mem_cut( &insert_ptr, len );

					// set NIP_MDNS_PTYPE_ANSWCONTINUED if necessary
					if ( nip_mdns.send == NIP_MDNS_PTYPE_QUERY )
					{
						nip_mdns.send = NIP_MDNS_PTYPE_ANSWCONTINUED;
						// Set Truncate-Bit in DNS header to signal more answers are
						// coming.
						nip_mdns_parse_packet( NIP_UDP_TX_BUF, NIP_MDNS_PTASK_SET_TC, &insert_ptr, 0 );
					}
				goto quit;

		}

		query_done:
		// all answers sent? -> Update Query in list

		// if query is not local, delete after transmission of all answers
		if ( !(q.type & NIP_MDNS_QTYPE_LOCAL_BIT ) )
		{
			if ( NIP_E_OK == nip_mem_cut( &nip_mdns_queries, sizeof( struct nip_mdns_query ) ))
			{
				// remove subscriptions
				NIP_MDNS_UNSUBSCRIBE_QUERY( q.id );
			}
			else
			{
				// something went wrong, come back and try again, later.
				goto quit;
			}
		}
		else
		{
			// save query data
			l = nip_mem_ptr( &nip_mdns_queries );
			if ( l == NULL )
			{
				nip_error = NIP_E_LOCKED;
				goto quit;
			}
			nip_memcpy( l, &q, sizeof( struct nip_mdns_query ));
			nip_mem_release_ptr( &nip_mdns_queries );

			/// @todo remove REMOTE or PROBE Query und remove subscriptions
			// increment list offset
			nip_mdns_queries.ptr = (void*)((nip_mem_size_t)nip_mdns_queries.ptr + sizeof( struct nip_mdns_query ));
		}
	}


	listsize = nip_mem_buf_used( nip_mdns_queries.id );
	if ( (nip_mem_size_t) nip_mdns_queries.ptr == listsize )
	{
		// Checked all Cache items? -> reset offset
		nip_mdns_queries.ptr = NULL;

		// Reset mDNS status
		nip_mdns.stat = NIP_MDNS_STAT_UP;
	}

	// Done with received Response-Packet -> release
	if ( nip_mdns.recv & NIP_MDNS_PTYPE_RESPONSE )
	{
		nip_udp_cancel_recv( nip_mdns_sock );
		nip_mdns.recv = NIP_MDNS_PTYPE_NONE;
	}

	quit:
#if NIP_MDNS_DISCARD_ON_OUT_OF_MEMORY == 1
		if ( res == NIP_E_OUT_OF_MEMORY
		  && count == 0
		  && nip_mdns.recv != NIP_MDNS_PTYPE_NONE )
		{
			// remove packet from receive buffer to free needed memory
			nip_udp_cancel_recv( nip_mdns_sock );
			nip_mdns.recv = NIP_MDNS_PTYPE_NONE;

			// try again.. we might have enough space now
			nip_disp_notify( NIP_DISP_CHECK_MDNS );
		}
#endif

		if  ( nip_mdns.send != NIP_MDNS_PTYPE_NONE )
		{
			// transmit only if there was any data added to the packet
			if ( count > 0 )
				NIP_CURR_CMD = &nip_mdns_transmit;
			else
			{
				nip_udp_cancel_send( nip_mdns_sock );
				nip_mdns.send = NIP_MDNS_PTYPE_NONE;
			}
		}
		return;
}

/** Dispatcher function to be called upon receiving mDNS events
 */
void nip_mdns_disp_check( void )
{
	struct nip_dns_header    *header;
	struct nip_udp_sock_addr rem_addr;

	// mDNS ready to receive?
	if ( ( nip_mdns.stat == NIP_MDNS_STAT_DOWN )
	  || ( nip_mdns.stat == NIP_MDNS_STAT_INIT
	    && nip_mdns_init() != NIP_E_OK )
		)
	{
		// discard all incoming mdns packets
		while ( nip_mdns_sock != NIP_UDP_NO_SOCKET
		     && nip_udp_recvfrom( nip_mdns_sock, NULL, 0, NULL ) > 0 )
		{
			nip_udp_cancel_recv( nip_mdns_sock );
		}
		return;
	}

	// Done parsing previously received mDNS packet?
	if ( nip_mdns.stat != NIP_MDNS_STAT_QUERY_CHECK
	  && nip_mdns.stat != NIP_MDNS_STAT_CACHE_CHECK
	  && nip_mdns.recv != NIP_MDNS_PTYPE_NONE )
	{
		nip_udp_cancel_recv( nip_mdns_sock );
		nip_mdns.recv = NIP_MDNS_PTYPE_NONE;
	}

	// No mDNS packet being processed right now?
	if ( nip_mdns.recv == NIP_MDNS_PTYPE_NONE )
	{
		do
		{
			// check UDP socket for new data
			nip_mdns.recv_size = nip_udp_recvfrom( nip_mdns_sock, NULL, 0, &rem_addr );

			if ( nip_mdns.recv_size > 0 )
			{

				/// @todo check that the packet originated on the local link

				// discard packets with invalid size
				if ( nip_mdns.recv_size < sizeof( struct nip_dns_header ) )
				{
					goto discard;
				}

				header = nip_udp_data_ptr( nip_mdns_sock, NIP_UDP_RX_BUF );
				// failed to retrieve pointer to packet data?.. try again later
				if ( header == NULL )
					break;

				// check packet type
				if ( (header->qr_opcode_aa_tc_rd & NIP_DNS_QR_MASK) > 0 )
					nip_mdns.recv = NIP_MDNS_PTYPE_RESPONSE;
				else
				{
					if ( header->qa_count[ NIP_MDNS_QDCOUNT_OFFSET ] != 0
					  && header->qa_count[ NIP_MDNS_NSCOUNT_OFFSET ] != 0 )
						nip_mdns.recv = NIP_MDNS_PTYPE_PROBE;
					else
						nip_mdns.recv = NIP_MDNS_PTYPE_QUERY;
				}

				// check header fields, according to mDNS standard
				// Discard own Probes
				if ( (header->qr_opcode_aa_tc_rd & NIP_DNS_OP_MASK) > 0
				  || (header->ra_z_ad_cd_rcode   & NIP_DNS_RC_MASK) > 0
				  || ( nip_mdns.recv == NIP_MDNS_PTYPE_PROBE
				    && nip_memcmp( NIP_IP_ADDR( nip_mdns_if_id ), rem_addr.ip, 4 ) == 0 )
				   )
				{
					nip_udp_ptr_release( nip_mdns_sock, NIP_UDP_RX_BUF );
					goto discard;
				}

				// end header check
				nip_udp_ptr_release( nip_mdns_sock, NIP_UDP_RX_BUF );
			}

			continue;

			discard:
				nip_mdns.recv = NIP_MDNS_PTYPE_NONE;
				nip_udp_cancel_recv( nip_mdns_sock );

		// -> repeat until we have valid or no packet
		} while ( nip_mdns.recv == NIP_MDNS_PTYPE_NONE && nip_mdns.recv_size > 0 );
	}

	if ( nip_mdns.stat == NIP_MDNS_STAT_QUERY_CHECK )
	{
		NIP_CURR_CMD = &nip_mdns_disp_check_queries;
		goto moveon;
	}

	// check cache
	NIP_CURR_CMD = &nip_mdns_disp_check_cache;

	/// @todo loop through queries
		/// @todo -> loop through answers in packet
				/// @todo answer matches local query?
				///       -> update cache
				///       -> update subscriptions
				/// @todo answer matches remote query?
				///       -> update subscriptions for matching cache entries

	/// @todo loop through queries
		/// @todo remote query? -> loop through subscription list
		///       -> retrieve answers from cache and fill packet
		/// @todo local query?
		///       -> check timer
		/// @todo remove query marked for deletion and all notifications done?
		///       -> remove from list
		///       -> remove subscriptions

	/// @todo loop through namestrings to remove unused names

	moveon:
	return;

}


#endif /* NIP_ENABLE_MDNS == 1 */
