/*##############################################################################

nIP - nano IP stack

File        : dispatcher.c

Description : Functions to serve ethernet interfaces

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#include "dispatcher.h"
#include "net/net_if.h"
#include "net/arp.h"
#include "app/dhcp.h"
#include "net/autoip.h"
#include "app/mdns.h"

#if (NIP_DISP_ENABLE==1)

extern nip_net_if_t nip_net_if_list[];
nip_disp_state_t nip_disp;

/** Notify dispatcher about one or more events */
// void nip_disp_notify( uint8_t events )
// {
// 	nip_disp.flags |= events;
// }

/** Notify dispatcher to handle incoming ethernet data on given interface */
void nip_disp_notify_if( nip_net_if_id_t net_if_id, uint8_t event )
{
	nip_disp.if_flags[ net_if_id ] |= event;
// 	nip_disp.flags |= NIP_DISP_CHECK_IF;
	nip_disp_notify( NIP_DISP_CHECK_IF );
}

/** Checks if dispatcher is running.
 *
 * @return NIP_E_OK if dispatcher is not running, or NIP_E_DISP_RUNNING else
 */
nip_error_t nip_disp_check()
{
	if ( (nip_disp.flags & NIP_DISP_RUNNING) != 0 )
		return NIP_E_DISP_RUNNING;
	else
		return NIP_E_OK;
}

/** Call nIP stack dispatcher. The dispatcher will check for accumulated events
 * and act accordingly. That is, it will initiate packet reception, transmission
 * and retransmission. Events are used to put the dispatcher on track.
 *
 * Only one instance of the dispatcher can be running at a time. It will return
 * NIP_E_DISP_RUNNING if there is another instance active. That instance will
 * take care of all events so no event will be lost.
 *
 * For blocking operations that need to wait for the dispatcher to finish in
 * order to complete what they're doing, the nip_disp_check() function may be
 * used to check the dispatcher status.
 *
 * @note Interrupting the dispatcher may lead to deadlock in non-multitasking
 * environments, if you attempt to wait for the dispatcher to to finish within
 * the interrupt. E.g. Do NOT wait for the already running dispatcher to handle
 * your event/request/packet if you're not in a multicasting environment.
 */
nip_error_t nip_dispatcher()
{
	void (*tmp_cmd)(void);

	// only allow one instance of dispatcher to be active
	if (nip_disp_check() == NIP_E_DISP_RUNNING) return NIP_E_DISP_RUNNING;

	// set dispatcher status
	nip_disp.flags |= NIP_DISP_RUNNING;

	// Repeat processing until all pending events have been handled. This has to
	// be a loop since new events may be added as the dispatcher is running.
	while ( nip_disp.flags > NIP_DISP_RUNNING )
	{
		NIP_CURR_CMD = NULL;

		// Handle Network Interface Events
		if ( nip_disp.flags & NIP_DISP_CHECK_IF )
		{
			// reset check_if flag
			nip_disp.flags &= ~NIP_DISP_CHECK_IF;

			// check all interfaces for events
			SET_CURR_IF( 0 );
			#if NIP_NET_IF_COUNT>1
			do
			{
			#endif
				// check for newly received (ethernet) packets
				if ( nip_disp.if_flags[ NIP_CURR_IF ] & NIP_DISP_IF_ETH_R )
				{
// 					NIP_CURR_CMD = NIP_DC_HARD_RECV;
					nip_disp.if_flags[ NIP_CURR_IF ] &= ~NIP_DISP_IF_ETH_R;
					NIP_CURR_CMD = nip_net_if_list[ NIP_CURR_IF ].phy_conf.hard_disp_recv;

					// set check_if flag to make sure we didn't miss any event
					nip_disp_notify( NIP_DISP_CHECK_IF );
					goto eval_cmd;
				}

				if ( nip_disp.if_flags[ NIP_CURR_IF ] & NIP_DISP_IF_TIMER )
				{
					nip_disp.if_flags[ NIP_CURR_IF ] &= ~NIP_DISP_IF_TIMER;
					nip_disp.if_flags[ NIP_CURR_IF ] = nip_disp.if_flags[NIP_CURR_IF]
					  #if NIP_DHCP_ENABLE == 1
						 | NIP_DISP_IF_DHCP
					  #endif
					  #if NIP_AUTOIP_ENABLE == 1
						 | NIP_DISP_IF_AUTOIP
					  #endif
					  #if NIP_IGMP_ENABLE == 1
						 | NIP_DISP_IF_GROUP
					  #endif
							;
				}

				#if ( NIP_DHCP_ENABLE == 1 )
				if ( nip_disp.if_flags[ NIP_CURR_IF ] & NIP_DISP_IF_DHCP )
				{
					nip_disp.if_flags[ NIP_CURR_IF ] &= ~NIP_DISP_IF_DHCP;
					NIP_CURR_CMD = &nip_dhcp_disp_check;
					// set check_if flag to make sure we didn't miss any event
					nip_disp_notify( NIP_DISP_CHECK_IF );
					goto eval_cmd;
				}
				#endif // ( NIP_DHCP_ENABLE == 1 )

				#if ( NIP_AUTOIP_ENABLE == 1 )
				if ( nip_disp.if_flags[ NIP_CURR_IF ] & NIP_DISP_IF_AUTOIP )
				{
					nip_disp.if_flags[ NIP_CURR_IF ] &= ~NIP_DISP_IF_AUTOIP;
					NIP_CURR_CMD = &nip_autoip_disp_check;
					// set check_if flag to make sure we didn't miss any event
					nip_disp_notify( NIP_DISP_CHECK_IF );
					goto eval_cmd;
				}
				#endif

				#if ( NIP_IGMP_ENABLE == 1 )
				if ( nip_disp.if_flags[ NIP_CURR_IF ] & NIP_DISP_IF_GROUP )
				{
					nip_disp.if_flags[ NIP_CURR_IF ] &= ~NIP_DISP_IF_GROUP;
					NIP_CURR_CMD = &nip_ip_disp_check_group;
					// set check_if flag to make sure we didn't miss any event
					nip_disp_notify( NIP_DISP_CHECK_IF );
					goto eval_cmd;
				}
				#endif

			#if NIP_NET_IF_COUNT>1
			} while ( ++NIP_CURR_IF < NIP_NET_IF_COUNT );
			#endif
		}

		if ( nip_disp.flags & NIP_DISP_TIMER )
		{
			nip_disp.flags &= ~NIP_DISP_TIMER;
			nip_disp.flags |=
					NIP_DISP_CHECK_TRANS
					#if NIP_TCP_ENABLE == 1
						| NIP_DISP_CHECK_TCP
					#endif
					#if NIP_MDNS_ENABLE == 1
						| NIP_DISP_CHECK_MDNS
					#endif
					;
		}

		// Transmission events
		// check for IP address resolution flag
		if ( nip_disp.flags & NIP_DISP_CHECK_TRANS )
		{

			nip_disp.flags &= ~NIP_DISP_CHECK_TRANS;
			// find transmissions with status NIP_NET_IF_RESOLV_ADDR
#if NIP_ARP_ENABLE == 1
			NIP_CURR_CMD = &nip_arp_disp_resolve_any;
#endif
			goto eval_cmd;
		}

		// mDNS events
#if NIP_MDNS_ENABLE == 1
		if ( nip_disp.flags & NIP_DISP_CHECK_MDNS )
		{

			nip_disp.flags &= ~NIP_DISP_CHECK_MDNS;
			// find transmissions with status NIP_NET_IF_RESOLV_ADDR
			NIP_CURR_CMD = &nip_mdns_disp_check;
			goto eval_cmd;
		}
#endif

#if NIP_TCP_ENABLE == 1
		// TCP events
		// check for IP address resolution flag
		if ( nip_disp.flags & NIP_DISP_CHECK_TCP )
		{

			nip_disp.flags &= ~NIP_DISP_CHECK_TCP;
			// find transmissions with status NIP_NET_IF_RESOLV_ADDR
			NIP_CURR_CMD = &nip_tcp_disp_check;
			goto eval_cmd;
		}
#endif /* NIP_TCP_ENABLE */

		// Handle UDP Events if UDP has been enabled
		/// @todo Check if using an UDP state machine makes more sense than the
		/// current, direct approach of nip_udp_transmit(). Check nip_udp_disp_send()
		/// for further details.
// 		#if ( NIP_UDP_ENABLE == 1 )
// 		if ( nip_disp.flags & NIP_DISP_CHECK_UDP )
// 		{
// 			nip_disp.flags &= ~NIP_DISP_CHECK_UDP;
// 			NIP_CURR_CMD = &nip_udp_disp_send;
// 			goto eval_cmd;
// 		}
// 		#endif

		// Execute Dispatcher Command Chain. A command is a pointer to a function
		// which may relocate that pointer during it's runtime. The dispatcher
		// will execute the new function after the previous function returned which
		// will possibly lead to chain of functions to be executed. In a non-
		// multitasking environment it is necessary that any chain of functions
		// implements defined exit points which will allow continuation of the
		// chain at a later, event-driven (e.g. timer) point of time. Any function
		// chain must ensure that no deadlock situation can arise from its
		// execution.
		eval_cmd:
		while ( NIP_CURR_CMD != NULL )
		{
			tmp_cmd = NIP_CURR_CMD;
			NIP_CURR_CMD = NULL;
			tmp_cmd();

		}


	};

	///\todo Take result value from nip_disp - Structure
	nip_disp.flags &= ~NIP_DISP_RUNNING;
	return NIP_E_OK;
}
#endif /*(NIP_DISP_ENABLE==1)*/
