/*##############################################################################

nIP - nano IP stack

File        : mem.c

Description : NIP dynamic memory management unit

Copyright notice:

Copyright (C) 2007 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##############################################################################*/

#include "mem.h"
#include "nip_error.h"
#include "os_core.h"

#if (NIP_MEM_ENABLE == 1)

///@todo debug-block.. remove when not needed
nip_mem_block_t *debug_block;

/** pool of dynamically allocatable memory */
static uint8_t nip_mem_pool[ NIP_MEM_DYNAMIC_SIZE ];

/** Initialize dynamic memory pool. All current data in the pool will be lost.
 */
void nip_mem_init()
{
	nip_mem_block_t *current = (nip_mem_block_t *)nip_mem_pool;

	current->id          = 1;
	current->res_length  = NIP_MEM_DYNAMIC_SIZE;
	current->min_length  = 0;
	current->used_length = 0;

	// initial memory block will be set to unused, that is, its memory
	// is available for allocation.
	current->flags       = NIP_MEM_FLGS_UNUSED;

}

/** free bytes from the memory block at the given address */
void nip_mem_free_bytes_ex( nip_mem_block_t *block, nip_mem_size_t size )
{
	/// \todo move all block-shrink-code here

}

/** find memory address of block */
static nip_mem_block_t *nip_mem_locate_block( nip_mem_handle_t block_id )
{
	nip_mem_block_t *block;

	// find block in pool
	for ( block = (nip_mem_block_t*) nip_mem_pool;
		  (uint8_t*)block < &nip_mem_pool[NIP_MEM_DYNAMIC_SIZE];
		  block = (nip_mem_block_t*)&((uint8_t*)block)[block->res_length])
	{
		if ( block->id == block_id )
		{
			return block;
		}
	}

	return NULL;
}

/** Free entire memory block.*/
void nip_mem_free( nip_mem_handle_t block_id )
{
	nip_mem_block_t *block;

	// find block in pool
	block = nip_mem_locate_block( block_id );
	if ( block != NULL )
	{
		// found block; reset block flags
		block->flags       = NIP_MEM_FLGS_UNUSED;
		block->used_length = 0;
		block->min_length  = 0;
		///\todo possibly defragment if previous or next block is also unused
		nip_error = NIP_E_OK;
		return;
	}
	nip_error = NIP_E_INVALID_BLOCK;
}

/** Make two blocks from one. First block will be resized to `'new size'' bytes
 * including header, the second block will receive remaining memory. The newly
 * created block will receive the given `'next_id''.
 * @warning: This function does _NOT_ do an sanity check. Make sure the given
 * parameters will work for the given block, before calling this function!!
 */
static nip_mem_block_t *nip_mem_split( nip_mem_block_t *block, nip_mem_size_t new_size, nip_mem_handle_t next_id )
{
	nip_mem_size_t next_size = block->res_length - new_size;

	// shrink block
	block->res_length = new_size;

	// create new unused block
	block = (nip_mem_block_t*)&(((uint8_t *)block)[block->res_length]);
	block->id          = next_id;
	block->res_length  = next_size;
	block->min_length  = 0;
	block->used_length = 0;
	block->flags = NIP_MEM_FLGS_UNUSED;

#if NIP_MEM_DEBUG == 1
	if ( block->res_length == 0 )
	{
		// DEBUG something went wrong. CHECK THIS!!!!!
		return NIP_MEM_NULL;
	}
#endif
	return block;
}
/** Search for unused memory in pool and collect at a specific position.
 *
 * The freed memory may be appended to an existing block (new_block_id ==
 * free_at_block_id) or a new block with the id new_block_id will be created.
 *
 * @todo There are many possible strategies to aquire needed memory. The current
 * approach is to take the required memory equally from all blocks. A better
 * strategy may be to take memory from blocks according to a which-block-would-
 * care-least algorithm.
 *
 * @param needed_size      number of bytes to look for
 * @param free_at_block_id block after which to position the freed memory
 * @param new_block_id     id of new block to create or the same value as
 *   free_at_block_id if freed memory is to be appended to the block specified
 *   in free_at_block_id.
 *
 * @return Pointer to modified memory block. On Error NULL will be returned
 * and an error code will be written to the global nip_error variable. Possible
 * errors are
 *  - NIP_E_OUT_OF_MEMORY if not enough memory could be freed, or
 *  - NIP_E_PARAM_FAILED if free_at_block_id==new_block_id but block was locked
 *    and can therefor not be extended.
 */
nip_mem_block_t *nip_mem_defrag( nip_mem_size_t needed_size,
	nip_mem_handle_t free_at_block_id, nip_mem_handle_t new_block_id )
{
	nip_mem_size_t     request_memory;    // memory needed to be freed
	nip_mem_size_t     free_available;    // amount of free available memory
	nip_mem_size_t     smallest_available;// smallest amount of free mem
	nip_mem_size_t     total_available;   // total free memory available
	nip_mem_size_t     freed_memory = 0;  // total memory freed so far
	nip_mem_size_t     unused_memory;// one block's reserved but unused memory
	nip_mem_size_t     mem_free = 0; // how much memory to free per block
	nip_mem_size_t     tmp_free;     // how much memory to free for a certain block
	nip_mem_size_t     mem_count;    // counter for mem move operation
	nip_mem_handle_t   block_count;
	nip_mem_block_t   *block;        // current block being worked with
	nip_mem_block_t   *first_block;  // first block to free memory at
	uint8_t           *insert_pos;   // position to insert memory at
	uint8_t            scan_pos;     /* 0x01: before block to add memory to,
	                                    0x02: at block to add memory to
	                                    0x03: after block to add memory to
	                                    0x04: at end of memory */
	uint8_t            free_unused_block; // unused block is to be freed
	uint8_t            i = 0;

	request_memory = needed_size;
	// add size of block-header to request_memory if necesary
	if ( new_block_id != free_at_block_id )
		request_memory += sizeof( nip_mem_block_t );

	// First run starts at beginning of memory pool.
	first_block = NULL;

	// position to free memory at yet unknown
	if ( free_at_block_id == NIP_MEM_NULL )
		insert_pos = nip_mem_pool;
	else
		insert_pos = NULL;

	do /* outer loop */
	{
		scan_pos = 1;
		// reset scanner position
		if ( free_at_block_id == NIP_MEM_NULL )
		{
			scan_pos = 3;
			block    = (nip_mem_block_t *)(nip_mem_pool + freed_memory);
		}
		if ( first_block == NULL || request_memory <= freed_memory )
		{
			// No start position has been set or enough memory has been freed.
			// in the latter case we start at the beginning, as the scanner
			// will just forward to the block where the freed memory has to be
			// appended or a new block has to be formed.
			block = (nip_mem_block_t *)nip_mem_pool;
		}
		else
		{
			block    = first_block;
			if ( (uint8_t*)block >= insert_pos )
				scan_pos = 3;
		}

		// initialize values for memory analysis
		block_count        = 0 ;
		smallest_available = 0;
		total_available    = 0;
		free_available     = 0;

		// reset start block for next run
		first_block = NULL;

		// start scanner
		do /* inner loop (scanner) */
		{

#if NIP_MEM_DEBUG == 1
			if ( (uint8_t *)block > &nip_mem_pool[ NIP_MEM_DYNAMIC_SIZE ]
			  || block == NULL
			  || block->res_length < 8 )
			{
				// DEBUG: sanity check of block pointer
				return NULL;
			}
#endif

			// reset variables
			free_unused_block = 0;

			// check if block is locked
			if ( NIP_MEM_BLOCK_LOCKED( block ) )
			{
				//=== block is locked ===//

				switch ( scan_pos )
				{
					case 1: // scanner is positioned before the block to add mem to.
						// Locked block restarts scanner at that position because
						// previously found memory cannot be moved to the requested
						// position.
						first_block = NULL;
						total_available    = 0;
						smallest_available = 0;
						free_available     = 0;
						break;

					case 3: // scanner is positioned after the block to add mem to.
						// Locked block stops scanner at current position because any
						// free memory beyond that block cannot be moved to the
						// requested position.
						scan_pos = 4;
						break;

					default:
						break;
				}

				// The block to add memory to must not be locked
				if ( block->id == free_at_block_id )
				{
					nip_error = NIP_E_PARAM_FAILED; // cancel function.
					return NULL;
				}

			} else
			{
				//=== block is not locked ===//

				// The unused_memory variable will only be used to signal
				// unused memory of blocks that are in use.
				unused_memory = 0;
				tmp_free = 0;

				// unused blocks may be freed entirely, if it needs to.
				if ( !NIP_MEM_BLOCK_USED( block )
				  && request_memory > freed_memory
				  && block->id != new_block_id
// 				  && (block->res_length < (request_memory-freed_memory+sizeof(nip_mem_block_t)))
					)
				{

/*					// If the free_at_block_id-block is unused, it can be used
					// as new block. It will therefor be given the new_block_id
					// and its size will be deducted from the requested memory.
					if ( block->id == free_at_block_id )
					{
						block->id = new_block_id;
						free_at_block_id = new_block_id;
						if ( request_memory > block->res_length - sizeof( nip_mem_block_t ) )
							request_memory -= block->res_length;
						else
							request_memory = 0;
					}
					// If the current block is unused and the insert position for
					// freed memory has been determined already, the entire block
					// will be freed and moved to that position.
					else*/ if ( i==1  )
					{
						tmp_free = request_memory - freed_memory;
						/** @todo Free Blocks NEED NOT be splitted. If only parts of a
						 * block need to be freed, that should be done  using the mechanism
						 * that's being used for any other non-free block with unused
						 * memory (see further down this file). For further discussion
						 * see Bug #106. */

						/** The following line is a workaround to Bug #106 */
						if ( tmp_free < sizeof( nip_mem_block_t) )
							tmp_free = sizeof( nip_mem_block_t );

						// block larger than needed amount of memory?
						if ( block->res_length > (tmp_free + sizeof( nip_mem_block_t )) )
						{

							// Split block, so only the needed amount of memory will
							// be moved. The remaining memory will form a new block,
							// given the ID of the current block, which will eventually
							// be (re)moved.
							nip_mem_split( block, tmp_free, block->id );

						}
						else
							tmp_free = block->res_length;

						free_unused_block = 1;
					}
					// If the current block is unused but no insert position for
					// freed memory has yet been determined, add the block's size
					// to the free_available memory counter for later use.
					else
					{
						free_available += block->res_length;
					}

					// if the previous block has been blocked use this one as
					// starting point.
					/// @todo  Should this line be moved down to apply to all
					/// unused blocks? I would think so! (jon)
					if ( first_block == NULL )
						first_block = block;

				} else
				if ( block->id != new_block_id )
				{
				/*DEBUG*/if (!NIP_MEM_BLOCK_USED( block ))
				/*DEBUG*/free_unused_block = 0;
					// check how much memory is unused in this block. Mem_free bytes
					// will be taken from that block. It's unused_memory will therefor
					// be reduced later.
					unused_memory = block->res_length - sizeof( nip_mem_block_t );
					if ( block->min_length < block->used_length)
						unused_memory -= block->used_length;
					else
						unused_memory -= block->min_length;

					// make sure we don't free more memory than is available
					if ( mem_free > unused_memory )
						tmp_free = unused_memory;
					else
						tmp_free = mem_free;
				}
				else
				{
					// don't free memory from the block we're collecting memory for
// 					unused_memory = 0;
				}

				// Free memory from block. Tmp_free will be 0 for the first run
				// of the scanner, so no memory will be freed during the first
				// run, as it's just used for memory analysis. (see below)
				/// @todo check if the following commented unused_memory condition
				/// makes any sense. It's been there and probably been cause of a
				/// bug but I'm not sure why I ever put it there, so there may be
				/// some sense to it.
 				if ( tmp_free > 0 /*&& unused_memory > 0*/ )
				{
					block->res_length -= tmp_free;
					freed_memory      += tmp_free;
					if ( unused_memory > 0 )
						unused_memory     -= tmp_free;
					switch ( scan_pos )
					{
						case 1:
							// Move-left following blocks by tmp_free-Bytes
							mem_count = tmp_free + block->res_length;

							for (
								;
								&((uint8_t*)block)[mem_count] < insert_pos;
								mem_count++ )
							{
								((uint8_t*)block)[mem_count-tmp_free] = ((uint8_t*)block)[mem_count];
							}

							break;
						case 3:
							// Move-right current and previous blocks by tmp_free-Bytes
							mem_count = (uint8_t*)block - insert_pos + block->res_length;
							while ( mem_count > 0 )
							{
								mem_count--;
								((uint8_t*)insert_pos)[mem_count+tmp_free] = ((uint8_t*)insert_pos)[mem_count];
							}

							// Advance the block pointer by the amount of freed memory.
							block = ( nip_mem_block_t *) ((uint8_t*) block + tmp_free);
							break;
					}
				}

				// Analyse how much this block may still be shrinked.
				total_available += unused_memory;

				if ( unused_memory > 0 )
				{
					// if the previous block was locked, mark this as next block to
					// possibly free memory from.
					if ( first_block == NULL )
						first_block = block;

					block_count++;

					if ( unused_memory < smallest_available || smallest_available == 0)
						smallest_available = unused_memory;

				}
			}

			//=== the following will be done for all blocks, locked or not ===//

			// is this the block to free memory at? Set insert position for
			// freed memory.
			if ( block->id == free_at_block_id )
			{
				insert_pos = &(( uint8_t* )block)[block->res_length];
			}

			// Check if all requested memory has been collected. Continue to run
			// up to free_at_block_id-Block because the block pointer will be
			// needed later.
			if ( request_memory <= freed_memory && block->id == free_at_block_id )
				scan_pos = 2;
			else
			if ( free_unused_block )
			{
				// current block has been selected by free_unused_block already
				// TODO check if all is done
			}
			else
			{
				// move to next block
				block = (nip_mem_block_t*)(((uint8_t*)block)+ block->res_length);
				if ( block == ( nip_mem_block_t *) insert_pos )
				{
					// skip freed memory
					block = ( nip_mem_block_t* )( insert_pos + freed_memory );
					scan_pos = 3;
				}

				#if NIP_MEM_DEBUG == 1
					if ( (uint8_t *)block > &nip_mem_pool[ NIP_MEM_DYNAMIC_SIZE ]
					|| block == NULL
					|| ((uint8_t *)block < &nip_mem_pool[ NIP_MEM_DYNAMIC_SIZE ] && block->res_length < 8) )
					{
						// DEBUG: sanity check of block pointer
						return NULL;
					}
				#endif
			}

			// End of memory pool reached?
			if ( (uint8_t*)block == &nip_mem_pool[ NIP_MEM_DYNAMIC_SIZE ] )
				scan_pos = 4;

			// reduce mem_free, if its not all needed
			if ( request_memory < freed_memory + mem_free )
				mem_free = request_memory - freed_memory;

			// if free memory is available skip the usage of non-free but unused memory
			if ( free_available > 0 || request_memory <= freed_memory )
				mem_free = 0;

		} while ( scan_pos != 2 && scan_pos != 4 ); /* inner loop (scanner) */

		// Outer loop control:
		// * After first run of scanner we know how much memory can be be freed
		// * Subsequent runs of scanner free parts of available memory
		// * Quit if first run did not end up with enough free memory or
		// * Quit if enough memory has been freed. Create new block with
		//   new_block_id if necessary.
		// * Continue if more memory can and must be freed. Determine how much
		//   memory should be freed per block ( mem_free variable )

		if ( request_memory > total_available + free_available + freed_memory )
		{
			nip_error = NIP_E_OUT_OF_MEMORY;
			return NULL;
		}

		// enough memory has been freed and the scanner is at correct position?
		if ( request_memory <= freed_memory && scan_pos == 2 )
		{
			// create new block?
			if ( new_block_id != free_at_block_id )
			{
				block = ( nip_mem_block_t *) insert_pos;
				block->id = new_block_id;
				block->res_length = freed_memory;
				block->flags = NIP_MEM_FLGS_UNUSED;
				// default setting for minimum length is reserved length.
				block->min_length = freed_memory;
				block->used_length = 0;
			}
			else
			{
				block->res_length += freed_memory;
			}
#if NIP_MEM_DEBUG == 1
			/* DEBUG */ debug_block = (nip_mem_block_t*)&((uint8_t*)block)[block->res_length];
#endif
			nip_error = NIP_E_OK;
			return block;
		}
		// more memory needs to be freed?
		else if ( block_count > 0 && request_memory > freed_memory)
		{
			mem_free = ( request_memory - freed_memory ) / block_count;

			// to ensure fair distribution of free memory, take same amounts of
			// memory from all blocks.
			if ( mem_free > smallest_available )
				mem_free = smallest_available;

			if ( request_memory - freed_memory == 1 )
				mem_free = 1;
		}

		i = 1;
	} while (1); /* outer loop */

}

/** Set new minimum block length and resize block if necessary.
 * @return NIP_E_OK, NIP_E_INVALID_BLOCK, or NIP_E_OUT_OF_MEMORY. */
nip_error_t nip_mem_set_min_length( nip_mem_handle_t block_id, nip_mem_size_t min_length )
{
	nip_mem_block_t *block;
	nip_mem_size_t  block_user_size;

	// find block in pool
	block = nip_mem_locate_block( block_id );
	if ( block != NULL )
	{
		// if min_length has to be reduced, well... reduce it. :)
		// if min_length has to be increased, but does not exceed the reserved
		// memory.. no problem.

		// if block has to be increased, try to find memory for it.
		block_user_size = block->res_length - sizeof( nip_mem_block_t );
		if ( min_length > block_user_size )
		{
			block = nip_mem_defrag( min_length - block_user_size, block->id, block->id );
			if ( block == NULL )
				return nip_error;
		}
		block->min_length = min_length;
	} else
		return NIP_E_INVALID_BLOCK;

	return NIP_E_OK;

}

/** Allocate dynamic memory block. The amount of allocated memory may change
 * during a block's lifetime, but will always be at least min_length.
 *
 * @param size       number of bytes to initially allocate, if available.
 * @param min_length minimum number of bytes to always keep reserved for this
 *                   block
 * @param flags      memory block flags. NIP_MEM_FLG_USED will always be set.
 * @param res_length if not NULL, will be filled with number of bytes that have
 *                   actually been allocated for usage.
 *
 * @return memory block handle or NIP_MEM_NULL if min_length bytes of memory are
 *         not available for allocation.
 */

nip_mem_handle_t nip_mem_alloc( nip_mem_size_t size, nip_mem_size_t min_length, nip_mem_flags_t flags, nip_mem_size_t *res_length)
{
	nip_mem_block_t     *block;
	nip_mem_block_t     *next_block;
	nip_mem_block_t     *new_block;
	nip_mem_size_t       new_size;
	nip_mem_size_t       block_user_size;
	uint8_t              i;
	nip_mem_handle_t     new_block_id;
	nip_mem_handle_t     at_block_id;
	nip_error_t          result = NIP_E_OUT_OF_MEMORY;

// TODO check lock or lock memory management

	// find free memory
	// perform a maximum of 4 outer loops:
	// 0. look for unused blocks of wanted length
	// 1. look for unused blocks6 of minimum length
	// 2. look for any unused blocks
	// 3. look for any spot with an unused block-id
	for ( i = 0; i < 4; i++ ) /* outer loop */
	{
		at_block_id = 0;

		for ( block = ( nip_mem_block_t *) nip_mem_pool; block != NULL; block = next_block)
		{
			if ( block->id > NIP_MEM_NULL )
				at_block_id  = block->id;

#if NIP_MEM_DEBUG == 1
			if ( block->res_length == 0 )
			{
				// DEBUG something went wrong. CHECK THIS!!!!!
				return NIP_MEM_NULL;
			}
#endif
			calc_user_size:
			block_user_size = block->res_length - sizeof( nip_mem_block_t );

			// determine next memory block
			next_block = (nip_mem_block_t*)&(((uint8_t*)block)[block->res_length]);
			if ( (uint8_t*)next_block == &nip_mem_pool[ NIP_MEM_DYNAMIC_SIZE ] )
				next_block = NULL;
#if NIP_MEM_DEBUG == 1
			if ( next_block != NULL && ( next_block->res_length == 0
			   || &((uint8_t*)next_block)[next_block->res_length]>&nip_mem_pool[ NIP_MEM_DYNAMIC_SIZE ]))
			{
				// DEBUG something went wrong. CHECK THIS!!!!!
				return NIP_MEM_NULL;
			}
#endif
			// merge neighboring empty blocks
			if ( !NIP_MEM_BLOCK_USED( block ) && next_block != NULL && !NIP_MEM_BLOCK_USED( next_block ))
			{
				block->res_length += next_block->res_length;
				goto calc_user_size;
			}

			if ( i==0 && !NIP_MEM_BLOCK_USED( block ) && block_user_size >= size )
			{
				// Found free block. Check to reduce block size. Reducing block size
				// requires a second block to be created which will hold the remaining
				// memory. First of all check if there is space for a new block header.
				if ( block_user_size >= size + sizeof( nip_mem_block_t ) )
				{

					// Check if there is a free id available for new block.
					if (( next_block != NULL && (next_block->id > block->id+1 ))
					||( next_block == NULL && block->id+1 < NIP_MEM_MAX_BLOCKS ))
					{
						next_block = nip_mem_split( block, size + sizeof( nip_mem_block_t ), block->id+1 );
#if NIP_MEM_DEBUG == 1
						if ( block->res_length == 0 || next_block->res_length == 0
						|| &((uint8_t*)next_block)[next_block->res_length]>&nip_mem_pool[ NIP_MEM_DYNAMIC_SIZE ])
						{
							// DEBUG something went wrong. CHECK THIS!!!!!
							return NIP_MEM_NULL;
						}
#endif

					}
					// else large block of memory than necessary will be reserved
					// because no free block-id is available to form a new unused
					// block.
					/** \todo Free memory could be moved to a position with available
					 * block-id to form a new unused block. */
				}

				// Memory has been allocated -> positive result.
				result = NIP_E_OK;

				i = 10; /* exit outer loop */
				break;
			}
			else
			if (( i==1 && !NIP_MEM_BLOCK_USED( block ) && block->id > NIP_MEM_NULL && block_user_size >= min_length && !NIP_MEM_BLOCK_LOCKED( block ) )
			 || ( i==2 && !NIP_MEM_BLOCK_USED( block ) && block->id > NIP_MEM_NULL && !NIP_MEM_BLOCK_LOCKED( block ))
			 || ( i==3 && next_block != NULL && at_block_id+1 < next_block->id)
			 || ( i==3 && next_block == NULL && at_block_id+1 < NIP_MEM_MAX_BLOCKS ))
			{
				if ( i>2 && block == ( nip_mem_block_t *) nip_mem_pool && block->id > 1)
				{
					// create new first block in pool
					new_size     = size;
					new_block_id = 1;
					at_block_id  = 0;
				}
				if ( i>2 )
				{
					// create new block after current block
					new_size     = size;
					new_block_id = block->id+1;
				}
				else
				{
					// fill up current, unused block
					new_size     = size - block_user_size;
					new_block_id = block->id;
				}

				new_block = nip_mem_defrag( new_size, block->id, new_block_id);

				// if not enough memory for wanted size, try minimum size
				if ( new_block == NULL && nip_error == NIP_E_OUT_OF_MEMORY )
				{
					new_size  -= size - min_length;
					new_block  = nip_mem_defrag( new_size, block->id, new_block_id);
				}

				if ( new_block != NULL )
				{
					block = new_block;
					// init block
					result = NIP_E_OK;
					i = 10; /* exit outer loop */
					break;
				}
			}
		}
	}

	nip_error = result;

	if ( result == NIP_E_OK )
	{
		block->min_length = min_length;
		block->flags = flags | NIP_MEM_FLG_USED;
		if ( res_length != NULL )
			*res_length = block->res_length - sizeof( nip_mem_block_t );
		return block->id;
	}

	// else
	return NIP_MEM_NULL;

}

/** Write data to memory block.
 *
 * The data will be appended after the last byte of data in the block. If the
 * block is too small to hold the data, attempts will be made to allocate the
 * needed memory. If that fails NIP_E_OUT_OF_MEMORY will be returned.
 *
 * @param block_id  id of block to write data to
 * @param buf       address of buffer to copy data from
 * @param size      number of bytes to copy from buffer
 *
 * If buf is NULL the block will still be increased by size bytes but initialized
 * with Zeros. That special case is usually being used to allocate memory to
 * be accessed directly, later. (see nip_mem_obtain_ptr() )
 *
 * @return NIP_E_OK or NIP_E_OUT_OF_MEMORY or NIP_E_INVALID_BLOCK
 */
nip_error_t nip_mem_write( nip_mem_handle_t block_id, void *buf, nip_mem_size_t size )
{
	nip_mem_block_t *block;
	nip_mem_size_t  unused_bytes;

	// find block in pool
	block = nip_mem_locate_block( block_id );
	if ( block != NULL )
	{
		unused_bytes = block->res_length - block->used_length - sizeof( nip_mem_block_t );
		// Make sure enough memory is available. If not, try to allocate.
		if ( size > unused_bytes )
		{
			if ( (block = nip_mem_defrag( size - unused_bytes, block_id, block_id )) == NULL )
			{
				return NIP_E_OUT_OF_MEMORY;
			}
		}

		if ( buf != NULL )
		{
			// Copy data from buffer to block
			nip_memcpy(
				&((uint8_t*)block)[ sizeof( nip_mem_block_t ) + block->used_length ],
				buf,
				size);
		}
		else
		{
			// no source buffer given -> initialize with 0
			nip_memset(
				&((uint8_t*)block)[ sizeof( nip_mem_block_t ) + block->used_length ],
				0,
				size);
		}
		block->used_length += size;
		return NIP_E_OK;
	}
	return NIP_E_INVALID_BLOCK;
}

/** Cut len bytes from the given location. The location has to be a managed block.
 */
nip_error_t nip_mem_cut( nip_mem_ptr_t *location, nip_mem_size_t len )
{
	uint8_t *l;
	nip_mem_size_t bufsize = nip_mem_buf_used( location->id);

	if ( len + (nip_mem_size_t)location->ptr > bufsize )
		return NIP_E_PARAM_FAILED;

	l = nip_mem_ptr( location );
	if ( l == NULL ) return nip_error;

	nip_memmove( l, l+len, bufsize-len-(nip_mem_size_t)location->ptr );

	nip_mem_release_block( location->id );

	// resize block to new length
	nip_mem_set_used( location->id, bufsize - len );

	return NIP_E_OK;
}

/** Allocate memory at dst and copy data from src to dst. No data will be
 * overwritten at dst. All current data at dst will be moved by len bytes. After
 * successfully inserting the data, the dst-pointer will be placed at the end of
 * the inserted data, so it can be directly used for follow-up insertions.
 *
 * @param dst has to point to a managed memory block. If the offset component of
 * dst points beyond the size of the dst-block or is -1, the new data will be
 * Appended to the dst-block. So to simply append data to a block you can use
 * (nip_mem_size_t)-1 as offset.
 * @param src may either point to a managed memory block or some other buffer.
 * @param len amount of bytes to be copied from src to dst.
 *
 * @return NIP_E_OK on success or
 *  - NIP_E_OUT_OF_MEMORY if not enough memory could be allocated at dst,
 *  - NIP_E_BLOCK_LOCKED  if dst or source block were locked
 *  - NIP_E_INVALID_BLOCK or
 *  - NIP_E_PARAM_FAILED  if src or dst pointed to invalid address/block
 */
nip_error_t nip_mem_insert( nip_mem_ptr_t *dst, nip_mem_ptr_t *src, nip_mem_size_t len )
{
	nip_error_t    res;
	nip_mem_size_t dst_size;
	uint8_t        *dst_buf, *src_buf;

	if ( dst == NULL || src == NULL )
		return NIP_E_PARAM_FAILED;

	dst_size = nip_mem_buf_used( dst->id );

	if ( (nip_mem_size_t)dst->ptr > dst_size )
		dst->ptr = (void*)dst_size;

	// allocate len-Bytes at dst
	res = nip_mem_write( dst->id, NULL, len );
	if ( res != NIP_E_OK )
		goto quit;

	// get pointers to source and destination buffers
	src_buf = nip_mem_ptr( src );

	if ( ( dst_buf = nip_mem_ptr( dst ) ) == NULL )
	{
		res = nip_error;
		goto release_src;
	}

	// move data at dst by len
	nip_memmove( dst_buf+len, dst_buf, dst_size - (nip_mem_size_t)dst->ptr );

	// copy data from source to dst, or if src_buf is NULL fill dst_buf with 0
	if ( src_buf != NULL )
		nip_memcpy( dst_buf, src_buf, len );
	else
		nip_memset( dst_buf, 0x00, len );

	// move dst pointer
	dst->ptr += len;

	nip_mem_release_ptr( dst );

	release_src:
		nip_mem_release_ptr( src );

	quit:
		return res;
}

/** Read data from block at a specific position.
 *
 * The data will be copied from a specific position in the block into the given
 * buffer. The data will not be removed from the block afterwards. If buf points
 * to NULL nothing will be copied, but the number of bytes that would've been
 * copied if buf had not been NULL will be returned.
 *
 * @return number of bytes actually copied or 0 on error. Possible errors are
 *  - NIP_E_EOF given position did not contain any data
 *  - NIP_E_INVALID_BLOCK given block_id was unknown
 */
nip_mem_size_t nip_mem_read_at_pos( nip_mem_handle_t block_id, uint8_t *buf,
                                    nip_mem_size_t size, nip_mem_size_t pos)
{
	nip_mem_block_t *block;
	nip_mem_size_t data_length;

	// find block in pool
	block = nip_mem_locate_block( block_id );
	if ( block == NULL )
	{
		nip_error = NIP_E_INVALID_BLOCK;
		return 0;
	}

	data_length = block->used_length;
	if ( pos >= data_length )
	{
		nip_error = NIP_E_EOF;
		return 0;
	}

	// check available data size
	if ( pos + size > data_length )
		size = data_length - pos;

	// copy data to buffer
	if ( buf != NULL )
		nip_memmove( buf, &((uint8_t*)block)[ sizeof( nip_mem_block_t ) + pos ], size );

	return size;
}


/** Read data from block.
 *
 * The data will be copied from the block into the given buffer. The read data
 * will be removed from the block. This operation will usually be used in
 * network buffers where data is just being collected and read once.
 *
 * If buf is NULL the requested amount of bytes will be read/deleted from the
 * buffer, but not copied anywhere.
 *
 * @param block_id  id of block to read data from
 * @param buf       pointer to buffer to copy data to
 * @param size      number of bytes to be copied
 *
 * @return number of bytes actually copied or 0 on error. Possible errors are
 *  - NIP_E_EOF given position did not contain any data
 *  - NIP_E_INVALID_BLOCK given block_id was unknown
 */
nip_mem_size_t nip_mem_read( nip_mem_handle_t block_id, uint8_t *buf, nip_mem_size_t size )
{
	nip_mem_block_t *block;
	nip_mem_size_t bytes_copied;

	// find block in pool
	block = nip_mem_locate_block( block_id );
	if ( block == NULL )
	{
		nip_error = NIP_E_INVALID_BLOCK;
		return 0;
	}

	bytes_copied = nip_mem_read_at_pos( block_id, buf, size, 0 );

	if ( bytes_copied > 0 )
	{
		// remove bytes from block
		block->used_length -= bytes_copied;
		nip_memmove(
			&((uint8_t*)block)[ sizeof( nip_mem_block_t ) ],
			&((uint8_t*)block)[ sizeof( nip_mem_block_t ) + bytes_copied],
			block->used_length
				);
	}

	return bytes_copied;
}

/** Move memory inbetween blocks by reading from the beginning of one block and
 * appending to the end of another block. Block sizes will be adjusted, if
 * necessary.
 *
 * In case not enough memory was available for the move operation,
 * NIP_E_OUT_OF_MEMORY will be written to nip_error and the number of bytes
 * which have successfully been moved be returned.
 *
 * @return number of bytes moved.
 */
nip_mem_size_t nip_mem_move( nip_mem_handle_t dst_block, nip_mem_handle_t src_block, nip_mem_size_t size )
{
	nip_mem_size_t  moved_size     = 0;
	nip_mem_size_t  available_size = size;
	nip_mem_block_t *block;

	do
	{
		if ( available_size > size )
			available_size = size;

		// attempt to add available_size bytes to destination block
		nip_error = nip_mem_write( dst_block, NULL, available_size );

		if ( nip_error == NIP_E_OUT_OF_MEMORY )
		{
			if ( available_size == 1 )
			{
				// no chance... tried smallest possible portion.
				break;
			}

			// try to allocate smaller size.
			available_size /= 2;
		}
		else
		if ( nip_error == NIP_E_OK )
		{
			// get pointer to destination memory block
			block = nip_mem_locate_block( dst_block );
			if ( block == NULL ) break;

			// append data to block
			nip_mem_read( src_block, &((uint8_t*)block)[sizeof(nip_mem_block_t)+block->used_length-available_size], available_size );

			moved_size += available_size;
			size       -= available_size;
		}
		else
		{
			// whatever happened.. quit.
			break;
		}
	} while ( size > 0 );

	return moved_size;
}

/** Obtain pointer on memory block data to perform direct operations on memory.
 *
 * The pointer is usually used for direct access to memory, such as using
 * structures. For simply reading from or writing to the block the functions
 * nip_mem_read() and nip_mem_write() SHOULD be used.
 *
 * The block will be locked for the time the pointer is handed out and will
 * therefor not be available for dynamic scaling or defragmentation, which will
 * in turn affect the entire memory management. MAKE SURE to obtain pointers for
 * as short a period as possible, to not impose those negative on the management
 * system any longer than needed.
 *
 * Use the nip_mem_release_block() function to unlock the block. The given pointer
 * MAY NOT be used after releasing it.
 *
 * @return Pointer to the block payload or NULL if the block_id was invalid or
 * the block has been locked by another process already. The nip_error variable
 * will be set to NIP_E_BLOCK_LOCKED, NIP_E_INVALID_BLOCK if NULL is returned.
 */
void *nip_mem_obtain_ptr( nip_mem_handle_t block_id )
{
	nip_mem_block_t *block;

	// find block in pool
	block = nip_mem_locate_block( block_id );
	if ( block == NULL || !NIP_MEM_BLOCK_USED( block ) )
	{
		nip_error = NIP_E_INVALID_BLOCK;
		return NULL;
	}

	if ( NIP_MEM_BLOCK_LOCKED( block ) )
	{
		nip_error = NIP_E_BLOCK_LOCKED;
		return NULL;
	}

	block->flags |= NIP_MEM_FLG_LOCKED;

	return ++block;
}

/** Unlocks memory block. Use this function after each nip_mem_obtain_ptr()
 * call.
 */
void nip_mem_release_block( nip_mem_handle_t block_id )
{
	nip_mem_block_t *block;

	// find block in pool
	block = nip_mem_locate_block( block_id );
	if ( block == NULL || !NIP_MEM_BLOCK_LOCKED( block ) )
	{
		return;
	}

	block->flags &= ~NIP_MEM_FLG_LOCKED;
}


/** return pointer to location specified by p. Depending on the id attribute
 * of p the result will point to an offset of the block with that ID, or if
 * id is set to NIP_MEM_NULL, the ptr attribute will be simply returned.
 *
 * @note You should call nip_mem_release_ptr() after you're done working with
 * the pointer, to make sure no ressources will remain blocked indefinitely.
 */
void *nip_mem_ptr( nip_mem_ptr_t *p )
{
	uint8_t *res;

	if ( p == NULL )
		return NULL;

	// no ID given? -> treat as regular pointer
	if ( p->id == NIP_MEM_NULL )
		return p->ptr;

	// check offset, which MUST NOT exceed buffer size
	if ( (nip_mem_size_t)p->ptr >= nip_mem_buf_used( p->id ) )
		return NULL;

	res = nip_mem_obtain_ptr( p->id );
	if ( res != NULL )
		res = res + (nip_mem_size_t)p->ptr;

	return res;
}

/** Counterpart to nip_mem_ptr. Release previously obtained pointer and unlock
 * any ressources that may have been locked because of it.
 */
void nip_mem_release_ptr( nip_mem_ptr_t *p )
{
	if ( p != NULL )
	{
		nip_mem_release_block( p->id );
	}
}


/** Return used buffer size. On error 0 will be returned but since 0 is also a
 * valid size you must check nip_error to ensure an error occured.
 */
nip_mem_size_t nip_mem_buf_used( nip_mem_handle_t block_id )
{
	nip_mem_block_t *block;

	// find block in pool
	block = nip_mem_locate_block( block_id );
	///@todo remove debug-block line whenever not needed anymore. This is just used
	/// to debug the block's content outside the memory management functions.
	//debug_block = block;
	if ( block == NULL )
	{
		nip_error = NIP_E_INVALID_BLOCK;
		return 0;
	}

	return block->used_length;
}

/** Set Buffer used size to specific value.
 *
 * @warning Only do this if you're sure about the used buffer size as it may
 * lead to data loss or corrupted buffers if set to a wrong value.
 */
void nip_mem_set_used( nip_mem_handle_t block_id, nip_mem_size_t new_size )
{
	nip_mem_block_t *block;

	// find block in pool
	block = nip_mem_locate_block( block_id );
	if ( block == NULL )
	{
		nip_error = NIP_E_INVALID_BLOCK;
	}
	else
		block->used_length = new_size;
}

#endif /* NIP_MEM_ENABLE == 1 */
