/*##############################################################################

nIP - nano IP stack

File        : arp.c

Description : Functions to serve ethernet interfaces

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#include "dispatcher.h"
#include "inet.h"
#include "net/arp.h"
#include "net/net_if.h"
#include "net/ip.h"
#include "net/numbers.h"
#include "net/autoip.h"
#include "os_core.h"


#if (NIP_ARP_ENABLE == 1)

extern nip_net_if_t nip_net_if_list[];
extern nip_disp_state_t nip_disp;

/** Array with ARP Cache entries */
static struct ARP_CACHE arp_cache[ NIP_ARP_CACHE_SIZE ];

// uint8_t uint32_gt( uint32_t *i1, uint32_t *i2 )
// {
// 	return ( *i1 > *i2 );
// }

#define uint32_gt( i1, i2 ) *i1>*i2

/** Find ARP cache entry by Protocol address.
 *
 * If pr_addr is NULL or was not found in cache, a pointer to an unused or
 * expired cache entry will be returned in `'entry''. If no unused or expired
 * cache entry is available a pointer to the entry with the oldest timestamp
 * will be returned.
 *
 * @return NIP_ERROR if pr_addr is not NULL and no unexpired entry with that
 * address has been found or if pr_addr is NULL and no unused or expired entry
 * has been found. NIP_SUCCESS, else. In any case `'entry'' will point to a
 * cache entry to be used to store new address information.
 */
static nip_success_t nip_arp_cache_lookup( uint8_t *pr_addr, struct ARP_CACHE **entry )
{
	arp_cache_id_t   cnt;
	nip_success_t    res;
	struct ARP_CACHE *e;

	cnt = 0;
	*entry = &arp_cache[ cnt ];
	do
	{
		e = &arp_cache[ cnt ];

		// remember entry with oldest timestamp
		if ( uint32_gt( &(*entry)->timestamp, &e->timestamp ) ) *entry = e;

		if ( uint32_gt( &e->timestamp, &nip_tickcount() ) )
		{
			res = NIP_SUCCESS;
		}
		else if ( pr_addr == NULL )
		{
			*entry = e;
			return NIP_SUCCESS;
		}
		else
			res = NIP_ERROR;

		// Found entry with matching protocol address. Return ERROR or SUCCESS
		// depending on timestamp.
		if ( pr_addr != NULL && nip_memcmp( pr_addr, e->pr_addr, NIP_ARP_PR_SIZE ) == 0 )
		{
			*entry = e;
			return res;
		}

	} while ( ++cnt < NIP_ARP_CACHE_SIZE );

	// if the searched entry was not found, return ERROR
	return NIP_ERROR;
}

/** Fill default values for ARP packet and transmit.
 *
 * The fields op-code, tpa and tha have to be set prior to calling this function.
 */
static void nip_arp_fill_defaults_and_send( nip_net_if_id_t if_id, struct ARP_PACKET *arp )
{
	nip_net_if_t *net_if    = &nip_net_if_list[ if_id ];
	uint8_t      *ll_target = arp->tha;

	arp->hrd = htons( net_if->phy_conf.type );
	arp->pro = NIP_ARP_PR_TYPE;
	arp->hln = net_if->phy_conf.hw_addr_size;
	arp->pln = NIP_ARP_PR_SIZE;
	nip_memcpy( arp->sha, net_if->phy_conf.hw_addr, net_if->phy_conf.hw_addr_size );
	nip_memcpy( arp->spa, net_if->ip_conf.addr    , NIP_ARP_PR_SIZE );

	// ARP requests are always being broadcasted on the link layer
	if ( arp->op[1] == NIP_ARP_OP_REQUEST
#if NIP_AUTOIP_ENABLE == 1
	// if sender IP address is link local address, always use link-layer
	// broadcast to transmit ARP packets.
	|| ( arp->spa[0] == 169 && arp->spa[1] == 254 )
#endif
	)
	{
		ll_target = net_if->phy_conf.hw_brdcst_addr;
	}

	// send ARP packet
	net_if->phy_conf.hard_send_init(
		if_id,
		NULL,
		NIP_NUM_PR_ARP,
		ll_target,
		arp->sha,
		sizeof( struct ARP_PACKET )
	);

	net_if->phy_conf.hard_send( if_id, (uint8_t*)arp, sizeof( struct ARP_PACKET ) );
}

/** prepare ARP request packet. */
static void nip_arp_fill_request( struct ARP_PACKET *arp, uint8_t *tpa )
{
	nip_memset ( arp, 0, sizeof( struct ARP_PACKET ) );
	arp->op[1]  = NIP_ARP_OP_REQUEST;
	nip_memcpy( arp->tpa, tpa, NIP_ARP_PR_SIZE );
}

/** Find Transmission with status NIP_NET_IF_RESOLV_ADDR and process if time
 * for retransmission has come or address was found in cache.
 */
void nip_arp_disp_resolve_any( void )
{
	nip_net_trans_id_t id = 0;
	nip_net_if_trans_t *trans;
	struct ARP_CACHE   *cache;

#if NIP_NET_MAX_TRANSMISSIONS > 1
	do
	{
#endif
		trans = &nip_net_trans_list[ id ];
		if ( trans->status == NIP_NET_IF_RESOLV_ADDR )
		{
			if ( trans->time < nip_tickcount()
			  || NIP_SUCCESS == nip_arp_cache_lookup( trans->params.ip.ll_nw_addr, &cache ) )
			{
				// because we cannot be sure that there won't be another transmission
				// waiting for address resolution, we'll set the flag for another
				// check.
				nip_disp_notify( NIP_DISP_CHECK_TRANS );
				nip_disp.next.common.trans = trans;

				// continue with ip_send command
				NIP_CURR_CMD = &nip_ip_disp_send;
				return;
			}
		}
#if NIP_NET_MAX_TRANSMISSIONS > 1
	} while ( ++id < NIP_NET_MAX_TRANSMISSIONS );
#endif
}

/** Retrieve hardware address from ARP cache or initiate ARP request
 *
 * The current transmission's status will be set to NIP_NET_IF_RTS and ip send
 * will be called upon successfull
 */
void nip_arp_disp_resolve( void )
{
	struct ARP_CACHE   *entry;
	struct ARP_PACKET  arp;
	nip_net_if_trans_t *trans = nip_disp.next.common.trans;

	// find entry in arp cache
	if ( NIP_SUCCESS == nip_arp_cache_lookup( trans->params.ip.ll_nw_addr, &entry ) )
	{
		// return hardware address
		nip_memcpy( nip_disp.next.common.remote_hw_addr, entry->hw_addr, NIP_ARP_HW_SIZE );
		trans->status = NIP_NET_IF_RTS;
		NIP_CURR_CMD = &nip_ip_disp_send;
	}
	else
	{
		// update transmission
		trans->retrans_cnt++;
		trans->time = nip_tickcount() + NIP_ARP_RETRY_INTVL;
		if ( trans->retrans_cnt > NIP_ARP_MAX_RETRIES )
		{
			trans->status = NIP_NET_IF_SEND_HNF;
			NIP_CURR_CMD = &nip_ip_disp_send;
		}

		// build and send ARP request
		nip_arp_fill_request( &arp, trans->params.ip.ll_nw_addr );
		nip_arp_fill_defaults_and_send( trans->if_id, &arp );
	}
}

#if NIP_AUTOIP_ENABLE == 1

/** Send ARP probe or announcement packet. IP address will be taken from AutoIP
 * configuration. The configured IP address of NIP_CURR_IF _has_ to be 0.0.0.0
 * for ARP probes, or the current link-local address for ARP announcements. In
 * the latter case it _has_ to match the address contained in the AutoIP-config.
 */
void nip_arp_disp_probe_announcement( void )
{
	struct ARP_PACKET arp;
#if NIP_NET_IF_COUNT > 1
	struct nip_autoip_conf *conf   = & NIP_AUTOIP_CONF( NIP_CURR_IF );
	#define AUTOIP_CONF_ATTR(attr) conf-> attr
#else
	#define AUTOIP_CONF_ATTR(attr) nip_autoip_conf.attr
#endif

	// build and send ARP request
	nip_arp_fill_request( &arp, AUTOIP_CONF_ATTR(addr));
	nip_arp_fill_defaults_and_send( NIP_CURR_IF, &arp );
}

#endif /* NIP_AUTOIP_ENABLE */

/** ARP Receive function to be called by dispatcher.
 *
 * This function will read and evaluate an ARP packet from the network interface
 * which's ID is given in the global nip_disp->param->if_id structure.*/
void nip_disp_arp_receive( void )
{
	// the following variables help reduce program memory usage big time
	// --> arp has local variables worth 37 Bytes.

	nip_net_if_id_t if_id   = NIP_CURR_IF;
	nip_net_if_t    *net_if = &nip_net_if_list[ if_id ];
	struct ARP_PACKET arp;
	struct ARP_CACHE *cache;

	// receive arp packet
	if ( sizeof( struct ARP_PACKET ) != net_if->phy_conf.hard_read( if_id, (uint8_t*) &arp, sizeof( struct ARP_PACKET ) ) )
	{
		// There's something wrong with the packet. It's not large enough.
		// --> forget about it!
		return;
	}

	// check hardware type: Ethernet = 0x0001
	if ( ntohs(arp.hrd) != net_if->phy_conf.type )
		return;

	// check protocol type: IP = 0x0800
	if ( arp.pro != NIP_ARP_PR_TYPE )
		return;

	// read address sizes, Ethernet: 6, IP: 4
	if ( arp.hln != net_if->phy_conf.hw_addr_size || arp.pln != NIP_ARP_PR_SIZE )
		return;

	if ( arp.op[0] == 0x00 )
	{
#if NIP_AUTOIP_ENABLE == 1
		// check ARP packet for IP address conflicts
		nip_autoip_arpcheck( if_id, &arp );
#endif

		// check opcode: 0x0001 = Request
		if ( arp.op[1] == NIP_ARP_OP_REQUEST )
		{
			// check ip
			if ( nip_memcmp( arp.tpa, net_if->ip_conf.addr, 4) == 0  )
			{
				// build reply
				nip_memcpy( arp.tha, arp.sha, NIP_ARP_PR_SIZE + net_if->phy_conf.hw_addr_size );
				arp.op[1]  = NIP_ARP_OP_REPLY;

				nip_arp_fill_defaults_and_send( NIP_CURR_IF, &arp );
			}
		}
		// check opcode: 0x0002 = Response
		else if ( arp.op[1] == NIP_ARP_OP_REPLY )
		{
			///\todo check if we need to check arp.tpa at this point.
			// add response to arp cache. The lookup function will either return
			// the old cache entry for the given protocol address or will return
			// a free slot where the address can be cached.
			nip_arp_cache_lookup( arp.spa, &cache );
			nip_memcpy( cache->hw_addr, arp.sha, arp.hln );
			nip_memcpy( cache->pr_addr, arp.spa, arp.pln );
			cache->timestamp = nip_tickcount() + NIP_ARP_CACHE_LIFE;

			// check if the response helped any pending transmission
			NIP_CURR_CMD = &nip_arp_disp_resolve_any;
		}
	}
}

#endif /* NIP_ARP_ENABLE == 1 */
