/*##############################################################################

nIP - nano IP stack

File        : autoip.c

Description : Implementation of Dynamic Configuration of IPv4 Link-Local
              Addresses (according to RFC 3927)

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#include "dispatcher.h"
#include "inet.h"
#include "net/autoip.h"
#include "net/arp.h"
#include "net/net_if.h"
#include "net/ip.h"
#include "os_core.h"

#if ( NIP_AUTOIP_ENABLE == 1 )


#if NIP_NET_IF_COUNT > 1
	struct nip_autoip_conf nip_autoip_conf[ NIP_NET_IF_COUNT ] ;
	#define NIP_AUTOIP_CONF( if_id ) nip_autoip_conf[if_id]
#else
	struct nip_autoip_conf nip_autoip_conf;
	#define NIP_AUTOIP_CONF( if_id ) nip_autoip_conf
#endif

/** Enable Dynamic Configuration of IPv4 Link-Local Addresses for given
 * interface.
 */
void nip_autoip_start( nip_net_if_id_t net_if )
{
	NIP_AUTOIP_CONF(net_if).state = NIP_AUTOIP_STAT_INIT;
	NIP_AUTOIP_CONF(net_if).t     = nip_tickcount() + NIP_AUTOIP_PROBE_WAIT;
}


/** Function to be called by dispatcher to check AutoIP state
 */
void nip_autoip_disp_check( void )
{
	// use different mechanisms to access autoip configuration for the single and
	// the multi interface cases to minimize use of program memory in either case
#if NIP_NET_IF_COUNT > 1
	struct nip_autoip_conf *conf   = & NIP_AUTOIP_CONF( NIP_CURR_IF );
	#define CONF_ATTR(attr) conf-> attr
#else
	#define CONF_ATTR(attr) nip_autoip_conf.attr
#endif
	nip_time_t             tc = nip_tickcount();
	nip_net_if_t           *net_if   = & nip_net_if_list[ NIP_CURR_IF ];
	uint8_t                *addr_ptr = &net_if->phy_conf.hw_addr[ net_if->phy_conf.hw_addr_size-4];

	// Don't perform any checks, if AutoIP is disabled. Also double check, that
	// we're not interfering with any other IP-setting instance (e.g. DHCP).
	if ( CONF_ATTR(state) == NIP_AUTOIP_STAT_DISABLED
	  || (net_if->ip_conf.state == NIP_IP_CONF_STAT_CONFIGURED
	     && memcmp( net_if->ip_conf.addr, CONF_ATTR(addr), 4 ) != 0)
		)
	{
		return;
	}

	// Upon conflict choose new IP address
	if ( CONF_ATTR(state) == NIP_AUTOIP_STAT_CONFLICT )
	{
		if ( CONF_ATTR(conflict_cnt) < 255 )
			CONF_ATTR(conflict_cnt)++;
		else
			CONF_ATTR(addr)[2]++;

		CONF_ATTR(addr)[3] += addr_ptr[0]+addr_ptr[3];

		// set new timeout for next probe
		if ( CONF_ATTR(conflict_cnt) > NIP_AUTOIP_MAX_CONFLICTS )
			CONF_ATTR(t) += NIP_AUTOIP_RATE_LIMIT_INTERVAL;

		// reset probe mechanism
		CONF_ATTR(cnt)          = 0;
		CONF_ATTR(state)        = NIP_AUTOIP_STAT_PROBE;
		net_if->ip_conf.state   = NIP_IP_CONF_STAT_UNCONFIGURED;
	}

	// all states except CONFLICT which is handled above are time triggered
	if ( tc >= CONF_ATTR(t) )
	{
		switch ( CONF_ATTR( state ) )
		{
			case NIP_AUTOIP_STAT_INIT:
				CONF_ATTR(conflict_cnt) = 0;
				// select IP address based on hardware address
				CONF_ATTR(addr)[0] = 169;
				CONF_ATTR(addr)[1] = 254;
				CONF_ATTR(addr)[2] = addr_ptr[0]+addr_ptr[1];
				CONF_ATTR(addr)[3] = addr_ptr[2]+addr_ptr[3];

				// send probes
				CONF_ATTR(cnt)          = 0;
				CONF_ATTR(state)        = NIP_AUTOIP_STAT_PROBE;
				goto probe;

			case NIP_AUTOIP_STAT_PROBE:
				goto probe;

			case NIP_AUTOIP_STAT_ANNOUNCE:
				goto announce;

			default:
				// don't do anything if autoip is UP or DISABLED
				// just wait for conflicts
				break;
		}
	}

	return;

probe:

	// check counter and switch to ANNOUNCE state if PROBE_NUM has been
	// reached.
	if ( ++CONF_ATTR(cnt) < NIP_AUTOIP_PROBE_NUM )
	{
		// set timeout for next probe
		CONF_ATTR(t) = tc + NIP_AUTOIP_PROBE_MAX;
	}
	else
	{
		// announce new address
		CONF_ATTR(cnt)   = 0;
		CONF_ATTR(t)     = tc + NIP_AUTOIP_ANNOUNCE_WAIT;
		CONF_ATTR(state) = NIP_AUTOIP_STAT_ANNOUNCE;
	}

	// clear IP address and send ARP Probe
	nip_memset( net_if->ip_conf.addr, 0, 4 );
	NIP_CURR_CMD = &nip_arp_disp_probe_announcement;

	return;

announce:

	// check announcement counter
	if ( ++CONF_ATTR(cnt) < NIP_AUTOIP_ANNOUNCE_NUM )
	{
		CONF_ATTR(t) = tc + NIP_AUTOIP_ANNOUNCE_INTERVAL;
	}
	else
	{
		CONF_ATTR(cnt)   = 0;
		CONF_ATTR(state) = NIP_AUTOIP_STAT_UP;
	}

	// set IP address and send ARP announcement
	nip_ip_addr_set( NIP_CURR_IF, CONF_ATTR(addr) );
	NIP_CURR_CMD = &nip_arp_disp_probe_announcement;

	return;
}

/** Check incoming ARP packet for conflict with current link-local configuration
 */
void nip_autoip_arpcheck( nip_net_if_id_t if_id, struct ARP_PACKET *arp )
{
// 	nip_net_if_t    *net_if = &nip_net_if_list[ if_id ];
#if NIP_NET_IF_COUNT > 1
	struct nip_autoip_conf *conf   = & NIP_AUTOIP_CONF( NIP_CURR_IF );
	#define CONF_ATTR(attr) conf-> attr
#else
	#define CONF_ATTR(attr) nip_autoip_conf.attr
#endif

	// if sender address matches own address -> conflict
	if ( nip_memcmp( arp->spa, CONF_ATTR(addr), 4 ) == 0 )
	{
		NIP_AUTOIP_CONF(if_id).state = NIP_AUTOIP_STAT_CONFLICT;
		nip_disp_notify_if( if_id, NIP_DISP_IF_AUTOIP );
	}

}


#endif // NIP_AUTOIP_ENABLE
