/*##############################################################################

nIP - nano IP stack

File        : ip.c

Description : Functions do IP routing

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#############################################################################*/

#include "dispatcher.h"
#include "inet.h"
#include "os_core.h"
#include "net/net_if.h"
#include "net/arp.h"
#include "net/ip.h"
#include "net/udp.h"
#include "app/dhcp.h"
#include "net/autoip.h"
#include "net/tcp.h"
#include "app/mdns.h"

uint16_t nip_ipv4_id;
uint8_t  nip_ip_localhost[4] = {127,0,0,1};
uint8_t  nip_ip_broadcast[4] = {255,255,255,255};
uint8_t  nip_ip_null[4]      = {0,0,0,0};

#if (NIP_IP_ENABLE == 1)

#if (NIP_MULTICAST_ENABLE == 1)
	uint8_t nip_ip_all_hosts[4] = {124,0,0,1};
	#if ( NIP_IP_MAX_HOST_GROUPS == 1 )
	struct nip_ip_group_membership nip_ip_group;
	#else
	struct nip_ip_group_membership nip_ip_groups[ NIP_IP_MAX_HOST_GROUPS ];
	#endif
#endif


#if (NIP_ICMP_ENABLE == 1) && defined ( NIP_ICMP_TEST_NO_MEM_MGR )
struct ICMP_HEADER nip_test_no_mem_icmp;
#endif

// forward declarations
void nip_ip_disp_deliver( void );

/** Calculates internet header checksum in network byte order.
 * @param buf  pointer to buffer to perform checksum calculation on. MUST NOT be NULL.
 * @param size number of bytes to utilize for checksum calculation. Size MUST be
 * an even number greater than 0.
 */
uint16_t nip_internet_checksum( uint16_t *buf, uint16_t size )
{
/*	uint32_t sum32 = 0;
	uint16_t sum16 = 0;
	uint16_t test = 0;
	size /= 2;
	do
	{
		test = ntohs( *buf );
		sum32 += test;
		sum16 += *buf;
		if ( sum16 < *buf) sum16++;
		if ( --size == 0  ) break;
		buf++;
	} while ( 1 );
	sum32  = ( sum32 & 0x0000FFFF) + ( sum32>>16 );
	sum32  = ( sum32 & 0x0000FFFF) + ( sum32>>16 );
	sum32  = ~sum32;
	sum32  = htons( sum32 );
	sum16 = ~sum16;
	return sum16;*/

/*	uint16_t sum = 0;
	uint16_t test = 0;
	size /= 2;
	do
	{
		test = ntohs( *buf );
		sum += test;
		if ( sum <= test ) sum++;
		if ( --size == 0  ) break;
		buf++;
	} while ( 1 );
	sum  = ~sum;
	sum  = htons( sum );
	return sum;*/


	uint16_t sum = 0;
	size /= 2;
	do
	{
		sum += *buf;
		if ( sum < *buf ) sum++;
		buf++;
	} while ( --size > 0 );
	return ~sum;
}

/** Set subnet mask.
 *
 * Broadcast address will be determined and set automatically from the subnet
 * mask.
 */
void nip_ip_subnet_set( nip_net_if_id_t if_id, uint8_t *mask )
{
	nip_net_if_t *net_if = &nip_net_if_list[ if_id ];
	uint8_t      i;

	// set subnet mask
	nip_memcpy( net_if->ip_conf.subnet_mask, mask, 4 );

	// set broadcast address
	i = 0;
	do
	{
		net_if->ip_conf.brdcst_addr[i] = net_if->ip_conf.addr[i] | ~net_if->ip_conf.subnet_mask[i];
	} while ( ++i < 4 );
}

/** Perform tasks, that have to be done, when IP interface goes down, such as
 * resetting TCP connections, sending mDNS notifications,..
 */
void nip_ip_down( nip_net_if_id_t if_id )
{
	nip_net_if_t *net_if = &nip_net_if_list[ if_id ];

	///@todo reset TCP connections

	///@todo send mDNS bye. To do this, this has to be embedded into the
	///dispatcher statemachine somehow.
#if NIP_MDNS_ENABLE == 1
	nip_mdns_interface_down();
#endif

	net_if->ip_conf.state = NIP_IP_CONF_STAT_UNCONFIGURED;
}

/** Perform tasks that have to be done, when IP interface goes up, such as
 * announcing group registrations, performing mDNS checks and lookups,..
 */
void nip_ip_up( nip_net_if_id_t if_id )
{

	// if IGMP is enabled: Report group memberships
#if NIP_IGMP_ENABLE == 1
	nip_ip_query_groups( if_id );
#endif

	// if MDNS is enabled, Report Host name
#if NIP_MDNS_ENABLE == 1
	nip_mdns_interface_up( if_id );
#endif

}

/** Set interface IP address.
 *
 * Subnet mask and broadcast address will be set according to the addresses
 * class. To choose a different subnet mask call nip_ip_subnet_set
 */
void  nip_ip_addr_set( nip_net_if_id_t if_id, uint8_t *addr )
{
	nip_net_if_t *net_if = &nip_net_if_list[ if_id ];

	// Things need only be updated, if ip address changed
	if ( nip_memcmp( addr, net_if->ip_conf.addr, 4 ) != 0 )
	{
		nip_ip_down( if_id );

		nip_memcpy( net_if->ip_conf.addr, addr, 4 );

		// determine subnet mask
		nip_memset( net_if->ip_conf.subnet_mask, 0, 4 );
		net_if->ip_conf.subnet_mask[0] = 255;
		if ( (addr[0] & 0x80) == 0x80 )
			net_if->ip_conf.subnet_mask[1] = 255;
		if ( (addr[0] & 0xC0) == 0xC0 )
			net_if->ip_conf.subnet_mask[2] = 255;

		// we still have to call nip_ip_subnet_set() to determine the broadcast addr
		nip_ip_subnet_set( if_id, net_if->ip_conf.subnet_mask );

		// set ip-conf status
		net_if->ip_conf.state = NIP_IP_CONF_STAT_CONFIGURED;


		if ( net_if->phy_conf.flags & NIP_PHY_IFF_UP )
		{
			nip_ip_up( if_id );
		}
	}
}

/** Try to aquire an IP address using DHCP or AutoIP
 */
void nip_ip_addr_autoconf( nip_net_if_id_t if_id )
{

	nip_ip_down( if_id );

#if NIP_DHCP_ENABLE   == 1
	nip_dhcp_start( if_id );
#endif
#if NIP_AUTOIP_ENABLE == 1
	nip_autoip_start( if_id );
#endif

}

/** Resolve route for packet and return transaction to transmit through. If the
 * trans-parameter is not NULL, that transaction will be used, if routing was
 * successfull.
 *
 * If not given, the network interface for the transmission will be determined
 * and a transaction reserved for it. The transaction will be configured with
 * the local link target, which is either the target host or a local gateway,
 * depending on the routing table.
 *
 * @note The network interface of the transmission will _NOT_ be set for only
 * local targetted transmissions, so don't rely on it!
 *
 * @param ip_addr pointer to ip address of target
 * @param if_id   optional network interface id if you want to restrict routing
 *                on an interface.
 * @param trans   optional pointer to pre selected transaction
 *
 * @return pointer to transaction or NULL on error. On error the nip_error
 * variable will be set and be either
 * - NIP_E_AGAIN            out of mem for transmission, try again later
 * - NIP_E_NO_ROUTE_TO_HOST no interface will deliver to given ip_addr
 *
 * @todo function has not been tested for multiple interface environment.
 */
nip_net_if_trans_t *nip_ip_route( uint8_t *target_addr, nip_net_if_id_t if_id,
	nip_net_if_trans_t *trans )
{
	nip_net_if_id_t    test_if = 0;
	nip_net_if_id_t    transmit_if = NIP_NET_NO_IF; /* interface to transmit on */
	nip_net_if_id_t    gateway_if = NIP_NET_NO_IF;  /* interface with std gateway*/
	nip_net_trans_status_t oldstat;
	nip_net_trans_id_t test_trans;
	uint8_t            cnt,i,b,l;
	uint8_t            *local_addr = NULL;
	uint8_t            *gateway;
	nip_if_ip_conf_t   *ip;
#if NIP_MULTICAST_ENABLE == 1 && NIP_IP_MAX_HOST_GROUPS > 1
	nip_host_group_id_t group;
#endif
#if NIP_AUTOIP_ENABLE == 1
	uint8_t            autoip_target = 0;
#endif

	// Three steps for this function to complete
	// 1. Make sure we have a transmission
	// 2. Determine interface for transmission or local delivery
	// 3. Route determined, finish transmission configuration


	//************************************************************************
	// ** S T E P   1:  make sure we have transmission to work with
	if ( trans == NULL )
	{
		// find idle transaction for interface
		test_trans = 0;
#if ( NIP_NET_IF_TRANS_COUNT > 1 )
		do
		{
#endif
			trans = &nip_net_trans_list[ test_trans ];

#if ( NIP_NET_IF_TRANS_COUNT > 1 )
			if ( trans->status == NIP_NET_IF_IDLE )
			{
				// use this transmission
				break;
			}
		} while ( ++test_trans < NIP_NET_MAX_TRANSMISSIONS );
#endif

		if ( test_trans >= NIP_NET_MAX_TRANSMISSIONS )
		{
			// no transmission found -> try again later
			nip_error = NIP_E_AGAIN;
			return NULL;
		}

		// Reset transmission's content
		nip_memset( trans, 0, sizeof( nip_net_if_trans_t ) );
	}

	// set transmission state
	oldstat = trans->status;
	trans->status = NIP_NET_IF_ROUTING;

	//************************************************************************
	// ** S T E P   2:  Determine interface for transmission or local delivery

	// loopback-check ( see RFC 3330 for information)
	if ( target_addr[0] == 127 )
	{
		trans->flags.deliverlocal = 1;
		nip_memcpy( trans->params.ip.src_nw_addr, nip_ip_localhost, 4 );
	}
	else
	{
	// routing table is implicit as each interface has a subnet mask and possibly
	// a standard gateway.
#if ( NIP_NET_IF_COUNT > 1 )
	do
	{
#endif // ( NIP_NET_IF_COUNT > 1 )

		// multicast or broadcast, use first/selected interface
		/// @todo find way to handle multicast and broadcast on multiple interfaces
		ip = & nip_net_if_list[ test_if ].ip_conf;

		cnt = 0; i = 0; b = 0; l = 0;
		/* Perform subnet mask check on all four bytes of target address. Use
		// i as counter for successfully checked bytes. Also check for possible
		// local network broadcast.*/
		do
		{
			// check subnet
			if ( ( target_addr[cnt] & ip->subnet_mask[cnt]) == (ip->addr[cnt] & ip->subnet_mask[cnt]) )
			{
				i++;
			}

			/* Check for broadcasts. That may be either 255.255.255.255 as global case
			// or x.y.z.255 as special case. Both have in common that subnet-mask and
			// ip address put together will always set all bits to 1.
			// eg:
			//   host-ip:      10.20.144.1
			//   subnet mask:  255.255.248.0
			//   broadcast-ip: 10.20.151.255
			//   --> 255.255.248.0 | 10.20.151.255 == 255.255.255.255*/
			if ( target_addr[cnt] == ip->brdcst_addr[cnt] ) b++;

			// check for locally matching Interface address
			if ( target_addr[cnt] == ip->addr[cnt] ) l++;
		} while ( ++cnt < 4 );

		// local loopback of broadcasts
		if ( nip_memcmp( target_addr, nip_ip_broadcast, 4 ) == 0 || b == 4 )
		{
			l = 4;
		}

		// check multicast addresses
		if ( target_addr[0] >= 0xE0 )
		{
			b = 4;
#if NIP_MULTICAST_ENABLE == 1
	#if NIP_IP_MAX_HOST_GROUPS == 1
			#define GROUP_ATTRi(attr) nip_ip_group.attr
	#else
			group = 0;
			#define GROUP_ATTRi(attr) nip_ip_groups[group].attr
			do {
	#endif
			if ((GROUP_ATTRi(state) != NIP_IP_GROUP_NON
			///\todo include interface check for groups?
			 && nip_memcmp( GROUP_ATTRi(addr), target_addr, 4 ) == 0)
			 || nip_memcmp( nip_ip_all_hosts, target_addr, 4 ) == 0)
			{
				l = 4;
			}
	#if NIP_IP_MAX_HOST_GROUPS != 1
			} while ( ++group < NIP_IP_MAX_HOST_GROUPS );
	#endif
#endif
		}

		/* If the target address was identified as broadcast for any interface
		// or matched it's local address we have to locally deliver the packet.
		// if that happens we have to make sure we have the source address set
		// to that of the interface, just in case of loopback. */
		if ( l == 4 )
		{
			trans->flags.deliverlocal = 1;
			nip_memcpy( trans->params.ip.src_nw_addr, ip->addr, 4 );
		}

		// Check for remote route. If interface was given only check for
		// that interface.
		if ( (if_id == NIP_NET_NO_IF || test_if == if_id) &&  transmit_if == NIP_NET_NO_IF )
		{
			///@todo perfom link-local check
#if NIP_AUTOIP_ENABLE == 1
			if ( target_addr[0] == 169 && target_addr[1] == 254 )
				autoip_target = 1;
#endif
			/* if all 4 address bytes matched the interface's subnet, the packet
			// should be transmitted over this interface. If multiple interfaces match
			// that criteria, select only the first one.*/
			if ( b == 4
#if NIP_AUTOIP_ENABLE == 1
			 || (autoip_target == 1 && target_addr[2] == 255 && target_addr[3] == 255)
#endif
				)
			{
				local_addr  = target_addr;
				transmit_if = test_if;
				trans->flags.ll_broadcast = 1;
			}
			else
			if ( i == 4
#if NIP_AUTOIP_ENABLE == 1
			// if target OR source is local-link address: send locally
			 || autoip_target == 1
			 || (ip->addr[0] == 169 && ip->addr[1] == 254)
#endif
			// if interface has no configured IP address all packets will be sent to
			// local network.
			 || ( if_id != NIP_NET_NO_IF && nip_memcmp( ip->addr, nip_ip_null, 4 ) == 0 ))
			{
				local_addr  = target_addr;
				transmit_if = test_if;
			}
			else
			// subnet-check failed? Check if there is a gateway.
			// Only use the first standard gateway
			if ( ip->std_gateway[0] != 0 && local_addr == NULL)
			{
				gateway    = ip->std_gateway;
				gateway_if = test_if;
			}
		}

#if ( NIP_NET_IF_COUNT > 1 )
	} while ( ++test_if < NIP_NET_IF_COUNT );
#endif

	}

	// found no local target? Check gateway.
	if ( transmit_if == NIP_NET_NO_IF && gateway_if != NIP_NET_NO_IF )
	{
		local_addr  = gateway;
		transmit_if = gateway_if;
	}

	if ( transmit_if != NIP_NET_NO_IF )
	{
		trans->flags.forward = 1;
		nip_memcpy( trans->params.ip.ll_nw_addr , local_addr , 4 );
		nip_memcpy( trans->params.ip.src_nw_addr, ip->addr, 4 );
	}

	// return if interface check failed
	if ( ! (trans->flags.forward || trans->flags.deliverlocal) )
	{
		trans->status = oldstat;
		nip_error = NIP_E_NO_ROUTE_TO_HOST;
		return NULL;
	}

	//************************************************************************
	// ** S T E P   3:  Route determined, finish transmission configuration

	trans->if_id  = transmit_if;
	nip_memcpy( trans->params.ip.dst_nw_addr, target_addr, 4 );
	trans->status = NIP_NET_IF_ROUTED;

	return trans;
}


/** Route and send IP packet */
void nip_ip_disp_send( void )
{
	struct IP_HEADER   head;
	/// @todo use transmissions interface
	nip_net_if_trans_t *trans  = nip_disp.next.common.trans;
	nip_net_if_id_t    if_id;
	nip_net_if_t       *net_if;
	nip_phy_if_t       *phy;
	nip_if_ip_conf_t   *ip;
	uint8_t            *payload, *header;
	uint8_t i;
	uint16_t           len;


	switch( trans->status )
	{
		case NIP_NET_IF_ROUTE:
			/// @todo move this to seperate state to safe stack, because ip_send already
			/// eats up quite alot of stack.
			if ( NULL == nip_ip_route( trans->params.ip.dst_nw_addr, NIP_NET_NO_IF, trans ) )
			{
				// Routing failed. No route to host? We can eliminate the possibility of
				// NIP_E_DEV_BUSY, because we already provided the transmission.
				trans->status = NIP_NET_IF_SEND_NRT;
				goto finish;
			}
		case NIP_NET_IF_RTS:
		case NIP_NET_IF_RESOLV_ADDR:
			// Routing done? -> check if transmission needs to be forwarded on the net
			if ( trans->flags.forward == 0 )      goto deliverlocal;

			// determine interface parameters for remote transmission
			if_id  = trans->if_id;
			net_if = &nip_net_if_list[ if_id ];
			phy    = &net_if->phy_conf;
			ip     = &net_if->ip_conf;

			// transmission still routing or routed but not yet approved to be sent
			if ( trans->status == NIP_NET_IF_RTS ) goto rts;

			break;
		default:
			goto finish;
	}


	/// \todo set resolve timestamp

	/* resolve mac address */
	/// \todo check resolve timestamp and return error if necessary
	// Resolve mac address. We only know how to resolve Ethernet addresses, so
	// lets hope other network adapters don't need a local address.
	if ( phy->type == NIP_PHY_IF_ETH )
	{
		// multicast
		/*
		if ( (trans->params.ip.dst_nw_addr[0] & 0xF0) == 0xE0 )
		{
			/// \todo convert multicast ip to multicast mac
		}
		// broadcast
		else*/
#if (NIP_ARP_ENABLE==1)
		if ( trans->flags.ll_broadcast )
#endif
		{
			i = 0;
			do
				nip_disp.next.common.remote_hw_addr[i] = 0xFF;
			while (++i < 6 );
		}
#if (NIP_ARP_ENABLE==1)
		// unicast
		else
		{
			trans->status = NIP_NET_IF_RESOLV_ADDR;
			NIP_CURR_CMD = &nip_arp_disp_resolve;
			return;
		}
#endif
	}

rts:      /* ready to send */

	// build header
	/// \todo check fragmentation

	len = trans->header_size + trans->payload_size + NIP_IP_DEFAULT_HEADER_SIZE;

	head.ihl          = 5;
	head.version      = 4;
	head.tos          = 0;
	head.total_length = htons( len );
// 	head.id           = trans->params.ip.id;
	head.id           = ++nip_ipv4_id;
	head.frag_offset  = 0;
	head.ttl          = trans->params.ip.ttl>0 ? trans->params.ip.ttl : NIP_IP_TTL;
	head.protocol     = trans->params.ip.protocol;
	head.checksum     = 0x0000;
	nip_memcpy( head.src_addr, trans->params.ip.src_nw_addr, 4 );
	nip_memcpy( head.dst_addr, trans->params.ip.dst_nw_addr, 4 );
	head.checksum     = nip_internet_checksum( (uint16_t*)&head, NIP_IP_DEFAULT_HEADER_SIZE );

#if (NIP_ICMP_ENABLE == 1) && defined (NIP_ICMP_TEST_NO_MEM_MGR)
	if ( head.protocol == NIP_IP_PROTO_ICMP )
		payload = (uint8_t*)&nip_test_no_mem_icmp;
	else
#endif

	if ( trans->payload == NIP_MEM_NULL && trans->header == NIP_MEM_NULL )
	{
		// nothing to send
		goto finish;
	}

	payload = nip_mem_obtain_ptr( trans->payload );
	if ( payload == NULL && trans->payload != NIP_MEM_NULL )
		goto finish;

	header  = nip_mem_obtain_ptr( trans->header  );
	if ( header == NULL && trans->header != NIP_MEM_NULL )
	{
		nip_mem_release_block( trans->payload );
		goto finish;
	}

	if ( payload == NULL && header == NULL )
	{
		// payload invalid? -> cancel transmission
		if ( nip_error == NIP_E_INVALID_BLOCK )
		{
			trans->status = NIP_NET_IF_DONE;
			goto finish;
		}
		// payload locked -> try again later.
		trans->status = NIP_NET_IF_RESOLV_ADDR;
		return;
	}

	// initiate transfer
	phy->hard_send_init(
		if_id,
		NULL,
		htons(0x0800),
		nip_disp.next.common.remote_hw_addr,
		phy->hw_addr,
		len
	);


	/// \todo check success of hardware transmission and react accordingly
	phy->hard_send( if_id, (uint8_t*)&head, NIP_IP_DEFAULT_HEADER_SIZE );
	phy->hard_send( if_id, header , trans->header_size );
	phy->hard_send( if_id, payload + trans->payload_off, trans->payload_size );

#if (NIP_ICMP_ENABLE == 1) && defined (NIP_ICMP_TEST_NO_MEM_MGR)
	if ( head.protocol != NIP_IP_PROTO_ICMP )
#endif
	if ( payload != NULL )
		nip_mem_release_block( trans->payload );
	if ( header  != NULL )
		nip_mem_release_block( trans->header  );


deliverlocal:
	// attempt local delivery of packet
	NIP_CURR_CMD = &nip_ip_disp_deliver;
	return;

finish:
	// clean up after transmission failed or succeeded
	nip_reset_trans( trans );
}



#if (NIP_ICMP_ENABLE == 1)
/** Read ICMP packet from network interface. To be called by dispatcher.
 */
void nip_icmp_disp_receive( void )
{
	struct ICMP_HEADER *icmp;
	nip_net_if_trans_t *trans = nip_disp.next.common.trans;

	// map icmp header onto packet data
#if defined (NIP_ICMP_TEST_NO_MEM_MGR)
	icmp = &nip_test_no_mem_icmp;
#else
	icmp = nip_mem_obtain_ptr( trans->payload );
	if ( icmp == NULL )
	{
		// obtaining pointer impossible. -> drop packet
		/// \todo we could also introduce some timeout and try receiving later,
		/// but this case seems impossible anyways, as we get here straight from
		/// of ip receive (usually) which just successfully obtained that pointer
		/// before.
		goto discard;
		return;
	}
#endif
	// check checksum
	icmp->checksum = nip_internet_checksum( (uint16_t*) icmp,trans->payload_size);
	if ( icmp->checksum != 0x00 )
	{
		// checksum check failed -> dump packet
 		goto discard;
	}

	// check message type
	switch ( icmp->type )
	{
		case NIP_ICMP_T_ECHO_REQ:
			// turn packet into icmp echo reply
			icmp->type = NIP_ICMP_T_ECHO_REP;
			icmp->checksum = 0x00;
			// calculate and set checksum
			icmp->checksum = nip_internet_checksum( (uint16_t*) icmp, trans->payload_size );
			break;

		default:
			// unhandled ICMP messages
			goto discard;
	}

#if !defined (NIP_ICMP_TEST_NO_MEM_MGR)
	// done reading/editing packet
	nip_mem_release_block( trans->payload );
#endif
	// configure transmission
	trans->status             = NIP_NET_IF_ROUTE;
	trans->params.ip.protocol = NIP_IP_PROTO_ICMP;
	nip_memcpy( trans->params.ip.dst_nw_addr, trans->params.ip.src_nw_addr, 4 );

	// send reply
	NIP_CURR_CMD = &nip_ip_disp_send;
	return;

discard:
#if !defined (NIP_ICMP_TEST_NO_MEM_MGR)
	nip_mem_release_block( trans->payload );
#endif
	nip_reset_trans( trans );
	return;
}
#endif // (NIP_ICMP_ENABLE == 1)

/** Read IP data from network interface. To be called by dispatcher. */
void nip_ip_disp_receive( void )
{
	struct IP_HEADER   head;
	nip_net_if_id_t    if_id   = NIP_CURR_IF;
	nip_net_if_t       *net_if = &nip_net_if_list[ if_id ];
	nip_phy_if_t       *phy    = &net_if->phy_conf;
//	nip_if_ip_conf_t   *ip     = &net_if->ip_conf;
//	nip_net_trans_id_t tr = 0;
	nip_net_if_trans_t *trans;
	uint8_t            cnt;
	uint16_t           word;      // 16-bit word for different purposes
	uint16_t           *word_ptr; // pointer to a 16-bit word
	uint16_t           sum;
	cnt = sizeof( struct IP_HEADER );

	// receive ip header
	if ( cnt != phy->hard_read( if_id, (uint8_t*) &head, cnt ) )
	{
		// Something's wrong with the packet. It's not large enough to hold
		// the minimum IP header. -> forget about it.
		return;///\todo add to statistic
	}

	// check IP version. Currently only IPv4 is supported.
	if ( head.version != 0x04 )
	{
		return;///\todo add to statistic
	}
	// check destination
	trans = nip_ip_route( head.dst_addr, NIP_NET_NO_IF, NULL );
	if ( trans == NULL )
		return;///\todo add to statistic

	// configure transmission to be reset after being finished
	trans->flags.reset_status = 1;

	// Check options and calculate header checksum
	cnt  = 0;
	sum  = 0;
	word_ptr = (uint16_t*)&head;
	///\todo evaluate options to find end-of-option option. Options will always
	///have a type and length field, each sized on byte. Only option type 0 and
	///option type 1 have no length field. Quit reading the header when reaching
	///option type 0.
	do
	{
		/// It is possible to calculate the checksum using the network byte order
		/// as that seems to not matter for this specific checksum (carries flow
		/// over on both bytes of the 16-bit words being added).

		if ( cnt >= sizeof ( struct IP_HEADER )/2 )
		{
			word_ptr = &word;
			// options are being discarded and just read for checksum calculation
			if ( phy->hard_read( if_id, (uint8_t*) word_ptr, 2 ) != 2 ) break;
		}

		sum += *word_ptr;

		if ( sum < *word_ptr ) sum++; // carry
		if ( ++cnt >= head.ihl * 2 ) break;
		word_ptr++;

	} while ( 1 );

	if ( sum != 0xFFFF )
	{
		// checksum failed
		goto reset_trans;///\todo add to statistic
	}

	// ignore TTL for not forwarded packets (we don't forward)

	/// \todo Check fragmentation.
	// Drop fragmented packets until they're supported

	// determine payload length
	word = ntohs(head.total_length) - cnt*2;

#if (NIP_ICMP_ENABLE == 1) && defined (NIP_ICMP_TEST_NO_MEM_MGR)
	if ( head.protocol == NIP_IP_PROTO_ICMP )
	{
		word_ptr = (uint16_t*)&nip_test_no_mem_icmp;
	}
	else
	{
#endif
	/*
	// set up transmission for header.
	trans->status = NIP_NET_IF_RECEIVING;
	trans->header = nip_mem_alloc(
		NIP_IP_DEFAULT_HEADER_SIZE,
		NIP_IP_DEFAULT_HEADER_SIZE,
		NIP_MEM_FLG_DELREAD
	);
	*/


	// fill parameters for next call
	nip_memcpy( trans->params.ip.dst_nw_addr, head.dst_addr, 4 );
	nip_memcpy( trans->params.ip.src_nw_addr, head.src_addr, 4 );
	trans->params.ip.protocol = head.protocol;

	// reserve memory for  IP payload
	trans->payload = nip_mem_alloc(
		word,
		word, /* maximize min_length for size check */
		NIP_MEM_FLG_DELREAD,
		NULL
	);

	if ( trans->payload == NIP_MEM_NULL )
	{
		// not enough memory for payload -> drop packet
		goto reset_trans; /// \todo add to statistic
	}

	// Retrieve pointer to read data to
	word_ptr = nip_mem_obtain_ptr( trans->payload );

#if (NIP_ICMP_ENABLE == 1) && defined (NIP_ICMP_TEST_NO_MEM_MGR)
	}
#endif

	// mark transmission as used
	trans->status = NIP_NET_IF_RECEIVING;
	trans->payload_size = word;
	trans->flags.free_payload = 1;
	word = phy->hard_read( if_id, (uint8_t*) word_ptr, trans->payload_size );
	// Read IP Payload
	if ( trans->payload_size != word )
	{
		// reading packet failed: Packet was not as large as IP header said it was
		// -> free and drop any read data
		goto reset_trans;
	}

	// set buffer 'used size'. We have to do that because we have manually added
	// data to the buffer without the memory management's knowledge.
	nip_mem_set_used( trans->payload, trans->payload_size );

	// Release pointer (unlock block). IMPORTANT!!
	nip_mem_release_block( trans->payload );

	// save transmission id for future use
	nip_disp.next.common.trans = trans;

	// deliver payload.
	///@note to enable routing you could put nip_ip_disp_send instead of
	/// nip_ip_disp_deliver. Additional means to reduce ttl and save ip header
	/// should be introduced then, though.
	NIP_CURR_CMD = &nip_ip_disp_deliver;
	return;

	reset_trans:
			nip_reset_trans( trans );
}

/** Deliver IP data to transport layer. Any transmission with the flag
 * deliverlocal not set to 1 will be discarded.
 */
void nip_ip_disp_deliver( void )
{
	nip_net_if_trans_t *trans = nip_disp.next.common.trans;

	if ( trans->flags.deliverlocal == 1 )
	{

		// check protocol field and quit packet reading.
#if (NIP_ICMP_ENABLE == 1)
		if ( trans->params.ip.protocol == NIP_IP_PROTO_ICMP )
		{
			NIP_CURR_CMD = &nip_icmp_disp_receive;
			return;///\todo add to statistic
		}
#endif // (NIP_ICMP_ENABLE == 1)
#if (NIP_UDP_ENABLE == 1)
		else if ( trans->params.ip.protocol == NIP_IP_PROTO_UDP )
		{
			NIP_CURR_CMD = &nip_udp_disp_receive;
			return;///\todo add to statistic
		}
#endif // (NIP_UDP_ENABLE == 1)
#if (NIP_TCP_ENABLE == 1)
		else if ( trans->params.ip.protocol == NIP_IP_PROTO_TCP )
		{
			NIP_CURR_CMD = &nip_tcp_disp_receive;
			return;///\todo add to statistic
		}
#endif // (NIP_TCP_ENABLE == 1)
#if (NIP_IGMP_ENABLE == 1)
		else if ( trans->params.ip.protocol == NIP_IP_PROTO_IGMP )
		{
			NIP_CURR_CMD = &nip_igmp_disp_receive;
			return;///\todo add to statistic
		}
#endif // (NIP_TCP_ENABLE == 1)
	}
	trans->status = NIP_NET_IF_DONE;
	// no protocol handler found? -> discard transmission
	nip_reset_trans( trans );

	return;///\todo add to statistic

}

#if NIP_MULTICAST_ENABLE == 1

#if NIP_IP_MAX_HOST_GROUPS > 1
/** @return pointer to group membership for given interface and group address.
 * If given address is NULL a pointer to an unused membership will be returned.
 * NULL will be returned if the query could not be fulfilled.
 */
struct nip_ip_group_membership *nip_ip_find_group_membership( nip_net_if_id_t if_id, uint8_t *addr )
{
	nip_host_group_id_t group   = 0;
	#define GROUP_ATTRi(attr) nip_ip_groups[group].attr

	do
	{
		// found host group membership?
		if (( addr != NULL
		  && nip_memcmp( addr, GROUP_ATTRi(addr), 4 ) == 0
		  && GROUP_ATTRi(state) != NIP_IP_GROUP_NON
		  && GROUP_ATTRi(if_id) == if_id
		   )
		  ||
		   ( addr == NULL
		  && GROUP_ATTRi(state) == NIP_IP_GROUP_NON
			))
			return &nip_ip_groups[group];

		if ( ++group == NIP_IP_MAX_HOST_GROUPS )
		{
			// host is not in group
			return NULL;
		}
	} while ( 1 );

}
#endif

#if NIP_IGMP_ENABLE == 1

void nip_ip_disp_check_group( void )
{
	nip_mem_handle_t    buf;
	nip_net_if_trans_t  *trans;
	struct nip_igmp_msg *msg;

#if NIP_IP_MAX_HOST_GROUPS == 1
	#define GROUP_ATTRi(attr) nip_ip_group.attr
#else
	nip_host_group_id_t group   = 0;
	#define GROUP_ATTRi(attr) nip_ip_groups[group].attr

	do
	{
#endif
		if ( GROUP_ATTRi(state) == NIP_IP_GROUP_DELAYING )
		{
			// check timer
			if ( GROUP_ATTRi(t_report) <= nip_tickcount() )
			{
				// build report
				buf = nip_mem_alloc(
				  sizeof( struct nip_igmp_msg ),
				  sizeof( struct nip_igmp_msg ),
				  0,
				  NULL
				);

				// not enough memory? -> try again later
				if ( buf == NIP_MEM_NULL )
					return;

				// get transmission
				trans = nip_ip_route( nip_ip_all_hosts, NIP_CURR_IF, NULL );

				// no transmission or route available? -> try again later
				if ( trans == NULL )
				{
					nip_mem_free( buf );
					return;
				}

				// fill report
				msg = nip_mem_obtain_ptr( buf );
				msg->ver_type = NIP_IGMP_VERTYPE_REPORT;
				msg->checksum = 0;
				nip_memcpy( msg->addr, GROUP_ATTRi(addr), 4 );
				msg->checksum = nip_internet_checksum(
				                    (uint16_t *)msg,
				                    sizeof( struct nip_igmp_msg )
				                );
				nip_mem_release_block( buf );

				// add to transmission
				trans->payload = buf;

				// don't loopback, don't route
				trans->flags.deliverlocal = 0;
				trans->params.ip.ttl = 1;

				// transmit
				trans->status = NIP_NET_IF_RESOLV_ADDR;
				nip_disp_notify( NIP_DISP_CHECK_TRANS );

				// reset group membership status
				GROUP_ATTRi(state) = NIP_IP_GROUP_IDLE;
			}
		}

#if NIP_IP_MAX_HOST_GROUPS != 1
	} while  ( ++group < NIP_IP_MAX_HOST_GROUPS );
#endif

}

/** mark all groups memberships on given interface to send a report */
void nip_ip_query_groups( nip_net_if_id_t if_id )
{
	nip_host_group_id_t group   = 0;
#if NIP_IP_MAX_HOST_GROUPS == 1
	#define GROUP_ATTRi(attr) nip_ip_group.attr
#else
	#define GROUP_ATTRi(attr) nip_ip_groups[group].attr

	do
	{
#endif
		if ( GROUP_ATTRi(state) == NIP_IP_GROUP_IDLE && GROUP_ATTRi(if_id) == if_id)
		{
			// find some fairly random value for report-timeout
			GROUP_ATTRi(t_report) = nip_tickcount() + ((&group)[group] % 10);
		}
#if NIP_IP_MAX_HOST_GROUPS != 1
	} while ( ++group < NIP_IP_MAX_HOST_GROUPS );
#endif
}

/** Dispatcher function to be called upon receiving IGMP packet */
void nip_igmp_disp_receive( void )
{
	nip_net_if_trans_t *trans = nip_disp.next.common.trans;
	struct nip_igmp_msg *msg;
#if NIP_IP_MAX_HOST_GROUPS == 1
	#define GROUP_ATTR(attr) nip_ip_group.attr
#else
	struct nip_ip_group_membership *group;
	#define GROUP_ATTR(attr) group->attr
#endif


	msg = nip_mem_obtain_ptr( trans->payload );

	///@todo Check IGMP checksum

	if ( msg->ver_type == NIP_IGMP_VERTYPE_QUERY )
	{
		nip_ip_query_groups( NIP_CURR_IF );
	}
	else
	{
		//stop report timer for the group, if we're member of it.
#if NIP_IP_MAX_HOST_GROUPS == 1
		if ( GROUP_ATTR(state) == NIP_IP_GROUP_DELAYING
		&& nip_memcmp( GROUP_ATTR(addr), msg->addr, 4 ) == 0
		&& GROUP_ATTR(if_id) == NIP_CURR_IF )
			GROUP_ATTR(state) = NIP_IP_GROUP_IDLE;
#else
		group = nip_ip_find_group_membership( NIP_CURR_IF, msg->addr );
		if ( group != NULL )
			GROUP_ATTR(state) = NIP_IP_GROUP_IDLE;
#endif
	}

	nip_mem_release_block( trans->payload );
	nip_reset_trans( trans );
}

#endif /* NIP_IGMP_ENABLE */


/** Join Host Group
 * @return NIP_E_OK or NIP_E_OUT_OF_RESSOURCES. In the latter case you need to
 * increase the value of the preconfigured NIP_IP_MAX_HOST_GROUPS definition.
*/
nip_error_t nip_ip_join_host_group( nip_net_if_id_t if_id, uint8_t *addr )
{
	// find configuration for group membership. This is done differently
	// depending on the NIP_IP_MAX_HOST_GROUPS configuration to save program
	// memory.
#if NIP_IP_MAX_HOST_GROUPS == 1
	#define GROUP_ATTR(attr) nip_ip_group.attr

	if ( GROUP_ATTR(state) != NIP_IP_GROUP_NON
	  && ( GROUP_ATTR(if_id) != if_id
	    || nip_memcmp( addr, GROUP_ATTR(addr), 4 ) != 0)
		)
	return NIP_E_OUT_OF_RESSOURCES;

#else
	struct nip_ip_group_membership *group;
	#define GROUP_ATTR(attr) group->attr

	group = nip_ip_find_group_membership( if_id, addr );
	if ( group == NULL )
		group = nip_ip_find_group_membership( if_id, NULL );

	if ( group == NULL )
		return NIP_E_OUT_OF_RESSOURCES;
#endif


	// set state to join the group on the network if that has not already
	// happened
	if ( GROUP_ATTR(state) == NIP_IP_GROUP_NON )
	{
		// configure group membership
		GROUP_ATTR(if_id) = if_id;
		GROUP_ATTR(member_cnt) = 1;
		GROUP_ATTR(state)      = NIP_IP_GROUP_DELAYING;
		nip_memcpy( GROUP_ATTR(addr), addr, 4 );
#if NIP_IGMP_ENABLE == 1
		GROUP_ATTR(t_report)   = nip_tickcount();
		nip_disp_notify_if( if_id, NIP_DISP_IF_GROUP );
		nip_dispatcher();
#endif
	}
	else
	{
		GROUP_ATTR(member_cnt)++;
	}

	return NIP_E_OK;
}

/** Leave Host Group */
void nip_ip_leave_host_group( nip_net_if_id_t if_id, uint8_t *addr )
{
	// find configuration for group membership. This is done differently
	// depending on the NIP_IP_MAX_HOST_GROUPS configuration to save program
	// memory.
#if NIP_IP_MAX_HOST_GROUPS == 1
	#define GROUP_ATTR(attr) nip_ip_group.attr

	if ( GROUP_ATTR(state) == NIP_IP_GROUP_NON
	  || ( GROUP_ATTR(if_id) != if_id
	    && nip_memcmp( addr, GROUP_ATTR(addr), 4 ) != 0)
		)
	// host was not in group
	return;

#else
	struct nip_ip_group_membership *group;
	#define GROUP_ATTR(attr) group->attr

	group = nip_ip_find_group_membership( if_id, addr );
	if ( group == NULL )
		return;
#endif

	// reduce member count
	if ( --GROUP_ATTR(member_cnt) == 0 )
	{
		// leave group
		GROUP_ATTR(state) = NIP_IP_GROUP_NON;
	}
}

#endif /*NIP_MULTICAST_ENABLE*/

#endif /*NIP_IP_ENABLE*/
