/*##############################################################################

nIP - nano IP stack

File        : net_if.c

Description : Network interface control

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#include "nip_init.h"
#include "net/net_if.h"
#include "net/ethernet.h"
#include "net/ip.h"

#if (NIP_NET_ENABLE == 1)

/** network interface list */
nip_net_if_t       nip_net_if_list[NIP_NET_IF_COUNT];

/** network transmissions */
nip_net_if_trans_t nip_net_trans_list[NIP_NET_MAX_TRANSMISSIONS];


/** Reset transmission. (Free data pointers, clear status,...). The given
 * pointer will not actually be freed from memory.
 */
void nip_reset_trans( nip_net_if_trans_t *trans )
{
	nip_mem_free( trans->header );
	if ( trans->flags.free_payload == 1 )
		nip_mem_free( trans->payload );

	nip_memset( &trans->params, 0, sizeof( nip_net_if_trans_params_t) );

	trans->header       = NIP_MEM_NULL;
	trans->payload      = NIP_MEM_NULL;
	trans->header_size  = 0;
	trans->payload_size = 0;
	trans->retrans_cnt  = 0;

	if ( trans->flags.reset_status == 1 )
		trans->status = NIP_NET_IF_IDLE;

	trans->flags.reset_status = 0;
	trans->flags.free_payload = 0;

}

/** Create item in network interface list. The item configuration will be
 * initialized depending on the given type
 *
 * @return ID of network interface or NIP_NET_IF_ID_UNEFINED
 */
nip_net_if_id_t nip_net_if_create( nip_phy_if_type_t type )
{
	nip_net_if_id_t id;

	// search for free item in network interface list
	id = 0;
	#if NIP_NET_IF_COUNT>1
	do
	{
	#endif
		// found free item?
		if ( nip_net_if_list[id].id == NIP_NET_NO_IF )
		{
			// set item id and type
			nip_net_if_list[id].id = id;
			nip_net_if_list[id].phy_conf.type = type;

			#if NIP_NET_IF_COUNT>1
			// stop searching
			break;
			#endif
		}
	#if NIP_NET_IF_COUNT>1
	} while ( ++id < NIP_NET_IF_COUNT );
	#endif

	// no free item found?
	if ( id >= NIP_NET_IF_COUNT )
		return NIP_NET_NO_IF;

	// reset interface statistics
	nip_net_if_list[id].phy_conf.stats.cnt_discarded = 0;

	// configure item depending on type
	switch ( type )
	{
		// default configuration for ethernet interface
		#if (NIP_ETH_ENABLE==1)
		case NIP_PHY_IF_ETH:
			nip_eth_configure_if_defaults( id );
		break;
		#endif

		default:
			nip_net_if_list[id].phy_conf.flags = NIP_PHY_IFF_NOARP;
			nip_net_if_list[id].phy_conf.open  = NULL;
			nip_net_if_list[id].phy_conf.close = NULL;
			nip_net_if_list[id].phy_conf.hard_send_init = NULL;
			nip_net_if_list[id].phy_conf.hard_send      = NULL;
			nip_net_if_list[id].phy_conf.hard_read_init = NULL;
			nip_net_if_list[id].phy_conf.hard_read      = NULL;
		break;
	}

	return id;
}

/** free network interface
 *
 * The network interface will be brought down and freed.
 */
void nip_net_if_free  ( nip_net_if_id_t id )
{
	// bring down network interface
	if ( nip_net_if_list[id].phy_conf.flags & NIP_PHY_IFF_UP )
	{
		nip_net_if_down( id );
	}
	// unset list item
	nip_net_if_list[id].id = NIP_NET_IF_COUNT;
}

/** set network interface flags
 * @note The network interface has to be closed (down) when calling this function.
 */
nip_success_t nip_set_if_flags ( nip_net_if_id_t id, nip_phy_if_flags_t flags)
{
	if ( nip_net_if_list[id].phy_conf.flags & NIP_PHY_IFF_UP )
	{
		// error if interface is up
		return NIP_ERROR;
	}
	else
	{
		// update flags if interface is down
		nip_net_if_list[id].phy_conf.flags = flags;
	}

	// if UP flag was set, try to open interface
	if ( nip_net_if_list[id].phy_conf.flags & NIP_PHY_IFF_UP )
	{
		return nip_net_if_up( id );
	}
	else
	{
		// flags have been set successfully
		return NIP_SUCCESS;
	}
}


/** return pointer to network interface configuration struct. */
nip_net_if_t *nip_net_if_ptr   ( nip_net_if_id_t id)
{
	return &nip_net_if_list[id];
}

/** Bring up network interface. */
nip_success_t nip_net_if_up    ( nip_net_if_id_t id )
{
	nip_net_if_t *net_if = &nip_net_if_list[id];

	// check if interface's open method has been configured
	if ( net_if->phy_conf.open != NULL )
	{
		// call open method and update IFF_UP-flag
		if ( net_if->phy_conf.open( id ) == NIP_SUCCESS )
			net_if->phy_conf.flags |= NIP_PHY_IFF_UP;
	}

	// return success if interface is up
	if ( net_if->phy_conf.flags & NIP_PHY_IFF_UP )
	{
		// no IP address configured? -> start autoconfiguration of IP address
		#if (NIP_IP_ENABLE==1)
		if ( net_if->ip_conf.state != NIP_IP_CONF_STAT_CONFIGURED )
			nip_ip_addr_autoconf( id );
		else
			nip_ip_up( id );
		#endif

		return NIP_SUCCESS;
	} else
		return NIP_ERROR;
}

/** Bring down network interface. */
nip_success_t nip_net_if_down  ( nip_net_if_id_t id)
{
	nip_net_if_t *net_if = &nip_net_if_list[id];

	// check if interface's close method has been configured
	if ( net_if->phy_conf.open != NULL )
	{
		// call close method and update IFF_UP-flag
		if ( net_if->phy_conf.close( id ) == NIP_SUCCESS )
			net_if->phy_conf.flags &= ~NIP_PHY_IFF_UP;
	}

	// return error if interface is not down
	if ( net_if->phy_conf.flags & NIP_PHY_IFF_UP )
		return NIP_ERROR;
	else
		return NIP_SUCCESS;
}

#endif /* NIP_NET_ENABLE==1 */
