/*##############################################################################

nIP - nano IP stack

File        : test_mem.c

Description : Test nIP memory management

Copyright notice:

Copyright (C) 2007 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##############################################################################*/

#include "nip_init.h"
#include "test/test_mem.h"
#include "mem.h"
#include "nip_error.h"

#ifdef NIP_TEST_MEM

/** performs several memory management checks
 *
 * 1. allocate more than maximum memory
 *
 *
 * @return 0 on SUCCESS or number of check that failed
 */
uint8_t nip_test_mem()
{
	nip_mem_size_t   res_length;
	nip_error_t      res_error;
	nip_mem_handle_t block_id;
	nip_mem_size_t   size, size1, size2, size3;
	nip_mem_size_t   unused;

	res_length = 0;
	block_id   = 0;
	// 1. Allocate memory block with minimum requested size of
	// NIP_MEM_DYNAMIC_SIZE+1. That should fail with NIP_E_OUT_OF_MEMORY error.
	block_id = nip_mem_alloc(
		NIP_MEM_DYNAMIC_SIZE+1, /* wanted size */
		NIP_MEM_DYNAMIC_SIZE+1, /* minimum size */
		0,
		&res_length
		);

	if ( block_id != NIP_MEM_NULL
	  || nip_error != NIP_E_OUT_OF_MEMORY
	  || res_length != 0 )
		return 1;

	res_length = 0;
	block_id   = 0;
	// 2. Allocate memory block with minimum requested size of
	// NIP_MEM_DYNAMIC_SIZE-sizeof(nip_mem_block_t)+1. That should fail with
	// NIP_E_OUT_OF_MEMORY error.
	block_id = nip_mem_alloc(
		NIP_MEM_DYNAMIC_SIZE+1, /* wanted size */
		NIP_MEM_DYNAMIC_SIZE-sizeof(nip_mem_block_t)+1, /* minimum size */
		0,
		&res_length
		);

	if ( block_id != NIP_MEM_NULL
	  || nip_error != NIP_E_OUT_OF_MEMORY
	  || res_length != 0 )
		return 2;

	res_length = 0;
	block_id   = 0;
	// 3. Allocate memory block with minimum requested size of
	// NIP_MEM_DYNAMIC_SIZE-sizeof(nip_mem_block_t), but too large size. That
	// succeed creating a block of minimum size that would take up all space.
	// The correct size has to be returned. The given Block-ID should be 1.
	block_id = nip_mem_alloc(
		NIP_MEM_DYNAMIC_SIZE+1, /* wanted size */
		NIP_MEM_DYNAMIC_SIZE-sizeof(nip_mem_block_t), /* minimum size */
		0,
		&res_length
		);

	if ( block_id != 1
	  || nip_error != NIP_E_OK
	  || res_length != NIP_MEM_DYNAMIC_SIZE-sizeof(nip_mem_block_t) )
		return 3;

	// 4. Allocating any amount of memory should fail now.
	block_id = nip_mem_alloc(10, 10, 0, &res_length );
	if ( block_id != NIP_MEM_NULL
	  || nip_error != NIP_E_OUT_OF_MEMORY
	  || res_length != NIP_MEM_DYNAMIC_SIZE-sizeof(nip_mem_block_t) ) /* should not have changed*/
		return 4;

	// 5. Free allocated memory block 1. No check to that, but as it's simple
	// we'll assume it works.
	nip_mem_free( 1 );

	// 6.-8. Allocate three blocks filling up maximum space.
	block_id   = 0;
	size1 = 0;
	size       =  (NIP_MEM_DYNAMIC_SIZE-3*sizeof(nip_mem_block_t)) * 0.2;
	block_id = nip_mem_alloc(
		size, /* wanted size */
		size, /* minimum size */
		0,
		&size1
		);

	if ( block_id != 1
	  || nip_error != NIP_E_OK
	  || size1 != size )
		return 6;

	block_id   = 0;
	size2 = 0;
	size       =  (NIP_MEM_DYNAMIC_SIZE-3*sizeof(nip_mem_block_t)) * 0.3;
	block_id = nip_mem_alloc(
		size, /* wanted size */
		size, /* minimum size */
		0,
		&size2
		);

	if ( block_id != 2
	  || nip_error != NIP_E_OK
	  || size2 != size )
		return 7;

	block_id   = 0;
	res_length = 0;
	size       =  (NIP_MEM_DYNAMIC_SIZE-3*sizeof(nip_mem_block_t)) * 0.5;
	block_id = nip_mem_alloc(
		size, /* wanted size */
		size, /* minimum size */
		0,
		&size3
		);

	if ( block_id != 3
	  || nip_error != NIP_E_OK
	  || size3 < size )     /* size3 may be larger than expected due to rounding
	                            inaccuracy which may result in one or two unused
	                            bytes to be added to the last block in memory
	                            because they are too few to form a new block. */
		return 8;
	unused = size3 - size;

	// 9. free first block
	nip_mem_free( 1 );

	// 10. Allocate less than the available amount of memory. Since block-1 has
	// bin freed and is therefor unused. That block should be allocated and
	// because there is no unused Id between block-1 and block-2 the newly
	// allocated block will use up all the space of block-1.
	size       = size1 / 2;
	block_id   = 0;
	res_length = 0;
	block_id   = nip_mem_alloc(
		size, /* wanted size */
		size, /* minimum size */
		0,
		&res_length
		);

	if ( block_id != 1
	  || nip_error != NIP_E_OK
	  || res_length != size1 )
		return 10;

	// 11. Free first and second block
	nip_mem_free( 1 );
	nip_mem_free( 2 );

	// 12. Allocate less than the available amount of memory. Since block-1 and
	// block-2 are unused block-1 should be allocated and the remaining memory
	// be put into block-2.
	size       = size1 / 2;
	block_id   = 0;
	res_length = 0;
	block_id   = nip_mem_alloc(
		size, /* wanted size */
		size, /* minimum size */
		0,
		&res_length
	);

	if ( block_id != 1
	  || nip_error != NIP_E_OK
	  || res_length != size )
		return 12;

	// 13. Allocate remaining memory for block 2
	size       = size1 + size2 - size;
	block_id   = 0;
	res_length = 0;
	block_id   = nip_mem_alloc(
		size,
		size,
		0,
		&res_length
	);

	if ( block_id != 2
	  || nip_error != NIP_E_OK
	  || res_length != size )
		return 13;

	// 14. Reduce minimum length of block 1 to 10
	res_error = nip_mem_set_min_length( 1, 10 );
	if ( res_error != NIP_E_OK )
		return 14;

	// 15. Set minimum length of block 1 to maximum possible length
	res_error = nip_mem_set_min_length( 1, size1/2 );
	if ( res_error != NIP_E_OK )
		return 15;

	size1 = size1/2+unused;
	// 16. gather unused bytes
	if ( unused > 0 )
	{
		res_error = nip_mem_set_min_length( 1, size1 );
		if ( res_error != NIP_E_OK )
			return 16;
	}

	// 17. Increase minimum length of block 1 over maximum size
	res_error = nip_mem_set_min_length( 1, size1+1 );
	if ( res_error != NIP_E_OUT_OF_MEMORY )
		return 17;

	// check inserting new block in between two existing blocks
	// 18. free block 2, reduce min-size of block 1 and block 3 (19.)and allocate
	// new block with memory from all the others (20.).
	nip_mem_free( 2 );
	size1 = size1/2;
	res_error = nip_mem_set_min_length( 1, size1 );
	if ( res_error != NIP_E_OK )
		return 18;
	
	// 19. reduce block 3 min-length
	size3 = size3/2;
	res_error = nip_mem_set_min_length( 3, size3 );
	if ( res_error != NIP_E_OK )
		return 19;

	// 20. allocate new block
	size = size1+size3;
	block_id   = 0;
	res_length = 0;
	block_id   = nip_mem_alloc(
		size,
		size,
		0,
		&size2
	);

	if ( block_id != 2
	  || nip_error != NIP_E_OK
	  || size2 < size )
		return 20;


	/// \todo check missing first block-case
	/// \todo check freeing of memory
	/// \todo check locked-block-case
	/// \todo more?


	// check inreasing a block into the space of an unused block
	// 21. free block 2, increase size of block1
	nip_mem_free( 2 );
	res_error = nip_mem_set_min_length( 1, size1+size2 );
	if ( res_error != NIP_E_OK )
		return 21;

	return 0;
}

#endif
